/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>

#include <gdk/gdkprivate.h>
#include "filebrowser.h"
#include "imageubox.h"
#include "imageviewer.h"
#include "gimdb.h"

#include "gtkcompat.h"
#include "profile.h"
#include "prefs.h"
#include "main.h"
#include "browsers.h"

/* 
   Image Browser 
*/

//#define DEBUG

#define DIAPORAMA_TIMEOUT 2000 /* 2 secs */

#define ZOOM_FACTOR 0.2

#define LBL_FONT_HEIGHT_EXTRA 10
#define PROGRESS_UPDATE_INTERVAL_1 0.2
#define PROGRESS_UPDATE_INTERVAL_2 0.03
#define progress_get_value( progress) gtk_progress_get_percentage_from_value( progress, gtk_progress_get_value( progress))

typedef struct _ProgressData ProgressData;
typedef enum {
    IMAGE_BROWSER_SORT_BY_NAME,
    IMAGE_BROWSER_SORT_BY_IMAGE_SIZE,
    IMAGE_BROWSER_SORT_BY_FILE_SIZE,
    IMAGE_BROWSER_REVERSE_SORT
} ImageBrowserSortType;

struct _ProgressData {
    ImageBrowser *browser;
    gfloat percentage;
    gfloat increment;

    gfloat next_update;
};

static void          image_browser_parent_set( GtkWidget *w, GtkWidget *p, ImageBrowser *browser);
static void          image_browser_delete_toplevel( GtkWidget *w, GdkEventAny *event, ImageBrowser *browser);
static void          image_browser_free( ImageBrowser *browser);
static void          image_browser_size_allocate( GtkWidget *w, 
						  GtkAllocation *allocation, 
						  ImageBrowser *browser);
static void          image_browser_set_images_cb( ImageBrowser *browser);
static gint          wait_making_previews_end( ImageBrowser *browser);
static gint          image_browser_update_progress_1( Image *image, gfloat percentage, ProgressData *data);
static gint          image_browser_update_progress_2( Image *image, gfloat percentage, ImageBrowser *browser);
static void          image_browser_make_previews( ImageBrowser *browser, gint page_no);
static void          image_browser_goto_page_cb( ImageBrowser *browser);
static gint          image_browser_button_press( UBox *ubox, 
						 GdkEventButton *event,
						 ImageBrowser *browser);
static void          image_browser_select_image_cb( UBox *ubox, gint cell_no, ImageBrowser *browser);
static void          image_browser_unselect_image_cb( UBox *ubox, gint cell_no, ImageBrowser *browser);
static void          image_browser_toggle_views( ImageBrowser *browser);
static gint          image_browser_diaporama( ImageBrowser *browser);
static void          image_browser_update_status_bar( ImageBrowser *browser);
static void          image_browser_set_sort( ImageBrowser *browser, ImageBrowserSortType type);
static void          image_browser_set_sort_cb( ImageBrowser *browser);


static void          ifact_first_page( GtkWidget *w, ImageBrowser *browser);
static void          ifact_last_page( GtkWidget *w, ImageBrowser *browser);
static void          ifact_previous_page( GtkWidget *w, ImageBrowser *browser);
static void          ifact_next_page( GtkWidget *w, ImageBrowser *browser);
static void          ifact_previous_image( GtkWidget *w, ImageBrowser *browser);
static void          ifact_next_image( GtkWidget *w, ImageBrowser *browser);
static void          ifact_show_preview1( GtkWidget *w, ImageBrowser *browser);
static void          ifact_show_preview2( GtkWidget *w, ImageBrowser *browser);
static void          ifact_show_preview3( GtkWidget *w, ImageBrowser *browser);
static void          ifact_toggle_views( GtkWidget *w, ImageBrowser *browser);
static void          ifact_fullscreen( GtkWidget *w, ImageBrowser *browser);
static void          ifact_sort_by_name( GtkWidget *w, ImageBrowser *browser);
static void          ifact_sort_by_image_size( GtkWidget *w, ImageBrowser *browser);
static void          ifact_sort_by_file_size( GtkWidget *w, ImageBrowser *browser);
static void          ifact_sort_reverse( GtkWidget *w, ImageBrowser *browser);

static void          ifact_zoom_in( GtkWidget *w, ImageBrowser *browser);
static void          ifact_zoom_out( GtkWidget *w, ImageBrowser *browser);
static void          ifact_diaporama( GtkWidget *w, ImageBrowser *browser);
static void          ifact_select_all( GtkWidget *w, ImageBrowser *browser);
static void          ifact_unselect_all( GtkWidget *w, ImageBrowser *browser);
static gint          compare_names( Image *i1, Image *i2);
static gint          compare_dimensions( Image *i1, Image *i2);
static gint          compare_file_size( Image *i1, Image *i2);

static GList*        merge_list( ImageBrowser *browser, GList *a, GList *b);
static GList*        sort_list( ImageBrowser *browser, GList *list, gint num);

static char          *pref_id             = "jview/image_browser";

static char          *pref_preview_width  = "preview width";
static char          *pref_preview_height = "preview height";
static char          *pref_preview_ppage  = "previews per page";
static char          *pref_diapo_timeout  = "diaporama timeout";

ImageBrowser*
image_browser_new( GFunc image_free_func) {
    GtkWidget *hbox;
    GtkWidget *frame;
    ImageBrowser *browser;
    GtkArg *value;
    gint line_height;
    PrefDef main_prefs[] = {
	{ PREF_PUBLIC, pref_preview_width, 
	  "width of the previews" , 
	  "-pw", "--preview-width", 
	  GTK_TYPE_INT, "130"},
	{ PREF_PUBLIC, pref_preview_height,
	  "height of the previews", 
	  "-ph", "--preview-height",
	  GTK_TYPE_INT, "130"},
	{ PREF_PUBLIC, pref_preview_ppage, 
	  "maximum number of preview on each page", 
	  "-pn", "--previews-number",
	  GTK_TYPE_INT, "50"},
	{ PREF_PUBLIC, pref_diapo_timeout,
	  "Time to wait before displaying the next image ( in 1 / 1000 second)",
	  "-dt", "--diaporama-timeout",
	  GTK_TYPE_UINT, "5000"}
    };

    pref_new_from_defs( main_prefs, sizeof( main_prefs) / sizeof( PrefDef), pref_id);

    browser = g_new( ImageBrowser, 1);
    browser->images = NULL;
    browser->image_count = 0;
    value = pref_get_value( pref_preview_ppage, pref_id);
    browser->images_per_page = GTK_VALUE_INT( *value);
    browser->image_free = image_free_func;
    browser->page_no = 0;
    browser->making_previews = FALSE;
    browser->stop_making_previews = FALSE;
    value = pref_get_value( pref_preview_width, pref_id);
    browser->preview_width = GTK_VALUE_INT( *value);
    value = pref_get_value( pref_preview_height, pref_id);
    browser->preview_height = GTK_VALUE_INT( *value);
    browser->image_viewer = image_viewer_new( NULL, 
					      (ImageProgressFunc) image_browser_update_progress_2, 
					      browser);
    browser->compare = (GCompareFunc) compare_names;
    browser->reverse_sort = FALSE;
    value = pref_get_value( pref_diapo_timeout, pref_id);
    browser->diaporama_timeout = GTK_VALUE_UINT( *value);
    browser->diaporama_start.tv_sec = 0;
    browser->diaporama_tag = 0;
    browser->image_no = -1;
    browser->ww = 0;

    browser->mainbox = gtk_vbox_new( FALSE, 0);
    gtk_widget_set_name( browser->mainbox, "ImageBrowser");
        gtk_signal_connect( GTK_OBJECT( browser->mainbox), "parent_set",
			(GtkSignalFunc) image_browser_parent_set, browser);
    hbox = gtk_hbox_new( FALSE, 0);
    gtk_container_add( GTK_CONTAINER( browser->mainbox), hbox);
    gtk_widget_show( hbox);
    
    /* install preferences */

    browser->preview_box = image_ubox_new( browser->preview_width, browser->preview_height);

    gtk_widget_set_name( browser->preview_box, "PreviewBox");
    gtk_container_add( GTK_CONTAINER( hbox), browser->preview_box);
    gtk_widget_show( browser->preview_box);

    browser->viewer_box = gtk_scrolled_window_new( NULL, NULL);
    gtk_widget_set_name( browser->viewer_box, "ViewerBox");
    gtk_container_add( GTK_CONTAINER( hbox), browser->viewer_box);
    gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW( browser->viewer_box),
				    GTK_POLICY_AUTOMATIC,
				    GTK_POLICY_AUTOMATIC);
    gtk_container_add( GTK_CONTAINER( browser->viewer_box), browser->image_viewer);
    gtk_widget_show( browser->image_viewer);
    /*
    sep = gtk_hseparator_new();
    gtk_box_pack_start( GTK_BOX( browser->mainbox), sep, FALSE, TRUE, 2);
    gtk_widget_show( sep);
    */
    hbox = gtk_hbox_new( FALSE, 7);
    gtk_box_pack_end( GTK_BOX( browser->mainbox), hbox, FALSE, TRUE, 2);
    gtk_widget_show( hbox);

    frame = gtk_frame_new( NULL);
    gtk_box_pack_start( GTK_BOX( hbox), frame, FALSE, TRUE, 7);
    gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);
    gtk_widget_show( frame);

   
    line_height = frame->style->font->ascent + frame->style->font->descent + LBL_FONT_HEIGHT_EXTRA;
    
    browser->lbl_image_name = gtk_label_new( "");
    gtk_container_add( GTK_CONTAINER( frame), browser->lbl_image_name);
    gtk_widget_show( browser->lbl_image_name);

    
    frame = gtk_frame_new( NULL);
    gtk_box_pack_start( GTK_BOX( hbox), frame, FALSE, TRUE, 7);
    gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);
    gtk_widget_show( frame);
    
    browser->lbl_image_dim = gtk_label_new( "");
    gtk_container_add( GTK_CONTAINER( frame), browser->lbl_image_dim);
    gtk_widget_show( browser->lbl_image_dim);

    frame = gtk_frame_new( NULL);
    gtk_box_pack_start( GTK_BOX( hbox), frame, FALSE, TRUE, 7);
    gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);
    gtk_widget_show( frame);
    
    browser->lbl_image_size = gtk_label_new( "");
    gtk_container_add( GTK_CONTAINER( frame), browser->lbl_image_size);
    gtk_widget_show( browser->lbl_image_size);

    frame = gtk_frame_new( NULL);
    gtk_box_pack_end( GTK_BOX( hbox), frame, FALSE, TRUE, 7);
    gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);
    gtk_widget_show( frame);
    
    browser->lbl_page_info = gtk_label_new( "");
    gtk_container_add( GTK_CONTAINER( frame), browser->lbl_page_info);
    gtk_widget_show( browser->lbl_page_info);

    image_browser_update_status_bar( browser);

    browser->progress_bar = gtk_progress_bar_new();
    gtk_box_pack_end( GTK_BOX( browser->mainbox), browser->progress_bar, FALSE, TRUE, 2);
    gtk_progress_bar_update( GTK_PROGRESS_BAR( browser->progress_bar), 1.0);
    gtk_widget_set_usize( browser->progress_bar, -1, 8);
    gtk_widget_show( browser->progress_bar);

    gtk_signal_connect( GTK_OBJECT( browser->preview_box), "select_cell",
			(GtkSignalFunc) image_browser_select_image_cb,
			browser);
    gtk_signal_connect( GTK_OBJECT( browser->preview_box), "unselect_cell",
			(GtkSignalFunc) image_browser_unselect_image_cb,
			browser);
    gtk_signal_connect( GTK_OBJECT( browser->preview_box), "button_press_event",
			(GtkSignalFunc) image_browser_button_press, browser);
    ubox_set_selection_mode( UBOX( browser->preview_box), GTK_SELECTION_EXTENDED);


    return browser;
}

static void          
image_browser_parent_set( GtkWidget *w, GtkWidget *p, ImageBrowser *browser) {
    GtkWidget *toplevel;

    toplevel = gtk_widget_get_toplevel( w);
    gtk_signal_connect( GTK_OBJECT( toplevel), "delete_event",
			(GtkSignalFunc) image_browser_delete_toplevel, browser);
}

static void
image_browser_delete_toplevel( GtkWidget *w, GdkEventAny *event, ImageBrowser *browser) {
    browser->stop_making_previews = TRUE;
    browser->wait_callback = image_browser_free;
    browser->callback_data = image_browser_delete_toplevel;
    gtk_idle_add( (GtkFunction) wait_making_previews_end, browser);
}

void
image_browser_destroy( ImageBrowser *browser) {
    browser->stop_making_previews = TRUE;
    browser->wait_callback = image_browser_free;
    browser->callback_data = image_browser_destroy;
    gtk_idle_add( (GtkFunction) wait_making_previews_end, browser);
}
    
static void
image_browser_free( ImageBrowser *browser) {
    GtkWidget *w;

    if ( browser->callback_data == image_browser_delete_toplevel)
	w = gtk_widget_get_toplevel( browser->mainbox);
    else
	w = browser->mainbox;

    if ( browser->diaporama_tag)
	gtk_timeout_remove( browser->diaporama_tag);

    if ( browser->image_free)
	g_list_foreach( browser->images, browser->image_free, NULL);

    if ( browser->images)
	g_list_free( browser->images);

    g_free( browser);

    
    gtk_widget_destroy( w);
}


void
image_browser_set_images( ImageBrowser *browser, GList *images) {
    GList *list = NULL;
    g_return_if_fail( browser != NULL);
     
    browser->image_count = 0;
    browser->image_no = -1;
    
    while ( images) {
	list = g_list_append( list, images->data);
	images = images->next;
	browser->image_count++;
    }

    browser->stop_making_previews = TRUE;
    browser->wait_callback = image_browser_set_images_cb;
    browser->callback_data = list;
    gtk_idle_add( (GtkFunction) wait_making_previews_end, browser);
}


static void
image_browser_set_images_cb( ImageBrowser *browser) {
    /* !!! image shouldn't be freed before that */
    image_ubox_clear( UBOX( browser->preview_box), TRUE);

    if ( browser->images) {
	if ( browser->image_free)
	    g_list_foreach( browser->images, browser->image_free, NULL);
	g_list_free( browser->images);
    }
    
    browser->images = browser->callback_data;
    if ( browser->images)
	browser->images = sort_list( browser, browser->images, browser->image_count);
    
    image_browser_make_previews( browser, 0);
}


void
image_browser_append_image( ImageBrowser *browser, Image *image) {
    gint last_page;

    g_return_if_fail( browser != NULL);
    g_return_if_fail( image != NULL);
    
    browser->images = g_list_append( browser->images, image);
    last_page = browser->page_no == browser->image_count / browser->images_per_page;
    if ( !browser->making_previews && last_page &&
	 UBOX( browser->preview_box)->child_count < browser->images_per_page)
	image_ubox_append( UBOX( browser->preview_box), image, NULL, NULL);
}


static gint
wait_making_previews_end( ImageBrowser *browser) {
    if ( !browser->making_previews) {
	browser->stop_making_previews = FALSE;
	( *browser->wait_callback) ( browser);
	return FALSE;
    }

    return TRUE;
}

static gint
image_browser_update_progress_1( Image *image, gfloat percentage, ProgressData *data) {
  
    if ( percentage == 0.0) {
	data->next_update = PROGRESS_UPDATE_INTERVAL_1;
	
    } else if ( percentage == 1.0) {
	data->percentage += data->increment;
	if ( data->percentage < 1.0) {
	    gtk_progress_set_percentage( GTK_PROGRESS( data->browser->progress_bar), 
					 data->percentage);
	    
	    while ( gtk_events_pending())
		gtk_main_iteration();
	    
	    if ( data->browser->stop_making_previews)
		return FALSE;
	}

    } else if ( percentage > data->next_update) { 
	gfloat p;

	data->next_update += PROGRESS_UPDATE_INTERVAL_1;
	
	p = data->percentage + data->increment * percentage;
	if ( p < 1.0) {
	    gtk_progress_set_percentage( GTK_PROGRESS( data->browser->progress_bar), p);
	    
	    while ( gtk_events_pending())
		gtk_main_iteration();
	    
	    if ( data->browser->stop_making_previews)
		return FALSE;
	}
    } 
    
    return TRUE;
}

static gint
image_browser_update_progress_2( Image *image, gfloat percentage, ImageBrowser *browser) {

    if ( percentage == 0.0) {
	gtk_progress_set_percentage( GTK_PROGRESS( browser->progress_bar), 0.0);

	browser->diaporama_start.tv_sec = 0;

    } else if ( percentage == 1.0) {
	gtk_progress_set_percentage( GTK_PROGRESS( browser->progress_bar), 1.0);

	if ( browser->diaporama_tag)
	    gettimeofday( &browser->diaporama_start, NULL);

    } else {
	if ( percentage - progress_get_value( GTK_PROGRESS( browser->progress_bar)) >
	     PROGRESS_UPDATE_INTERVAL_2) {
	    gtk_progress_set_percentage( GTK_PROGRESS( browser->progress_bar), percentage);
	    
	    while ( gtk_events_pending())
		gtk_main_iteration();
	}
    }

    return TRUE;
}

    
static void
image_browser_make_previews( ImageBrowser *browser, gint page_no) {
    Image *image;
    GList *item;
    gint count, pcount;
    ProgressData progress;

    g_return_if_fail( browser != NULL);

    browser->making_previews = TRUE;
    browser->page_no = page_no;    
    browser->image_no = -1;
    item = g_list_nth( browser->images, page_no*browser->images_per_page);
    if ( !item)
	item = browser->images;

    /* pcount = Predicted count (or Progress count ?) :
       Guess how much images will be appended if there's no 
       call to image_browser_append_image for progress monitoring*/
    if ( browser->page_no == browser->image_count / browser->images_per_page)
	pcount = g_list_length( item);
    else
	pcount = browser->images_per_page;

    count = browser->images_per_page;
    progress.browser = browser;
    progress.percentage = 0.0;
    progress.increment = 1.0 / pcount;

    image_browser_update_status_bar( browser);
    gtk_progress_bar_update( GTK_PROGRESS_BAR( browser->progress_bar), 0.0);

    while ( item && count) {
	image = item->data;
	item = item->next;

	image_ubox_append( UBOX( browser->preview_box), image, 
			   (ImageProgressFunc) image_browser_update_progress_1, 
			   &progress);

	if ( browser->stop_making_previews)
	    break;

	count--;

    }
    gtk_progress_bar_update( GTK_PROGRESS_BAR( browser->progress_bar), 1.0);


    browser->making_previews = FALSE;

}

void
image_browser_goto_page( ImageBrowser *browser, gint page) {
    gint page_count;

    g_return_if_fail( browser != NULL);
   
    switch ( page) {
    case FIRST_PAGE:
	browser->page_no = 0;
	break;
    case LAST_PAGE:
	browser->page_no = browser->image_count / browser->images_per_page;
	break;
    case NEXT_PAGE:
	page_count = browser->image_count / browser->images_per_page;
	browser->page_no = MIN( page_count, browser->page_no + 1);
	break;
    case PREVIOUS_PAGE:
	browser->page_no = MAX( 0, browser->page_no - 1);
	break;
    default :
	g_return_if_fail( page >= 0);
    }


    browser->stop_making_previews = TRUE;
    browser->wait_callback = image_browser_goto_page_cb;
    gtk_idle_add( (GtkFunction) wait_making_previews_end, browser);

}

static void
image_browser_goto_page_cb( ImageBrowser *browser) {
    image_ubox_clear( UBOX( browser->preview_box), TRUE);

    image_browser_make_previews( browser, browser->page_no);
}

static void          
image_browser_select_image_cb( UBox *ubox, 
			       gint cell_no, 
			       ImageBrowser *browser) {
    g_return_if_fail( ubox != NULL);
    g_return_if_fail( cell_no >= 0);
    g_return_if_fail( browser != NULL);


}

static gint
image_browser_button_press( UBox *ubox, 
			    GdkEventButton *event,
			    ImageBrowser *browser) {
    UBoxCell *cell;
    gint cell_no;
    
    g_return_val_if_fail( browser != NULL, FALSE);
    if ( event->window != ubox->cells_area)
	return FALSE;

    cell_no = ubox_cell_from_coord( ubox, event->x, event->y);
    cell = ubox_get_cell( ubox, cell_no);
    if ( cell) {
	if ( event->type == GDK_2BUTTON_PRESS || event->button == 2) {
	    if ( !GTK_WIDGET_VISIBLE( browser->viewer_box))
		image_browser_toggle_views( browser);

	    
	    browser->image_no = cell_no + browser->page_no * browser->images_per_page;
	    image_browser_update_status_bar( browser);
	    image_viewer_set_image( IMAGE_VIEWER(browser->image_viewer), 
				    (Image*) cell->data);
	}
    }
    
    return FALSE;
}


static void          
image_browser_unselect_image_cb( UBox *ubox, 
				 gint cell_no, 
				 ImageBrowser *browser) {
    g_return_if_fail( ubox != NULL);
    g_return_if_fail( cell_no >= 0);
    g_return_if_fail( browser != NULL);


}

void
image_browser_view_image( ImageBrowser *browser,
			    gint image) {
    UBoxCell *cell;

    if ( !GTK_WIDGET_VISIBLE( browser->viewer_box)) 
	image_browser_toggle_views( browser);

    switch ( image) {
    case NEXT_IMAGE:
	if ( browser->image_no == -1)
	    browser->image_no = browser->page_no * browser->images_per_page;
	else if ( browser->image_no < browser->image_count) {
	    browser->image_no ++;
	    if ( browser->image_no >= (browser->page_no + 1)*browser->images_per_page) {
		 image_browser_goto_page( browser, NEXT_PAGE);
		 break;
	    }
	} else
	    return;

	break;
    case PREVIOUS_IMAGE:
	if ( browser->image_no == -1)
	    browser->image_no = browser->page_no * browser->images_per_page;
	else if ( browser->image_no > 0) {
	    browser->image_no --;
	    if ( browser->image_no < browser->page_no*browser->images_per_page) {
		 image_browser_goto_page( browser, PREVIOUS_PAGE);
		 break;
	    }

	} else
	    return;

	break;
    default:
	if ( image < 0 && image >= browser->image_count)
	    return;

	browser->image_no = image;
	break;
    }

    cell = ubox_get_cell( UBOX( browser->preview_box), 
			  browser->image_no - browser->page_no*browser->images_per_page);
    if ( cell) {
	ubox_moveto( UBOX( browser->preview_box), cell);
	image_viewer_set_image( IMAGE_VIEWER(browser->image_viewer), 
				(Image*) cell->data);
	image_browser_update_status_bar( browser);
    }
}

static void
image_browser_toggle_views( ImageBrowser *browser) {
    g_return_if_fail( browser != NULL);

    if ( GTK_WIDGET_VISIBLE( browser->viewer_box)) {
	gtk_widget_hide( browser->viewer_box);
	gtk_box_set_child_packing( GTK_BOX( browser->preview_box->parent),
				   browser->preview_box,
				   TRUE, TRUE, 0, GTK_PACK_START);
    } else {
	UBox *ubox;
	GtkRequisition requisition;
	ubox = UBOX( browser->preview_box);

	gtk_widget_show( browser->viewer_box);
	gtk_box_set_child_packing( GTK_BOX( browser->preview_box->parent),
				   browser->preview_box,
				   FALSE, FALSE, 0, GTK_PACK_END);

	gtk_widget_size_request( browser->preview_box, &requisition);
	gtk_widget_set_usize( browser->preview_box,
			      requisition.width,
			      -1);
    }
}

static void
image_browser_set_sort( ImageBrowser *browser, ImageBrowserSortType type) {
    switch( type) {
    case IMAGE_BROWSER_SORT_BY_NAME:
	browser->compare = (GCompareFunc) compare_names;
	break;
    case IMAGE_BROWSER_SORT_BY_IMAGE_SIZE:
	browser->compare = (GCompareFunc) compare_dimensions;
	break;
    case IMAGE_BROWSER_SORT_BY_FILE_SIZE:
	browser->compare = (GCompareFunc) compare_file_size;
	break;
    case IMAGE_BROWSER_REVERSE_SORT:
	browser->reverse_sort = !browser->reverse_sort;
	break;
    default:
	return;
    }

    browser->stop_making_previews = TRUE;
    browser->wait_callback =  image_browser_set_sort_cb;
    gtk_idle_add( (GtkFunction) wait_making_previews_end, browser);
}

static void
image_browser_set_sort_cb( ImageBrowser *browser) {

    if ( browser->images)
	browser->images = sort_list( browser, browser->images, browser->image_count);
    
    image_browser_make_previews( browser, 0);
}


static gint
image_browser_diaporama( ImageBrowser *browser) {
    struct timeval t;
    

    if ( !browser->making_previews && browser->diaporama_start.tv_sec) {
	g_return_val_if_fail( !gettimeofday( &t, NULL), FALSE);
	if ( PROF_ELAPSED_MSEC( browser->diaporama_start, t) > browser->diaporama_timeout) {

	    if ( browser->image_no >= browser->image_count) {
		browser->diaporama_tag = 0;
		return FALSE;
	    }
	    image_browser_view_image( browser, NEXT_IMAGE);
	    
	}
    }

    return TRUE;
}

    
static void
image_browser_update_status_bar( ImageBrowser *browser) {
    Image *image;
    gchar buff[32];

    image = g_list_nth_data( browser->images, browser->image_no);
    if ( image) {
	gtk_label_set( GTK_LABEL( browser->lbl_image_name), image->file_name);
	sprintf( buff, "%dx%dx%d",
		 image->width, image->height, image->flags & IMAGE_HAS_COLORS ? 24 : 8);
	gtk_label_set( GTK_LABEL( browser->lbl_image_dim), buff);
	if ( image->file_size > 4096)
	    sprintf( buff, "%d Kb", image->file_size / 1024);
	else
	    sprintf( buff, "%d bytes", image->file_size);
	gtk_label_set( GTK_LABEL( browser->lbl_image_size), buff);
    }
    sprintf( buff, "Page %d of %d", 
	     browser->page_no + 1, 1 + browser->image_count / browser->images_per_page);
    gtk_label_set( GTK_LABEL( browser->lbl_page_info), buff);
}

GtkItemFactory*
image_browser_get_menu_factory( ImageBrowser *browser) {
    GtkItemFactory *item_factory;

    GtkItemFactoryEntry menu_entries[] = {
    {"/_First page"              , "<control>F"          , ifact_first_page        , 2, NULL           },
    {"/_Last page"               , "<control>L"          , ifact_last_page         , 2, NULL           },
    {"/_Next page"               , "<control>N"          , ifact_next_page         , 2, NULL           },
    {"/_Previous page"           , "<control>P"          , ifact_previous_page     , 2, NULL           },
    /*
    {"/"                         , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/sort by/_name"            , NULL                  , ifact_sort_by_name      , 2, "<RadioItem>"  },
    {"/sort by/_image size"      , NULL                  , ifact_sort_by_image_size, 2, "/sort by/name"  },
    {"/sort by/_file size"       , NULL                  , ifact_sort_by_file_size , 2, "/sort by/name"  },
    {"/sort by/_reverse"         , NULL                  , ifact_sort_reverse      , 2, "/sort by/name"  },
    */
    {"/"                         , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/show preview"             , NULL                  , ifact_show_preview1     , 2, "<RadioItem>"  },
    {"/show list"                , NULL                  , ifact_show_preview2     , 2, "/show preview"},
    {"/show long list"           , NULL                  , ifact_show_preview3     , 2, "/show preview"},
    {"/"                         , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/Next image"               , "<shift><control>N"   , ifact_next_image        , 2, NULL           },
    {"/Previous image"           , "<shift><control>P"   , ifact_previous_image    , 2, NULL           },
    {"/"                         , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/Zoom _in"                 , "<control>I"          , ifact_zoom_in           , 2, NULL           },
    {"/Zoom _out"                , "<control>O"          , ifact_zoom_out          , 2, NULL           },
    {"/"                         , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/_Diaporama"               , "<control>D"          , ifact_diaporama         , 2, NULL           },
    {"/"                         , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/_Toggle views"            , "<control>T"          , ifact_toggle_views      , 2, NULL           },
    {"/Toggle fullscreen"        , "<control>K"          , ifact_fullscreen        , 2, NULL           },
    {"/"                         , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/Select _all"              , "<control>A"          , ifact_select_all        , 2, NULL           },
    {"/Unselect all"             , NULL                  , ifact_unselect_all      , 2, NULL           } };

    g_return_val_if_fail( browser != NULL, NULL);


    item_factory = gtk_item_factory_new( GTK_TYPE_MENU, "<ImageBrowser>", NULL);
    gtk_item_factory_create_items_ac( item_factory, 
				      sizeof( menu_entries) / sizeof( GtkItemFactoryEntry),
				      menu_entries,
				      browser, 
				      2);

    return item_factory;
}

static void         
ifact_first_page( GtkWidget *w, ImageBrowser *browser) {
    image_browser_goto_page( browser, FIRST_PAGE);
}

static void         
ifact_last_page( GtkWidget *w, ImageBrowser *browser) {
    image_browser_goto_page( browser, LAST_PAGE);
}

static void         
ifact_previous_page( GtkWidget *w, ImageBrowser *browser) {
    image_browser_goto_page( browser, PREVIOUS_PAGE);
}

static void         
ifact_next_page( GtkWidget *w, ImageBrowser *browser) {
    image_browser_goto_page( browser, NEXT_PAGE);
}

static void         
ifact_previous_image( GtkWidget *w, ImageBrowser *browser) {
    image_browser_view_image( browser, PREVIOUS_IMAGE);
}

static void         
ifact_next_image( GtkWidget *w, ImageBrowser *browser) {
    image_browser_view_image( browser, NEXT_IMAGE);
}

static void
image_browser_set_preview_type_cb( ImageBrowser *browser) {
    gint preview_type;

    preview_type = GPOINTER_TO_INT( browser->callback_data);
    image_ubox_set_type( UBOX( browser->preview_box), preview_type);

    image_ubox_clear(  UBOX( browser->preview_box), FALSE);
    image_browser_make_previews( browser, browser->page_no);

    if ( GTK_WIDGET_VISIBLE( browser->viewer_box)) {
	GtkRequisition requisition;
	gtk_widget_size_request( browser->preview_box, &requisition);
	
	/* xxx don't work */
	gtk_widget_set_usize( browser->preview_box, requisition.width, -1);
    }

}

static void          
ifact_show_preview1( GtkWidget *w, ImageBrowser *browser) {
    if ( !GTK_CHECK_MENU_ITEM( w)->active)
	return;

    browser->stop_making_previews = TRUE;
    browser->wait_callback = image_browser_set_preview_type_cb;
    browser->callback_data = GINT_TO_POINTER( IMAGE_UBOX_PREVIEW1);
    gtk_idle_add( (GtkFunction) wait_making_previews_end, browser);
}

static void          
ifact_show_preview2( GtkWidget *w, ImageBrowser *browser) {
    if ( !GTK_CHECK_MENU_ITEM( w)->active)
	return;

    browser->stop_making_previews = TRUE;
    browser->wait_callback = image_browser_set_preview_type_cb;
    browser->callback_data = GINT_TO_POINTER( IMAGE_UBOX_PREVIEW2);
    gtk_idle_add( (GtkFunction) wait_making_previews_end, browser);
}

static void          
ifact_show_preview3( GtkWidget *w, ImageBrowser *browser) {
    if ( !GTK_CHECK_MENU_ITEM( w)->active)
	return;

    browser->stop_making_previews = TRUE;
    browser->wait_callback = image_browser_set_preview_type_cb;
    browser->callback_data = GINT_TO_POINTER( IMAGE_UBOX_PREVIEW3);
    gtk_idle_add( (GtkFunction) wait_making_previews_end, browser);
}



static void          
ifact_sort_by_name( GtkWidget *w, ImageBrowser *browser) {
    image_browser_set_sort( browser, IMAGE_BROWSER_SORT_BY_NAME);
}

static void          
ifact_sort_by_image_size( GtkWidget *w, ImageBrowser *browser) {
    image_browser_set_sort( browser, IMAGE_BROWSER_SORT_BY_IMAGE_SIZE);
}

static void          
ifact_sort_by_file_size( GtkWidget *w, ImageBrowser *browser) {
    image_browser_set_sort( browser, IMAGE_BROWSER_SORT_BY_FILE_SIZE);
}

static void          
ifact_sort_reverse( GtkWidget *w, ImageBrowser *browser) {
    image_browser_set_sort( browser, IMAGE_BROWSER_REVERSE_SORT);
}


static void         
ifact_toggle_views( GtkWidget *w, ImageBrowser *browser) {

    image_browser_toggle_views( browser);
}


/* xxx BUGGY 
 */
static void         
ifact_fullscreen( GtkWidget *w, ImageBrowser *browser) {
    GtkWidget *top_widget;

    top_widget = gtk_widget_get_toplevel( browser->preview_box); 
  
    if ( browser->ww) {
	
	gtk_widget_set_usize( top_widget, browser->ww, browser->wh);
	gtk_widget_set_uposition( top_widget, browser->wx, browser->wy);

	browser->ww = 0;

    } else {
	browser->wx = top_widget->allocation.x;
	browser->wy = top_widget->allocation.y;
	browser->ww = top_widget->allocation.width;
	browser->wh = top_widget->allocation.height;
	
	gtk_widget_set_uposition( top_widget, 0, 0);
	gtk_widget_set_usize( top_widget, gdk_screen_width(), gdk_screen_height());
    }
	
}
    
static void         
ifact_zoom_in( GtkWidget *w, ImageBrowser *browser) {
    float scale_factor;
    
    scale_factor = image_viewer_get_scale_factor( IMAGE_VIEWER( browser->image_viewer));
    scale_factor *=  1 + ZOOM_FACTOR;
    image_viewer_set_scale_factor( IMAGE_VIEWER( browser->image_viewer),  scale_factor);
}
   
static void         
ifact_zoom_out( GtkWidget *w, ImageBrowser *browser) {
    float scale_factor;
    
    scale_factor = image_viewer_get_scale_factor( IMAGE_VIEWER( browser->image_viewer));
    scale_factor *= 1 - ZOOM_FACTOR;
    image_viewer_set_scale_factor( IMAGE_VIEWER( browser->image_viewer),  scale_factor);
}

static void         
ifact_diaporama( GtkWidget *w, ImageBrowser *browser) {
    if ( !browser->diaporama_tag && !GTK_WIDGET_VISIBLE( browser->image_viewer))
	image_browser_toggle_views( browser);

    if ( browser->diaporama_tag) {
	gtk_timeout_remove( browser->diaporama_tag);
	browser->diaporama_tag = 0;
    } else {
	gint timeout;

	timeout = MIN( browser->diaporama_timeout, 100);
	browser->diaporama_tag = gtk_timeout_add( timeout, (GtkFunction) image_browser_diaporama, browser);
	image_browser_view_image( browser, NEXT_IMAGE);
    }
}

static void         
ifact_select_all( GtkWidget *w, ImageBrowser *browser) {
    UBoxCell *cell;
    GSList *list;
    
    list = UBOX( browser->preview_box)->children;
    while ( list) {
	cell = list->data;
	list = list->next;

	ubox_select_cell( UBOX( browser->preview_box), cell);
    }

}

static void         
ifact_unselect_all( GtkWidget *w, ImageBrowser *browser) {
    UBoxCell *cell;
    GList *list;
    
    list = ubox_get_selection( UBOX( browser->preview_box));
    while ( list) {
	cell = list->data;
	list = list->next;

	ubox_unselect_cell( UBOX( browser->preview_box), cell);
    }

}

static gint          
compare_names( Image *i1, Image *i2) {
    gint cmp;
    
    cmp = strcmp( i1->path, i2->path);
    if ( cmp)
	return cmp;

    return strcmp( i1->file_name, i2->file_name);
}

static gint          
compare_dimensions( Image *i1, Image *i2) {
    gint cmp;

    cmp = (i1->flags & IMAGE_HAS_COLORS) - (i2->flags & IMAGE_HAS_COLORS);
    if ( cmp)
	return cmp;
    
    return i1->width * i1->height - i2->width * i2->height;
}

static gint          
compare_file_size( Image *i1, Image *i2) {
    return i1->file_size - i2->file_size;
}



/* adapted from gtkclist.c */
static GList*
merge_list( ImageBrowser *browser, GList *a, GList *b) {
    GList z = { 0 };                    /* auxiliary node */
    GList *c;
    gint cmp;
    
    c = &z;
    
    while (a || b) {
	if (a && !b) {
	    c->next = a;
	    a->prev = c;
	    c = a;
	    a = a->next;
	    break;
	} else if (!a && b) {
	    c->next = b;
	    b->prev = c;
	    c = b;
	    b = b->next;
	    break;
	} else /* a && b */ {
	    cmp = (*browser->compare) ( a->data, b->data);
	    if ((cmp >= 0 && browser->reverse_sort) ||
		(cmp <= 0 && !browser->reverse_sort)) {
		c->next = a;
		a->prev = c;
		c = a;
		a = a->next;
	    } else {
		c->next = b;
		b->prev = c;
		c = b;
		b = b->next;
	    }
	}
    }
    
    return z.next;
}


static GList*
sort_list( ImageBrowser *browser, GList *list, gint num) {
    GList *half;
    gint i;
    
    if (num == 1) {
	return list;
    } else {
	/* move "half" to the middle */
	half = list;
	for (i = 0; i < num / 2; i++)
	    half = half->next;
	
	/* cut the list in two */
	half->prev->next = NULL;
	half->prev = NULL;
	
	/* recursively sort both lists */
	return merge_list( browser,
			   sort_list( browser, list, num / 2),
			   sort_list( browser, half, num - num / 2));
    }
}


/* 
   Directory Browser
*/


static void         destroy_window_cb( GtkWidget *w, DirectoryBrowser *browser);
static void         directory_browser_select_file( FileBrowser *file_browser, 
						   char *directory, 
						   DirectoryBrowser *browser);

static void         ifact_quit( GtkWidget *w, DirectoryBrowser *browser);
static void         ifact_new_database( GtkWidget *w, DirectoryBrowser *browser);
static void         ifact_new_database2( GtkWidget *w, DirectoryBrowser *browser);
static void         ifact_load_database( GtkWidget *w, DirectoryBrowser *browser);
static void         ifact_load_database2( GtkWidget *w, DirectoryBrowser *browser);
static GtkWidget*   database_list_new( DirectoryBrowser *browser, GtkSignalFunc ok_func, gpointer func_data);
static GImDB*       database_list_get_database( GtkWidget *w);
static void         ifact_add_directory( GtkWidget *w, DirectoryBrowser *browser);
static void         ifact_add_directory2( GtkWidget *w, DirectoryBrowser *browser);
static void         ifact_add_selection( GtkWidget *w, DirectoryBrowser *browser);
static void         ifact_add_selection2( GtkWidget *w, DirectoryBrowser *browser);

static void         ifact_remove_database( GtkWidget *w, GImDB *gimdb);


static GList *gimdb_list = NULL;

static char  *pref_window_width  = "window width";
static char  *pref_window_height = "window height";

GtkWidget*
directory_browser_new( char *starting_directory,
		       gint preview_width,
		       gint preview_height,
		       gint images_per_page) {
    DirectoryBrowser *browser;
    GtkWidget *window;
    GtkWidget *mainbox;
    GtkWidget *menubar;
    GtkWidget *image_menu;
    GtkWidget *menuitem;
    GtkWidget *hpaned;
    GtkWidget *frame;
    GtkItemFactory *menubar_factory, *menu_factory;
    GtkArg *value;
    PrefDef   main_prefs[] = {
	{ PREF_PUBLIC, pref_window_width,
	  "width of the browser window", 
	  "-bw", "--browser-width",
	  GTK_TYPE_INT, "640"},
	{ PREF_PUBLIC, pref_window_height,
	  "height of the browser window", 
	  "-bh", "--browser-height",
	  GTK_TYPE_INT, "480"}
    };
    GtkItemFactoryEntry menu_entries[] = {
    {"/File/"                        , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/File/Quit"                    , "<control>Q"          , ifact_quit              , 2, NULL           },
    {"/Database/New"                 , NULL                  , ifact_new_database      , 2, NULL           },
    {"/Database/Load"                , NULL                  , ifact_load_database     , 2, NULL           },
    {"/Database/"                    , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/Database/Add selection"       , NULL                  , ifact_add_selection     , 2, NULL           },
    {"/Database/Add directory"       , NULL                  , ifact_add_directory     , 2, NULL           } };
    gint window_width, window_height;

    pref_new_from_defs( main_prefs, sizeof( main_prefs) / sizeof( PrefDef), pref_id);

    value = pref_get_value( pref_window_width, pref_id);
    window_width  = GTK_VALUE_INT( *value);
    value = pref_get_value( pref_window_height, pref_id);
    window_height = GTK_VALUE_INT( *value);

    browser = g_new( DirectoryBrowser, 1);
    browser->images = NULL;
    browser->path = NULL;
    browser->file_browser = file_browser_new( starting_directory ? 
					      starting_directory : "/",
					      FILE_BROWSER_SHOW_ALL);

    browser->image_browser = image_browser_new( (GFunc) image_free);
    
    menubar_factory = gtk_item_factory_new( GTK_TYPE_MENU_BAR, "<MenuBar>", NULL);
    gtk_item_factory_create_items_ac( menubar_factory, 
				      sizeof( menu_entries) / sizeof( GtkItemFactoryEntry),
				      menu_entries,
				      browser, 
				      2);
    menu_factory = image_browser_get_menu_factory( browser->image_browser);
    

    window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title( GTK_WINDOW( window), "jview: file system browser");
    gtk_signal_connect( GTK_OBJECT( window), "destroy",
			(GtkSignalFunc) destroy_window_cb, browser);
    gtk_window_add_accel_group( GTK_WINDOW( window), menubar_factory->accel_group);
    gtk_window_add_accel_group( GTK_WINDOW( window), menu_factory->accel_group);

    mainbox = gtk_vbox_new( FALSE, 0);
    gtk_container_add( GTK_CONTAINER( window), mainbox);
    gtk_widget_show( mainbox);

    menubar = gtk_item_factory_get_widget( menubar_factory, "<MenuBar>");
    gtk_box_pack_start( GTK_BOX( mainbox), menubar, FALSE, TRUE, 0);
    gtk_widget_show( menubar);
    
    image_menu = gtk_item_factory_get_widget( menu_factory, "<ImageBrowser>");

#ifdef GTK_HAVE_FEATURES_1_1_2
    menuitem = gtk_tearoff_menu_item_new();
    gtk_menu_prepend( GTK_MENU( image_menu), menuitem);
    gtk_widget_show( menuitem);
#endif

    menuitem = gtk_menu_item_new_with_label( "Browser");
    gtk_menu_item_set_submenu( GTK_MENU_ITEM( menuitem), image_menu);
    gtk_menu_bar_insert( GTK_MENU_BAR( menubar), menuitem, 1);
    gtk_widget_show( menuitem);

    hpaned = gtk_hpaned_new();
    gtk_container_add( GTK_CONTAINER( mainbox), hpaned);
    gtk_widget_show( hpaned);

    frame = gtk_frame_new( NULL);
    gtk_paned_add1( GTK_PANED( hpaned), frame);
    gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);
    gtk_widget_set_usize( frame, 0.25*window_width, window_height);
    gtk_widget_show( frame);

    gtk_container_add( GTK_CONTAINER( frame), browser->file_browser);
    gtk_signal_connect( GTK_OBJECT( browser->file_browser), "select_file",
			(GtkSignalFunc) directory_browser_select_file, browser);
    gtk_widget_show( browser->file_browser);

    frame = gtk_frame_new( NULL);
    gtk_paned_add2( GTK_PANED( hpaned), frame);
    gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);
    gtk_widget_set_usize( frame, 0.75*window_width, window_height);
    gtk_widget_show( frame);

    gtk_container_add( GTK_CONTAINER( frame), browser->image_browser->mainbox);
    gtk_widget_show( browser->image_browser->mainbox);

    gtk_widget_set_usize( window, window_width, window_height);

    return window;
}

static void
directory_browser_select_file( FileBrowser *file_browser, 
			       char *directory, 
			       DirectoryBrowser *browser) {
    DIR *dir;
    struct dirent *entry;
    GList *images;
    Image *image;

    g_return_if_fail( file_browser != NULL);
    g_return_if_fail( browser != NULL);
    
#ifdef DEBUG
    printf( "Reading directory %s\n", directory);
    fflush( stdout);
#endif

    if ( browser->path)
	g_free( browser->path);
    browser->path = g_strdup( directory);

    if ( browser->images)
	g_list_free( browser->images);

    images = NULL;
    dir = opendir( directory);
    if ( dir) {
	
	while ( (entry = readdir( dir))) {
	    if ( image_extension_supported( entry->d_name)) {
		image = image_new( entry->d_name, browser->path);
		images = g_list_append( images, image);
	    }
	}
	closedir( dir);
    }
    
    browser->images = images;
    image_browser_set_images( browser->image_browser, images);
}


static void         
destroy_window_cb( GtkWidget *w, DirectoryBrowser *browser) {
    main_quit();
}

void         
main_quit() {
    GList *item;
    char *home, *config_file = NULL;

    item = gimdb_list;
    while ( item) {
	gimdb_close( (GImDB*) item->data);
	item = item->next;
    }

    /* save preferences */
    home = getenv( "HOME");

    if ( home) {
	config_file = g_malloc( strlen( home) + CONFIG_FILE_NAME_SIZE + 3);
	sprintf( config_file, "%s/.%s", home, CONFIG_FILE_NAME);
    }

    if ( !home || !pref_save( config_file, NULL)) 
	g_warning( "Can't save preferences");
    
    if ( config_file) g_free( config_file);

    /* I don't know why it fails to quit somtimes */
    gtk_main_quit();
   
    gtk_exit( 0);
}

static void
ifact_quit(  GtkWidget *w, DirectoryBrowser *browser) {
    main_quit();
}


static void         
ifact_new_database( GtkWidget *w, DirectoryBrowser *browser) {
    GtkWidget *filesel;

    filesel = gtk_file_selection_new( "Create Database :");
    gtk_signal_connect_object( GTK_OBJECT( GTK_FILE_SELECTION( filesel)->cancel_button),
			       "clicked", (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT( filesel));
    gtk_signal_connect( GTK_OBJECT( GTK_FILE_SELECTION( filesel)->ok_button),
			"clicked", (GtkSignalFunc) ifact_new_database2, browser);
    gtk_widget_show( filesel);
    
}

static void         
ifact_new_database2( GtkWidget *w, DirectoryBrowser *browser) {
    GtkWidget *filesel;
    GtkWidget *window;
    GImDB *gimdb;
    gchar *file_name;

    filesel = gtk_widget_get_toplevel( w);
    g_return_if_fail( GTK_IS_FILE_SELECTION( filesel));

    file_name = gtk_file_selection_get_filename( GTK_FILE_SELECTION( filesel));
    if ( file_name && (gimdb = gimdb_create( file_name))) {
	window = gtk_widget_get_toplevel( gimdb->tree);
	gtk_signal_connect( GTK_OBJECT( window), "destroy",
			    (GtkSignalFunc) ifact_remove_database, gimdb);
	gimdb_list = g_list_append( gimdb_list, gimdb);
    }
    gtk_widget_destroy( filesel);
}

static void         
ifact_load_database( GtkWidget *w, DirectoryBrowser *browser) {
    GtkWidget *filesel;

    filesel = gtk_file_selection_new( "Load Database :");
    gtk_signal_connect_object( GTK_OBJECT( GTK_FILE_SELECTION( filesel)->cancel_button),
			       "clicked", (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT( filesel));
    gtk_signal_connect( GTK_OBJECT( GTK_FILE_SELECTION( filesel)->ok_button),
			"clicked", (GtkSignalFunc) ifact_load_database2, browser);
    gtk_widget_show( filesel);
}

static void         
ifact_load_database2( GtkWidget *w, DirectoryBrowser *browser) {
    GtkWidget *filesel;
    GtkWidget *window;
    GImDB *gimdb;
    gchar *file_name;

    filesel = gtk_widget_get_toplevel( w);
    g_return_if_fail( GTK_IS_FILE_SELECTION( filesel));

    file_name = gtk_file_selection_get_filename( GTK_FILE_SELECTION( filesel));
    if ( file_name && (gimdb = gimdb_load( file_name))) {
	window = gtk_widget_get_toplevel( gimdb->tree);
	gtk_signal_connect( GTK_OBJECT( window), "destroy",
			    (GtkSignalFunc) ifact_remove_database, gimdb);
	gimdb_list = g_list_append( gimdb_list, gimdb);
    }
    gtk_widget_destroy( filesel);
}

static GtkWidget*
database_list_new( DirectoryBrowser *browser, GtkSignalFunc ok_func, gpointer func_data) {
    GtkWidget *window;
    GtkWidget *mainbox, *box;
    GtkWidget *list;
    GtkWidget *button;
    GtkWidget *sep;
    GList *item;
    gint i;

    window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
    gtk_container_border_width( GTK_CONTAINER( window), 7);
    gtk_window_set_title( GTK_WINDOW( window), "Select a Database");

    mainbox = gtk_vbox_new( FALSE, 7);
    gtk_container_add( GTK_CONTAINER( window), mainbox);

    list = gtk_clist_new( 1);
    gtk_clist_set_policy( GTK_CLIST( list),
			  GTK_POLICY_AUTOMATIC,
			  GTK_POLICY_AUTOMATIC);
    gtk_container_add( GTK_CONTAINER( mainbox), list);
    item = gimdb_list;
    while ( item) {
	i = gtk_clist_append( GTK_CLIST( list), & ((GImDB*) item->data)->imdb->db_path);
	gtk_clist_set_row_data( GTK_CLIST( list), i, item->data);
	item = item->next;
    }

    sep = gtk_hseparator_new();
    gtk_box_pack_start( GTK_BOX( mainbox), sep, FALSE, TRUE, 7);

    box = gtk_hbox_new( FALSE, 15);
    gtk_box_pack_end( GTK_BOX( mainbox), box, FALSE, TRUE, 7);

    button = gtk_button_new_with_label( "Cancel");
    gtk_container_add( GTK_CONTAINER( box), button);
    gtk_signal_connect_object( GTK_OBJECT( button), "clicked",
			       (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT( window));

    button = gtk_button_new_with_label( "Ok");
    gtk_container_add( GTK_CONTAINER( box), button);
    gtk_signal_connect( GTK_OBJECT( button), "clicked",
			ok_func, func_data);
    GTK_WIDGET_SET_FLAGS( button, GTK_CAN_DEFAULT);
    gtk_widget_grab_default( button);
    gtk_object_set_user_data( GTK_OBJECT( button), list);

    gtk_widget_show_all( window);

    return window;
}

static GImDB*
database_list_get_database( GtkWidget *w) {
    GtkWidget *list;
    GtkCListRow *row;
    GList *item;
    gint i = 0;

    list = gtk_object_get_user_data( GTK_OBJECT( w));
    g_return_val_if_fail( list != NULL && GTK_IS_CLIST( list), NULL);

    item = GTK_CLIST( list)->row_list;
    while ( item) {
	row = GTK_CLIST_ROW( item);
	if ( row->state & GTK_STATE_SELECTED) 
	    return gtk_clist_get_row_data( GTK_CLIST( list), i);
	i++;
	item = item->next;
    }

    return NULL;
}

static void         
ifact_add_directory( GtkWidget *w, DirectoryBrowser *browser) {
    database_list_new( browser, ifact_add_directory2, browser);
}



static void
ifact_add_directory2( GtkWidget *w, DirectoryBrowser *browser) {
    GImDB *gimdb;

    gimdb = database_list_get_database( w);
    gtk_widget_destroy( gtk_widget_get_toplevel( w));

    if ( gimdb)
	gimdb_append_images( gimdb, browser->path, browser->images);
    else
	g_warning( "No database selected");

}


static void         
ifact_add_selection( GtkWidget *w, DirectoryBrowser *browser) {
     database_list_new( browser, ifact_add_selection2, browser);
}

static void         
ifact_add_selection2( GtkWidget *w, DirectoryBrowser *browser) {
    GImDB *gimdb;
    
    gimdb = database_list_get_database( w);
    gtk_widget_destroy( gtk_widget_get_toplevel( w));

    if ( gimdb) {
	GList *list;
	
	list = ubox_get_selection( UBOX( browser->image_browser->preview_box));
	if ( list) {
	    gimdb_append_images( gimdb, browser->path, list);
	    g_list_free( list);
	} else
	    g_warning( "No image selected");
    } else
	g_warning( "No database selected");
}


static void
ifact_remove_database( GtkWidget *w, GImDB *gimdb) {
    gimdb_list = g_list_remove( gimdb_list, gimdb);
}



