/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>

#include "filebrowser.h"
#include "gtkcompat.h"

#define CTREE_SPACING 5

#define TEST_COLORS

static void        file_browser_class_init( FileBrowserClass *kclass);
static void        file_browser_init( FileBrowser *fb);


static void          file_browser_tree_select_row( GtkCTree *tree, GtkCTreeNode *node, gint column);
static void          file_browser_tree_expand( GtkCTree *tree, GtkCTreeNode *node);
static void          tree_get_path( FileBrowser *fb, GtkCTreeNode *node);
static void          tree_expand_node( FileBrowser *fb, GtkCTreeNode *node);
static GtkCTreeNode* tree_find_file( FileBrowser *fb, GtkCTreeNode *parent, char *file_name);

#if 0
static    GdkColor lnk_color = { 0, 0x0000, 0xffff, 0xffff};      /* symbolic link */
static    GdkColor reg_color = { 0, 0x0000, 0x0000, 0x0000};      /* regular file  */
static    GdkColor dir_color = { 0, 0x0000, 0xffff, 0x0000};      /* directory    */
static    GdkColor chr_color = { 0, 0xffff, 0xffff, 0x0000};      /* character device */
static    GdkColor blk_color = { 0, 0xffff, 0xffff, 0x0000};      /* block device */
static    GdkColor fifo_color= { 0, 0xffff, 0x0000, 0x0000};      /* fifo */
static    GdkColor sock_color= { 0, 0xffff, 0x0000, 0xffff};      /* socket */
#endif

#ifdef TEST_COLORS
static    GdkColor lnk_color = { 0, 0x0000, 0xcd00, 0xcd00};      /* symbolic link */
static    GdkColor reg_color = { 0, 0x0000, 0x0000, 0x0000};      /* regular file  */
static    GdkColor dir_color = { 0, 0x2e00, 0x8b00, 0x5700};      /* directory    */
static    GdkColor chr_color = { 0, 0xff00, 0xd700, 0x0000};      /* character device */
static    GdkColor blk_color = { 0, 0xff00, 0xd700, 0x0000};      /* block device */
static    GdkColor fifo_color= { 0, 0xff00, 0x6300, 0x4700};      /* fifo */
static    GdkColor sock_color= { 0, 0xff00, 0x1400, 0x9300};      /* socket */
#endif

enum {
    SELECT_FILE,
    LAST_SIGNAL
};

static GtkCTreeClass *parent_class;

static guint file_browser_signals[LAST_SIGNAL] = {0};

GtkType
file_browser_get_type( void) {
    static GtkType file_browser_type = 0;

    if ( !file_browser_type) {
	GtkTypeInfo file_browser_info = {
	    "FileBrowser",
	    sizeof( FileBrowser),
	    sizeof( FileBrowserClass),
	    (GtkClassInitFunc) file_browser_class_init,
	    (GtkObjectInitFunc) file_browser_init,
	    /* reserved_1 */ NULL,
	    /* reserved_2 */ NULL,
	    (GtkClassInitFunc) NULL,
	};
	
	file_browser_type = gtk_type_unique( GTK_TYPE_CTREE, &file_browser_info);
    }

    return file_browser_type;
}


static void        
file_browser_class_init( FileBrowserClass *klass) {
    GtkObjectClass *object_class;
    GtkCTreeClass  *ctree_class;

    object_class = (GtkObjectClass *) klass;
    ctree_class  = (GtkCTreeClass*) klass;

    parent_class = gtk_type_class( gtk_ctree_get_type());

    file_browser_signals[ SELECT_FILE] = 
	gtk_signal_new( "select_file",
			GTK_RUN_FIRST,
			object_class->type,
			GTK_SIGNAL_OFFSET ( FileBrowserClass, select_file),
			gtk_marshal_NONE__POINTER,
			GTK_TYPE_NONE, 1, GTK_TYPE_POINTER);

    gtk_object_class_add_signals (object_class, file_browser_signals, LAST_SIGNAL);

    ctree_class->tree_expand = file_browser_tree_expand;
    ctree_class->tree_select_row = file_browser_tree_select_row;

    klass->select_file = NULL;

}


static void        
file_browser_init( FileBrowser *fb) {
    *fb->last_path = 0;

    fb->flags = FILE_BROWSER_SHOW_CTIME | 
	FILE_BROWSER_SHOW_ALL;
}

GtkWidget*
file_browser_new( char *root_path, 
		      guint flags) {
    GtkWidget *widget;
    FileBrowser *fb;
    GtkCTreeNode *node = NULL;
    GdkColormap *cmap;
    char *text[6] = {0, "", "", "", "", ""};
    char *root = "/", *rp, *tmp;

    g_return_val_if_fail( root_path != NULL, NULL);

    widget = gtk_type_new( file_browser_get_type());
    gtk_ctree_construct( GTK_CTREE( widget),
			 flags & FILE_BROWSER_LONG_LIST ? 6 : 1,
			 FALSE, NULL);
    
    cmap = gtk_widget_get_colormap( widget);
#ifdef TEST_COLORS
    g_return_val_if_fail( cmap != NULL, NULL);
    if ( cmap) {
	gdk_color_alloc( cmap, &dir_color);
	gdk_color_alloc( cmap, &lnk_color);
	if ( flags & FILE_BROWSER_SHOW_FILES) {
	    gdk_color_alloc( cmap, &reg_color);	
	    gdk_color_alloc( cmap, &chr_color);
	    gdk_color_alloc( cmap, &blk_color);
	    gdk_color_alloc( cmap, &fifo_color);
	    gdk_color_alloc( cmap, &sock_color);

	}
    }
#endif

    fb = FILE_BROWSER( widget);
    fb->flags = flags;
    gtk_clist_set_policy( GTK_CLIST( fb),
			  GTK_POLICY_AUTOMATIC,
			  GTK_POLICY_AUTOMATIC);
#ifndef GTK_HAVE_FEATURES_1_1_2
    gtk_ctree_set_auto_sort( GTK_CTREE( fb), TRUE);
#endif

    text[0] = root;
    node = gtk_ctree_insert( GTK_CTREE( fb),
			     NULL, NULL,
			     text, CTREE_SPACING,
			     NULL, NULL, NULL, NULL,
			     FALSE, TRUE);
    tree_expand_node( fb, node);
    
    rp = g_strdup( root_path);
    tmp = strtok( rp, "/");
    while ( tmp) {
	node = tree_find_file( fb, node, tmp);
	if ( !node)
	    break;
	gtk_ctree_expand( GTK_CTREE( fb), node);
	tmp = strtok( NULL, "/");
    }
    
    g_free( rp);
     

    return widget;
}

static void
file_browser_tree_select_row( GtkCTree *tree, GtkCTreeNode *node, gint column) {
    FileBrowser *fb;

    g_return_if_fail( tree != NULL);
    g_return_if_fail( IS_FILE_BROWSER( tree));
    g_return_if_fail( node != NULL);

    if ( parent_class->tree_select_row)
	( *parent_class->tree_select_row) ( tree, node, column);

    fb = FILE_BROWSER( tree);

    tree_get_path( fb, node);

    gtk_signal_emit( GTK_OBJECT( fb), 
		     file_browser_signals[ SELECT_FILE],
		     fb->last_path);
    

}

static void
file_browser_tree_expand( GtkCTree *tree, GtkCTreeNode *node) {

    g_return_if_fail( tree != NULL);
    g_return_if_fail( IS_FILE_BROWSER( tree));
    g_return_if_fail( node != NULL);
    
    tree_expand_node( FILE_BROWSER( tree), node);

    if ( parent_class->tree_expand)
	( *parent_class->tree_expand) ( tree, node);
}

static void
tree_expand_node( FileBrowser *fb, GtkCTreeNode *node) {
    GtkCTreeNode *child;
    char buffer[4096];
    DIR *dir;
    struct stat stat_buff;
    struct dirent *entry;
    gint  has_subdir;
    
    char file_size[12];
    char file_time[32];
    char mode[10];
    char *text[6] = {0};
    
    char *unknown = "unknown";

    g_return_if_fail( fb != NULL);
    g_return_if_fail( node != NULL);

    if ( GTK_WIDGET_VISIBLE( fb))
	gtk_clist_freeze( GTK_CLIST( fb));

    /* remove childs */
    child = GTK_CTREE_ROW( node)->children;
    while ( child) {
	gtk_ctree_remove( GTK_CTREE( fb), child);
	child = GTK_CTREE_ROW( node)->children;
    } 
	
        /* get the path of the node to expand */
    tree_get_path( fb, node);
    g_return_if_fail( *fb->last_path != 0);

    if ( stat( fb->last_path, &stat_buff))
	return;

    dir = opendir( fb->last_path);
    g_return_if_fail( dir != NULL);
    
   
    child = NULL;
    text[1] = file_size;
    text[4] = file_time;
    text[5] = mode;
    *mode = 0;
    while ( (entry = readdir( dir))) {
	
	if ( *entry->d_name == '.' &&
	     !(fb->flags & FILE_BROWSER_SHOW_ALL))
	    continue; /* don't show files starting with '.' */

	/* skip "." and ".." */
	if ( !strcmp( entry->d_name, ".") ||
	     !strcmp( entry->d_name, ".."))
	    continue; 

	snprintf( buffer, 4096, "%s/%s", fb->last_path, entry->d_name);

	if ( !stat( buffer, &stat_buff) &&
	     ( (fb->flags & FILE_BROWSER_SHOW_FILES) || 
	       S_ISDIR( stat_buff.st_mode))) {
	    if ( fb->flags & FILE_BROWSER_LONG_LIST) {
		struct passwd *user_info;
		struct group  *group_info;
		
		user_info = getpwuid( stat_buff.st_uid);
		group_info = getgrgid( stat_buff.st_gid);
		sprintf( file_time, "%s", ctime( &stat_buff.st_mtime));
		sprintf( file_size, "%ld", stat_buff.st_size);

		text[0] = entry->d_name;
		text[2] = user_info ? user_info->pw_name : unknown;
		text[3] = group_info ? group_info->gr_name : unknown;

		has_subdir = S_ISDIR( stat_buff.st_mode) && (stat_buff.st_nlink > 2);
		child = gtk_ctree_insert( GTK_CTREE( fb),
					  node, child, 
					  text,
					  CTREE_SPACING,
					  NULL, NULL, NULL, NULL,
					  has_subdir == FALSE, FALSE);
		
		
		if ( has_subdir) {
		    char *tmp[6] = { unknown};

		    gtk_ctree_insert( GTK_CTREE( fb),
				      child, NULL, 
				      tmp,
				      CTREE_SPACING,
				      NULL, NULL, NULL, NULL,
				      TRUE, FALSE);
		}

	    } else {
		text[0] = entry->d_name;

		has_subdir = S_ISDIR( stat_buff.st_mode) && (stat_buff.st_nlink > 2);
		child = gtk_ctree_insert( GTK_CTREE( fb),
					  node, child, 
					  text,
					  CTREE_SPACING,
					  NULL, NULL, NULL, NULL,
					  has_subdir == FALSE, FALSE);
		
		
		if ( has_subdir) 
		    gtk_ctree_insert( GTK_CTREE( fb),
				      child, NULL, 
				      &unknown,
				      CTREE_SPACING,
				      NULL, NULL, NULL, NULL,
				      TRUE, FALSE);
	    }
#ifdef TEST_COLORS
	    if ( !lstat( buffer, &stat_buff)) {
		if ( S_ISLNK( stat_buff.st_mode))
		    gtk_ctree_node_set_foreground( GTK_CTREE( fb), child, &lnk_color);
		else if ( S_ISREG( stat_buff.st_mode))
		    gtk_ctree_node_set_foreground( GTK_CTREE( fb), child, &reg_color);
		else if ( S_ISDIR( stat_buff.st_mode))
		    gtk_ctree_node_set_foreground( GTK_CTREE( fb), child, &dir_color);
		else if ( S_ISCHR( stat_buff.st_mode))
		    gtk_ctree_node_set_foreground( GTK_CTREE( fb), child, &chr_color);
		else if ( S_ISBLK( stat_buff.st_mode))
		    gtk_ctree_node_set_foreground( GTK_CTREE( fb), child, &blk_color);
		else if ( S_ISFIFO( stat_buff.st_mode))
		    gtk_ctree_node_set_foreground( GTK_CTREE( fb), child, &fifo_color);
		else if ( S_ISSOCK( stat_buff.st_mode))
		    gtk_ctree_node_set_foreground( GTK_CTREE( fb), child, &sock_color);
	    }
#endif
	}

    }
    
    closedir( dir);

#ifdef GTK_HAVE_FEATURES_1_1_2
    gtk_ctree_sort_node( GTK_CTREE( fb), node);
#endif

    if ( GTK_WIDGET_VISIBLE( fb))
	gtk_clist_thaw( GTK_CLIST( fb));

}
    

static GtkCTreeNode* 
tree_find_file( FileBrowser *fb, GtkCTreeNode *parent, char *file_name) {
    GtkCTreeNode *child;

    g_return_val_if_fail( fb != NULL, NULL);
    g_return_val_if_fail( parent != NULL && file_name != NULL, NULL);
    
    child = GTK_CTREE_ROW( parent)->children;
    while ( child) {
	if ( !strcmp( GTK_CELL_PIXTEXT( GTK_CTREE_ROW( child)->row.cell[0])->text, file_name))
	    return child;
	child = GTK_CTREE_ROW( child)->sibling;
    }

    return NULL;
}


static void
tree_get_path( FileBrowser *fb, GtkCTreeNode *node) {
    GList *dirs = NULL, *d;
    char *dir, *buffer = fb->last_path;

    g_return_if_fail( fb != NULL && node != NULL);

    
    *buffer = 0;

    while ( node) {

	dir = GTK_CELL_PIXTEXT (GTK_CTREE_ROW 
				(node)->row.cell[GTK_CTREE( fb)->tree_column])->text;
	g_return_if_fail( dir != NULL);

	dirs = g_list_prepend( dirs, dir);

	node = GTK_CTREE_ROW( node)->parent;
    }

    if ( dirs) {
	gint i, size;
	size = 4096;

	if ( strcmp( dir, "/") || !dirs->next) {
	    snprintf( buffer, size, "%s", dir);
	    i = strlen( dir);
	    buffer += i;
	    size -= i;
	}
	
	d = dirs->next;
	while ( d) {
	    dir = d->data;
	    d = d->next;
	    
	    snprintf( buffer, size, "/%s", dir);
	    i = strlen( dir) + 1;
	    buffer += i;
	    size -= i;
	}
	
	g_list_free( dirs);
    } 
}



