/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "gimdb.h"
#include "imdb.h"
#include "find.h"

#define BOX_SPACING 8
#define BORDER_WIDTH 6

typedef enum {
    FIND_NAME,
    FIND_WIDTH,
    FIND_HEIGHT,
    FIND_DEPTH,
    FIND_FILE_SIZE,
    FIND_PROPERTY,
    FIND_KEYWORD
} FindType;

typedef enum {
    EQUAL,
    SMALLER,
    BIGGER,
    BEGIN_WITH,
    END_WITH,
    REGEXP
} FindCompare;

typedef enum {
    SRC_GIMDBS,
    SRC_IMAGE_BROWSER
} FindSource;

typedef struct _FindParam {
    GtkWidget *hbox;
    FindCompare  compare;

    FindType type;
    union {
	struct _s {
	    char *string;
	    GtkWidget *entry;
	} s;

	struct _t {
	    guint32 number;
	    GtkWidget *entry;
	} t;

	struct _p {
	    GImDB *gimdb;

	    gint property_id;
	    gint value_id;
	    
	    GtkWidget *properties;
	    GtkWidget *values;
	} p;

    } u;
} FindParam;

	    
typedef struct _Find {
    FindSource type;
    GList *params;
    GList *result;
    guint idler_tag;

    
    GtkWidget *params_box;
    GtkWidget *find_button;
    ImageBrowser *browser;

    gfloat increment;
    gfloat percentage;

    union {
	struct _g {
	    GList *gimdbs;
	    GtkWidget *dbs;
	    
	    /* used by idle_finder */
	    GList *gimdb;
	    guint directory;
	    GList *image;
	} g;

	struct _b {
	    ImageBrowser *browser;

	    /* used by idle_finder */
	    GList *image;
	} b;

    } u;
} Find;

typedef struct _Int2Str {
    gint type;
    char *str;
} Int2Str;

static GtkWidget*        find_build( Find *find);
static void              find_destroy_cb( GtkWidget *w, Find *find);
static void              more_params( GtkWidget *w, Find *find);
static void              fewer_params( GtkWidget *w, Find *find);
static GtkWidget*        types_menu_new();
static GtkWidget*        compare_menu_new();
static GtkWidget*        database_selector_new( GList *gimdbs);
static GtkWidget*        property_menu_new( Find *find);
static void              property_values_menu_build( Find *find, FindParam *param, ImDBProperty *property);
static void              select_all_dbs( GtkWidget *w, Find *find);
static void              select_type( GtkWidget *w, Find *find);
static void              select_compare( GtkWidget *w, Find *find);
static void              select_property( GtkWidget *w, Find *find);
static void              select_property_value( GtkWidget *w, Find *find);
static void              exec_find( GtkWidget *w, Find *find);
static gint              find_gimdb_idler( Find *find);
static gint              find_browser_idler( Find *find);
static void              find_test_image( Find *find, Image *image);
static gint              strings_match( char *s1, char *s2, FindCompare cmp);
static gint              strings_match_case( char *s1, char *s2, FindCompare cmp);
static gint              guint32_match( guint32 i1, guint32 i2, FindCompare cmp);
static gint              _strncasecmp( char *s1, char *s2, gint n);

static Int2Str FindTypes[]  = {
    { FIND_NAME     , "name"               },
    { FIND_WIDTH    , "image width"        },
    { FIND_HEIGHT   , "image height"       },
    { FIND_DEPTH    , "image depth"        },
    { FIND_FILE_SIZE, "file size"          },
    { FIND_PROPERTY , "property"           },
    { FIND_KEYWORD  , "keyword"            },
    { 0             , NULL                 }
};

static Int2Str FindCompares[] = {
    { EQUAL        , "equal"       },
    { SMALLER      , "smaller"     },
    { BIGGER      , "bigger"     },
    { BEGIN_WITH   , "begin with"  },
    { END_WITH     , "end with"    },
    { REGEXP       , "match regexp"},
    { 0            , NULL          }
};

    
/* xxx All functions calling *_find_new should disable
   image insertion & deletion as we are working with their
   image list ( or we should first copy all images subject
   to the search )
   This is not done at the moment because after setting
   a widget state to GTK_STATE_INSENSITIVE, I can't restore
   the state of the widget ???
   */
GtkWidget*
gimdb_find_new( GList *gimdbs) {
    Find *find;
    GtkWidget *window, *hbox, *button;

    find = g_new( Find, 1);
    find->type = SRC_GIMDBS;
    find->params = NULL;
    find->result = NULL;
    find->idler_tag = 0;
    find->u.g.gimdbs = gimdbs;

    window = find_build( find);

    hbox = gtk_hbox_new( FALSE, BOX_SPACING);
    gtk_box_pack_start( GTK_BOX( find->params_box), hbox, FALSE, TRUE, 0);
    gtk_box_reorder_child( GTK_BOX( find->params_box), hbox, 0);
    gtk_widget_show( hbox);

    find->u.g.dbs = database_selector_new( gimdbs);
    gtk_box_pack_start( GTK_BOX( hbox), find->u.g.dbs, FALSE, TRUE, 0);
    gtk_widget_set_usize( find->u.g.dbs, 220 + BOX_SPACING, -1);
    gtk_widget_show( find->u.g.dbs);

    button = gtk_check_button_new_with_label( "All");
    gtk_box_pack_start( GTK_BOX( hbox), button, FALSE, TRUE, 0);
    gtk_signal_connect( GTK_OBJECT( button), "clicked",
			(GtkSignalFunc) select_all_dbs, find);
    gtk_widget_show( button);

    return window;
}

GtkWidget*
browser_find_new( ImageBrowser *browser) {
    Find *find;
    GtkWidget *window;

    find = g_new( Find, 1);
    find->type = SRC_IMAGE_BROWSER;
    find->params = NULL;
    find->result = NULL;
    find->idler_tag = 0;
    find->u.b.browser = browser;

    window = find_build( find);
 
    return window;
}

static GtkWidget*
find_build( Find *find) {
    GtkWidget *window, *mainbox, *hbox, *button, *frame, *result_frame;
    GtkItemFactory *ifact;
    GtkWidget *menubar, *menu, *menuitem;


    find->browser = image_browser_new( (GFunc) image_free); 

    window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title( GTK_WINDOW( window), "jview: search databases");
    gtk_container_border_width( GTK_CONTAINER( window), BORDER_WIDTH);
    gtk_signal_connect( GTK_OBJECT( window), "destroy",
			(GtkSignalFunc) find_destroy_cb, find);

 
    mainbox = gtk_vbox_new( FALSE, BOX_SPACING);
    gtk_container_add( GTK_CONTAINER( window), mainbox);

    frame = gtk_frame_new( "Parameters");
    gtk_box_pack_start( GTK_BOX( mainbox), frame, FALSE, TRUE, 0);
    //gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);

    find->params_box = gtk_vbox_new( FALSE, BOX_SPACING);
    gtk_container_border_width( GTK_CONTAINER( find->params_box), BORDER_WIDTH);
    gtk_container_add( GTK_CONTAINER( frame), find->params_box);

    more_params( NULL, find);
    
    result_frame = gtk_frame_new( "Result");
    gtk_box_pack_end( GTK_BOX( mainbox), result_frame, TRUE, TRUE, 0);
    gtk_frame_set_shadow_type( GTK_FRAME( result_frame), GTK_SHADOW_IN);

    hbox = gtk_hbox_new( FALSE, BOX_SPACING);
    gtk_box_pack_end( GTK_BOX( mainbox), hbox, FALSE, TRUE, 0);
    gtk_container_border_width( GTK_CONTAINER( hbox), BORDER_WIDTH);

    find->find_button = gtk_button_new_with_label( "Find");
    gtk_box_pack_start( GTK_BOX( hbox), find->find_button, FALSE, TRUE, 0);
    gtk_signal_connect( GTK_OBJECT( find->find_button), "clicked",
			(GtkSignalFunc) exec_find, find);

    button = gtk_button_new_with_label( "More Parameters");
    gtk_box_pack_start( GTK_BOX( hbox), button, FALSE, TRUE, 0);
    gtk_signal_connect( GTK_OBJECT( button), "clicked",
			(GtkSignalFunc) more_params, find);

    button = gtk_button_new_with_label( "Fewer Parameters");
    gtk_box_pack_start( GTK_BOX( hbox), button, FALSE, TRUE, 0);
    gtk_signal_connect( GTK_OBJECT( button), "clicked",
			(GtkSignalFunc) fewer_params, find);

    menubar = gtk_menu_bar_new();
    gtk_box_pack_end( GTK_BOX( hbox), menubar, FALSE, TRUE, 0);
    
    menuitem = gtk_menu_item_new_with_label( "Browser");
    gtk_menu_bar_append( GTK_MENU_BAR( menubar), menuitem);
   
    ifact = image_browser_get_menu_factory( find->browser);
    menu = gtk_item_factory_get_widget( ifact, "<ImageBrowser>");
    gtk_menu_item_set_submenu( GTK_MENU_ITEM( menuitem), menu);

    gtk_window_add_accel_group( GTK_WINDOW( window), ifact->accel_group);
    gtk_widget_show_all( window);

    gtk_container_add( GTK_CONTAINER( result_frame), find->browser->mainbox);
    gtk_widget_show( find->browser->mainbox);

    return window;
}

static void
find_destroy_cb( GtkWidget *w, Find *find) {
    GList *list;
    FindParam *param;
    
    if ( find->idler_tag)
	gtk_idle_remove( find->idler_tag);

    list = find->params;
    while ( list) {
	param = list->data;
	list = list->next;
	
	if ( (param->type == FIND_NAME || param->type == FIND_KEYWORD) &&
	     param->u.s.string) 
	    g_free( param->u.s.string);
	
	g_free( param);
    }
    if ( find->params)
	g_list_free( find->params);

    g_free( find);
}
    
static void
more_params( GtkWidget *w, Find *find) {
    FindParam *param;
    GtkWidget *type;
    GtkWidget *compare;
    GtkWidget *menu;

    param = g_new( FindParam, 1);
    bzero( param, sizeof( FindParam));
    param->type = FIND_NAME;
    find->params = g_list_append( find->params, param);

    param->hbox = gtk_hbox_new( FALSE, BOX_SPACING);
    gtk_box_pack_start( GTK_BOX( find->params_box), param->hbox, FALSE, TRUE, 0);
    gtk_widget_show( param->hbox);

    type = types_menu_new();
    gtk_box_pack_start( GTK_BOX( param->hbox), type, FALSE, TRUE, 0);
    menu = gtk_option_menu_get_menu( GTK_OPTION_MENU( type));
    gtk_signal_connect( GTK_OBJECT( menu), "selection_done", 
			(GtkSignalFunc) select_type, find);
    gtk_object_set_user_data( GTK_OBJECT( menu), param);
    gtk_widget_set_usize( type, 110, -1);
    gtk_widget_show( type);
    
    compare = compare_menu_new();
    gtk_box_pack_start( GTK_BOX( param->hbox), compare, FALSE, TRUE, 0);
    menu = gtk_option_menu_get_menu( GTK_OPTION_MENU( compare));
    gtk_signal_connect( GTK_OBJECT( menu), "selection_done", 
			(GtkSignalFunc) select_compare, find);
    gtk_object_set_user_data( GTK_OBJECT( menu), param);
    gtk_widget_set_usize( compare, 110, -1);
    gtk_widget_show( compare);

    param->u.s.entry = gtk_entry_new();
    gtk_box_pack_end( GTK_BOX( param->hbox), param->u.s.entry, TRUE, TRUE, 0);
    gtk_widget_show( param->u.s.entry);
}
    
static void              
fewer_params( GtkWidget *w, Find *find) {
    GList *list;
    FindParam *param;
    GtkWidget *parent;

    list = g_list_last( find->params);
    if ( list && list->prev) {
	param = list->data;
	
	parent = param->hbox->parent->parent;
	gtk_widget_destroy( param->hbox);
	gtk_widget_queue_resize( parent);
	find->params = g_list_remove_link( find->params, list);
    }
}

static GtkWidget*
types_menu_new() {
    GtkWidget *omenu, *menu;
    GtkWidget *menuitem;
    GSList    *group = NULL;
    Int2Str   *ass = FindTypes;

    omenu = gtk_option_menu_new();
    menu = gtk_menu_new();
    gtk_option_menu_set_menu( GTK_OPTION_MENU( omenu), menu);
    while ( ass->str) {
	menuitem = gtk_radio_menu_item_new_with_label( group, ass->str);
	gtk_menu_append( GTK_MENU( menu), menuitem);
	gtk_object_set_user_data( GTK_OBJECT( menuitem), ass); 
	gtk_widget_show( menuitem);

	group = gtk_radio_menu_item_group( GTK_RADIO_MENU_ITEM( menuitem));
	ass ++;
    }
    gtk_option_menu_set_history( GTK_OPTION_MENU( omenu), 0);

    return omenu;
}

static GtkWidget*
compare_menu_new() {
    GtkWidget *omenu, *menu;
    GtkWidget *menuitem;
    GSList    *group = NULL;
    Int2Str   *ass = FindCompares;

    omenu = gtk_option_menu_new();
    menu = gtk_menu_new();
    gtk_option_menu_set_menu( GTK_OPTION_MENU( omenu), menu);
    while ( ass->str) {
	menuitem = gtk_radio_menu_item_new_with_label( group, ass->str);
	gtk_menu_append( GTK_MENU( menu), menuitem);
	gtk_object_set_user_data( GTK_OBJECT( menuitem), ass);
	gtk_widget_show( menuitem);

	group = gtk_radio_menu_item_group( GTK_RADIO_MENU_ITEM( menuitem));
	ass ++;
    }
    gtk_option_menu_set_history( GTK_OPTION_MENU( omenu), 0);

    return omenu;
}

static GtkWidget*
database_selector_new( GList *gimdbs) {
    GtkWidget *omenu, *menu;
    GtkWidget *menuitem;
    GImDB     *gimdb;
    char      *name;
    gint first = TRUE;

    omenu = gtk_option_menu_new();
    menu = gtk_menu_new();
    gtk_option_menu_set_menu( GTK_OPTION_MENU( omenu), menu);
    while ( gimdbs) {
	gimdb = gimdbs->data;
	gimdbs = gimdbs->next;

	name = strrchr( gimdb->imdb->db_path, '/');
	g_return_val_if_fail( name != NULL, NULL);

	name ++;
	menuitem = gtk_check_menu_item_new_with_label( name);
	if ( first) {
	    gtk_check_menu_item_set_state( GTK_CHECK_MENU_ITEM( menuitem), 
					   GTK_STATE_ACTIVE);
	    first = FALSE;
	}
	gtk_menu_append( GTK_MENU( menu), menuitem);
	gtk_object_set_user_data( GTK_OBJECT( menuitem), gimdb);
	gtk_widget_show( menuitem);

    }
    gtk_option_menu_set_history( GTK_OPTION_MENU( omenu), 0);

    return omenu;
}

static GtkWidget*
property_menu_new( Find *find) {
    GtkWidget *omenu, *dbmenu, *menu;
    GtkWidget *menuitem;
    ImDBProperty *property;
    GImDB *gimdb;
    GList *list, *gimdbs;
    GSList *group = NULL;

    gimdbs = find->u.g.gimdbs;
    omenu = gtk_option_menu_new();
    dbmenu = gtk_menu_new();
    gtk_option_menu_set_menu( GTK_OPTION_MENU( omenu), dbmenu);

    while ( gimdbs) {
	gimdb = gimdbs->data;
	gimdbs = gimdbs->next;

	menuitem = gtk_menu_item_new_with_label( strrchr( gimdb->imdb->db_path, '/') + 1);
	gtk_menu_append( GTK_MENU( dbmenu), menuitem);
	gtk_widget_show( menuitem);
	    
	menu = gtk_menu_new();
	gtk_menu_item_set_submenu( GTK_MENU_ITEM( menuitem), menu);
	gtk_object_set_user_data( GTK_OBJECT( menu), gimdb);

	list = gimdb->imdb->properties;
	while ( list) {
	    property = list->data;
	    list = list->next;
	    
	    menuitem = gtk_radio_menu_item_new_with_label( group, property->name);
	    gtk_menu_append( GTK_MENU( menu), menuitem);
	    gtk_object_set_user_data( GTK_OBJECT( menuitem), property);
	    gtk_widget_show( menuitem);
	    
	    group = gtk_radio_menu_item_group( GTK_RADIO_MENU_ITEM( menuitem));
	}
    }

    return omenu;
}

static void
property_values_menu_build( Find *find, FindParam *param, ImDBProperty *property) {
    ImDBPropertyValue *value;
    GtkWidget *menu, *menuitem;
    GList *list;
    GSList *group = NULL;
    
    gtk_option_menu_remove_menu( GTK_OPTION_MENU( param->u.p.values));

    menu = gtk_menu_new();
    gtk_option_menu_set_menu( GTK_OPTION_MENU( param->u.p.values), menu);
    gtk_signal_connect( GTK_OBJECT( menu), "selection_done",
			(GtkSignalFunc) select_property_value, find);
    gtk_object_set_user_data( GTK_OBJECT( menu), param);


    list = property->values;
    while ( list) {
	value = list->data;
	list = list->next;

	menuitem = gtk_radio_menu_item_new_with_label( group, value->name);
	gtk_menu_append( GTK_MENU( menu), menuitem);
	gtk_object_set_user_data( GTK_OBJECT( menuitem), value);
	gtk_widget_show( menuitem);

	group = gtk_radio_menu_item_group( GTK_RADIO_MENU_ITEM( menuitem));
    }

}

static void              
select_all_dbs( GtkWidget *w, Find *find) {
    if ( find->u.g.dbs->state & GTK_STATE_INSENSITIVE)
	gtk_widget_set_state( find->u.g.dbs, GTK_STATE_NORMAL);
    else
	gtk_widget_set_state( find->u.g.dbs, GTK_STATE_INSENSITIVE);

}

static void
select_type( GtkWidget *w, Find *find) {
    FindParam *param;
    GtkWidget *active, *menu;
    GList *children, *child;
    Int2Str *ass;

    g_return_if_fail( w != NULL);
    param = gtk_object_get_user_data( GTK_OBJECT( w));
    g_return_if_fail( param != NULL);
    active = gtk_menu_get_active( GTK_MENU( w));
    g_return_if_fail( active != NULL);
    ass = gtk_object_get_user_data( GTK_OBJECT( active));
    g_return_if_fail( ass != NULL);

    if ( param->type == ass->type) 
	return;

    switch ( param->type) {
    case FIND_NAME:
    case FIND_KEYWORD:
	if ( param->u.s.string) {
	    g_free( param->u.s.string);
	    param->u.s.string = NULL;
	}
	gtk_container_remove( GTK_CONTAINER( param->hbox), param->u.s.entry);
	break;
    case FIND_WIDTH:
    case FIND_HEIGHT:
    case FIND_DEPTH:
    case FIND_FILE_SIZE:
	gtk_container_remove( GTK_CONTAINER( param->hbox), param->u.t.entry);
	break;
    case FIND_PROPERTY:
	gtk_container_remove( GTK_CONTAINER( param->hbox), param->u.p.properties);
	gtk_container_remove( GTK_CONTAINER( param->hbox), param->u.p.values);
	break;
    default:
	break;
    }

    param->type = ass->type;
    switch ( param->type) {
    case FIND_NAME:
    case FIND_KEYWORD:
	param->u.s.entry = gtk_entry_new();
	gtk_box_pack_end( GTK_BOX( param->hbox), param->u.s.entry, TRUE, TRUE, 0);
	gtk_widget_show( param->u.s.entry);
	break;
    case FIND_WIDTH:
    case FIND_HEIGHT:
    case FIND_DEPTH:
    case FIND_FILE_SIZE:
	param->u.t.entry = gtk_entry_new();
	gtk_box_pack_end( GTK_BOX( param->hbox), param->u.t.entry, TRUE, TRUE, 0);
	gtk_widget_show( param->u.t.entry);
	break;
    case FIND_PROPERTY:
	param->u.p.properties = property_menu_new( find);
	gtk_box_pack_start( GTK_BOX( param->hbox), param->u.p.properties, FALSE, TRUE, 0);
	gtk_widget_set_usize( param->u.p.properties, 110, -1);
	gtk_widget_show( param->u.p.properties);

	menu = gtk_option_menu_get_menu( GTK_OPTION_MENU( param->u.p.properties));
	child = children = gtk_container_children( GTK_CONTAINER( menu));
	while ( child) {
	    if ( GTK_IS_MENU_ITEM( child->data)) {
		GtkWidget *menuitem = child->data;

		menu = GTK_MENU_ITEM( menuitem)->submenu;
		gtk_signal_connect( GTK_OBJECT( menu), "selection_done",
				    (GtkSignalFunc) select_property, find);
		gtk_object_set_user_data( GTK_OBJECT( menu), param);
	    }
	    child = child->next;
	}
	if ( children)
	    g_list_free( children);

	param->u.p.values = gtk_option_menu_new();
	gtk_box_pack_start( GTK_BOX( param->hbox), param->u.p.values, FALSE, TRUE, 0);
	gtk_widget_set_usize( param->u.p.values, 110, -1);
	gtk_widget_show( param->u.p.values);

	menu = gtk_menu_new();
	gtk_option_menu_set_menu( GTK_OPTION_MENU( param->u.p.values), menu);

	break;
    default:
	break;
    };

}

static void
select_compare( GtkWidget *w, Find *find) {
    FindParam *param;
    GtkWidget *active;
    Int2Str   *ass;

    g_return_if_fail( w != NULL);
    param = gtk_object_get_user_data( GTK_OBJECT( w));
    g_return_if_fail( param != NULL);
    active = gtk_menu_get_active( GTK_MENU( w));
    g_return_if_fail( active != NULL);
    ass = gtk_object_get_user_data( GTK_OBJECT( active));
    g_return_if_fail( ass != NULL);

    param->compare = ass->type;
}

static void
select_property( GtkWidget *w, Find *find) {
    FindParam *param;
    ImDBProperty *property;
    GtkWidget *active;
  
    g_return_if_fail( w != NULL);
    param = gtk_object_get_user_data( GTK_OBJECT( w));
    g_return_if_fail( param != NULL);
    active = gtk_menu_get_active( GTK_MENU( w));
    g_return_if_fail( active != NULL);
    property = gtk_object_get_user_data( GTK_OBJECT( active));
    if ( !property)
	return;
    
    param->u.p.gimdb = gtk_object_get_user_data( GTK_OBJECT( active->parent));
    g_return_if_fail( param->u.p.gimdb != NULL);

    param->u.p.property_id = property->property_id;
    property_values_menu_build( find, param, property);
}

static void
select_property_value( GtkWidget *w, Find *find) {
    FindParam *param;
    ImDBPropertyValue *value;
    GtkWidget *active;

    g_return_if_fail( w != NULL);
    param = gtk_object_get_user_data( GTK_OBJECT( w));
    g_return_if_fail( param != NULL);
    active = gtk_menu_get_active( GTK_MENU( w));
    if ( !active)
	return;
    value = gtk_object_get_user_data( GTK_OBJECT( active));
    g_return_if_fail( value != NULL);

    param->u.p.value_id = value->id;
}

static void
exec_find( GtkWidget *w, Find *find) {
    GList *list;
    FindParam *param;
    guint32 n;
    //gint image_count;
    char *s;

    if ( find->idler_tag) {
	gtk_idle_remove( find->idler_tag);
	find->idler_tag = 0;
    }

    image_browser_set_images( find->browser, NULL);

    if ( find->result) {
	g_list_free( find->result);
	find->result = NULL;
    }

    list = find->params;
    while ( list) {
	param = list->data;

	switch( param->type) {
	case FIND_NAME:
	case FIND_KEYWORD:
	    s = gtk_entry_get_text( GTK_ENTRY( param->u.s.entry));
	    if ( !s) { 
		GList *tmp = list;
		list = list->next;
		find->params = g_list_remove_link( find->params, tmp);
		continue;
		break;
	    }
	    if ( param->u.s.string) 
		g_free( param->u.s.string);
	    param->u.s.string = g_strdup( s);
	    break;
	case FIND_WIDTH:
	case FIND_HEIGHT:
	case FIND_DEPTH:
	case FIND_FILE_SIZE:
	    s = gtk_entry_get_text( GTK_ENTRY( param->u.t.entry));
	    if ( !s) { 
		GList *tmp = list;
		list = list->next;
		find->params = g_list_remove_link( find->params, tmp);
		continue;
		break;
	    }
	    n = strtol( s, NULL, 10);
	    if ( errno == ERANGE) 
		g_warning( "%s", strerror( errno));
	    param->u.t.number = n;
	    break;
	case FIND_PROPERTY:
	    
	default:
	    break;
	}
	list = list->next;
    }

    /*
    image_count = 0;
    find->percentage = 0.0;

    if ( find->type == SRC_GIMDBS) {
	GImDB *gimdb;

	list = find->u.g.gimdbs;
	while ( list) {
	    gimdb = list->data;
	    list = list->next;

	    image_count += gimdb->imdb->header.image_count;
	}

    } else 
	image_count += find->u.b.browser->image_count;

    find->increment = 1.0 / image_count;
    */

    switch( find->type) {
    case SRC_GIMDBS:
	g_return_if_fail( find->u.g.gimdbs != NULL);
	find->u.g.gimdb = find->u.g.gimdbs;
	find->u.g.directory = 1;
	find->u.g.image = g_array_index( ((GImDB*)find->u.g.gimdb->data)->imdb->dirtree, 
					 ImDBDirectory, 1).images; 
	find->idler_tag = gtk_idle_add( (GtkFunction) find_gimdb_idler, find);
	break;
    case SRC_IMAGE_BROWSER:
	find->u.b.image = find->u.b.browser->images;
	find->idler_tag = gtk_idle_add( (GtkFunction) find_browser_idler, find);
	break;
    default:
    }

    gtk_widget_show( find->browser->mainbox->parent);

}

static gint
find_gimdb_idler( Find *find) {
    if ( !find->u.g.image) {
	if ( find->u.g.directory < 
	     ((GImDB*) find->u.g.gimdb->data)->imdb->header.directory_count) {
	    /* scan the next directory in the same database */
	    find->u.g.image = g_array_index( ((GImDB*)find->u.g.gimdb->data)->imdb->dirtree, 
					     ImDBDirectory, 
					     find->u.g.directory).images;
	    find->u.g.directory++;
	    return TRUE;
	} else if ( find->u.g.gimdb->next) {
	    /* scan the next database */
	    find->u.g.directory = 1;
	    find->u.g.gimdb = find->u.g.gimdb->next;
	    find->u.g.image = g_array_index( ((GImDB*)find->u.g.gimdb->data)->imdb->dirtree, 
					     ImDBDirectory, 1).images; 
	    return TRUE;
	} else {
	    /* end of search */
	    image_browser_set_images( find->browser, find->result);
	    find->idler_tag = 0;
	    return FALSE;
	}
    }
	    
    find_test_image( find, (Image*) find->u.g.image->data);
    find->u.g.image = find->u.g.image->next;

    return TRUE;
}

static gint 
find_browser_idler( Find *find) {
    if ( !find->u.b.image) {
	image_browser_set_images( find->browser, find->result);
	find->idler_tag = 0;
	return FALSE;
    }
    
    find_test_image( find, (Image*) find->u.b.image->data);
    find->u.b.image = find->u.b.image->next;

    return TRUE;
}

static void
find_test_image( Find *find, Image *image) {
    GList *list, *tmp;
    FindParam *param;
    ImDBImage *dbim;
    char *s;
    guint32 n;

    g_return_if_fail( find && image);

    list = find->params;
    while ( list) {
	param = list->data;
	list = list->next;

	switch( param->type) {
	case FIND_NAME:
	    if ( !strings_match_case( image->file_name, 
				      param->u.s.string,
				      param->compare))
		return;
	    break;
	case FIND_WIDTH:
	    if ( !guint32_match( image->width, 
				param->u.t.number, 
				param->compare))
		return;
	    break;
	case FIND_HEIGHT:
	    if ( !guint32_match( image->height, 
				param->u.t.number, 
				param->compare))
		return;
	    break;
	case FIND_DEPTH:
	    n = image->flags & IMAGE_HAS_COLORS ? 24 : 8;
	    if ( !guint32_match( n, 
				param->u.t.number, 
				param->compare))
		return;
	    break;
	case FIND_FILE_SIZE:
	    if ( !guint32_match( image->file_size, 
				param->u.t.number, 
				param->compare))
		return;
	    break;
	case FIND_PROPERTY:
	    dbim = (ImDBImage*) image;
	    if ( !imdb_image_find_property( param->u.p.gimdb->imdb,
					    dbim,
					    param->u.p.property_id,
					    param->u.p.value_id))
		return;
	    break;
	case FIND_KEYWORD:
	    dbim = (ImDBImage*) image;
	    tmp = dbim->keywords;
	    n = FALSE;
	    while ( tmp) {
		s = tmp->data;
		tmp = tmp->next;
		
		if ( !strings_match_case( s, 
					  param->u.s.string,
					  param->compare))
		    n = TRUE;
	    }
	    if ( !n)
		return;
	    break;
	default:
	    return;
	    break;
	}

    }

    find->result = g_list_append( find->result, image_copy( image));
}

static gint
strings_match( char *s1, char *s2, FindCompare cmp) {
    switch( cmp) {
    case EQUAL:
	return !strcmp( s1, s2);
	break;
    case SMALLER:
	return strcmp( s1, s2) < 0;
	break;
    case BIGGER:
	return strcmp( s1, s2) > 0;
	break;
    case BEGIN_WITH:
	return !strncmp( s1, s2, strlen( s2));
	break;
    case END_WITH:
	s1 += strlen( s1) - strlen( s2);
	return !strncmp( s1, s2, strlen( s2));
	break;
    case REGEXP:
	/* xxx not implemented */
    default:
	break;
    }
    
    return FALSE;
}

static gint
strings_match_case( char *s1, char *s2, FindCompare cmp) {
    switch( cmp) {
    case EQUAL:
	return !strcasecmp( s1, s2);
	break;
    case SMALLER:
	return strcasecmp( s1, s2) < 0;
	break;
    case BIGGER:
	return strcasecmp( s1, s2) > 0;
	break;
    case BEGIN_WITH:
	return !_strncasecmp( s1, s2, strlen( s2));
	break;
    case END_WITH:
	s1 += strlen( s1) - strlen( s2);
	return !_strncasecmp( s1, s2, strlen( s2));
	break;
    case REGEXP:
	/* xxx not implemented */
    default:
	break;
    }
    
    return FALSE;
}


static gint
guint32_match( guint32 i1, guint32 i2, FindCompare cmp) {
    switch( cmp) {
    case EQUAL:
	return i1 == i2;
	break;
    case SMALLER:
	return i1 < i2;
	break;
    case BIGGER:
	return i1 > i2;
	break;
    case BEGIN_WITH:
    case END_WITH:
    case REGEXP:
	/* xxx not implemented */
    default:
	break;
    }
    
    return FALSE;
} 

static gint
_strncasecmp( char *s1, char *s2, gint n) {
  
    while( *s1 && *s2 && n--) 
	if ( toupper( *s1) != toupper( *s2))
	    return toupper( *s1) -  toupper( *s2);

    return n == 0;
}
