/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <gtk/gtk.h>


#include "gimdb.h"
#include "ubox.h"
#include "imageubox.h"
#include "find.h"

#include "gtkcompat.h"

typedef struct _PropertyPanelData {
    GImDB *gimdb;

    GtkWidget *tree;
    GtkCTreeNode *tree_root;
    GtkWidget *entry;

    GList *properties;
    GList *values;

} PropertyPanelData;

typedef struct _PropertyMenuData {
    GImDB *gimdb;
    guint property_id;
    guint value_id;
} PropertyMenuData;


static void              init_prefs( GImDB *gimdb);
static void              gimdb_destroy_cb( GtkWidget *w, GImDB *gimdb);
static void              gimdb_tree_build( GImDB *gimdb, GImDBTreeType type);
static GtkWidget*        gimdb_init_gui( GImDB *gimdb);

static gint              gimdb_tree_select_row_cb( GtkWidget *w, 
						   GtkCTreeNode *node, 
						   gint column, GImDB *gimdb);
static GtkCTreeNode*     gimdb_tree_append_path( GImDB *gimdb, char *path);
static GImDBImageMap*    gimdb_find_map( GImDB *gimdb, guint map_id);
static void              gimdb_select_cell( UBox *ubox, gint cell_no, GImDB *gimdb);
static gint              gimdb_image_browser_button_press( UBox *ubox, GdkEventButton *event, GImDB *gimdb);

/* menu callback functions */
static void              ifact_informations( GtkWidget *w, GImDB *gimdb);
static void              ifact_close( GtkWidget *w, GImDB *gimdb);
static void              ifact_find( GtkWidget *w, GImDB *gimdb);
static void              ifact_properties( GtkWidget *w, GImDB *gimdb);

/* properties' menu */ 
static void              property_add( GtkWidget *w, PropertyPanelData *panel);
static void              property_add_value( GtkWidget *w, PropertyPanelData *panel);
static void              property_remove( GtkWidget *w, PropertyPanelData *panel);
static void              property_destroy( GtkWidget *w, PropertyPanelData *panel);
static void              property_select( GtkCTree *tree, 
					  GtkCTreeNode *node, 
					  gint column, 
					  PropertyPanelData *panel);
static void              property_key_press( GtkWidget *w, 
					     GdkEventKey *event, 
					     PropertyPanelData *panel);
static GtkCTreeNode*     ctree_get_selection( GtkCTree *tree);

/* properties' popup menu to quickly add/remove a property from a set of images */
static GtkWidget*        property_menu_build( GImDB *gimdb);
static void              property_menu_destroy( GtkWidget *w, PropertyMenuData *data);
static void              property_menu_activate( GtkCheckMenuItem *menu_item, PropertyMenuData *data);
static gint              images_have_property( GImDB *gimdb,
					       GList *images, 
					       ImDBProperty *property, 
					       ImDBPropertyValue *value);
#define GTK_CTREE_SPACING 5
#define GTK_BOX_SPACING 7


static void
init_prefs( GImDB *gimdb) {

}

GImDB*
gimdb_load( char *filename) {
    GImDB *gimdb = NULL;
    ImDB  *imdb;
    GtkWidget *window;

    g_return_val_if_fail( filename != NULL, NULL);

    imdb = imdb_load( filename);
 
    if ( imdb) {
	gimdb = g_new( GImDB, 1);
	init_prefs( gimdb);

	gimdb->imdb = imdb;
	gimdb->image_browser = image_browser_new( NULL);
	gimdb->tree = gtk_ctree_new( 1, 0);
	gimdb->image_map = NULL;
	gimdb_tree_build( gimdb, GIMDB_DIRECTORY_TREE);

	window = gimdb_init_gui( gimdb);
	gtk_widget_set_usize( window, 700, -1);
	gtk_widget_show( window);
    }

    return gimdb;
}

GImDB*
gimdb_create( char *filename) {
    GImDB *gimdb = NULL;
    ImDB  *imdb;
    GtkWidget *window;

    g_return_val_if_fail( filename != NULL, NULL);
    
    imdb = imdb_create( filename);

    if ( imdb) {
	gimdb = g_new( GImDB, 1);
	init_prefs( gimdb);

	gimdb->imdb = imdb;
	gimdb->image_browser = image_browser_new( NULL);
	gimdb->tree = gtk_ctree_new( 1, 0);
	gimdb->image_map = NULL;
	gimdb_tree_build( gimdb, GIMDB_DIRECTORY_TREE);

	window = gimdb_init_gui( gimdb);
	gtk_widget_set_usize( window, 700, -1);
	gtk_widget_show( window);
    }

    return gimdb;
}

void
gimdb_close( GImDB *gimdb) {
    GtkWidget *window;

    window = gtk_widget_get_toplevel( gimdb->tree);
    gtk_widget_destroy( window);
}

static void              
gimdb_destroy_cb( GtkWidget *w, GImDB *gimdb) {
    if ( gimdb->image_map) {
	g_list_foreach( gimdb->image_map, (GFunc) g_free, NULL);
	g_list_free( gimdb->image_map);
    }

    imdb_close( gimdb->imdb);

    g_free( gimdb);
}

				  
static void
gimdb_tree_build( GImDB *gimdb, GImDBTreeType type) {
    GImDBImageMap *map;

    g_return_if_fail( gimdb != NULL);
    g_return_if_fail( gimdb->tree != NULL);

    gtk_clist_clear( GTK_CLIST( gimdb->tree));

    if ( type == GIMDB_DIRECTORY_TREE) {
	ImDBDirectory *dbnode;
	GtkCTreeNode **nodes, *node, *parent;
	guint node_id, size;
	char *root_str = "/";

	g_return_if_fail( gimdb->imdb != NULL && gimdb->imdb->dirtree != NULL);
	
	if ( GTK_WIDGET_VISIBLE( gimdb->tree))
	    gtk_clist_freeze( GTK_CLIST( gimdb->tree));

#ifndef GTK_HAVE_FEATURES_1_1_2
	gtk_ctree_set_auto_sort( GTK_CTREE( gimdb->tree), TRUE);
#endif
	size = (gimdb->imdb->header.directory_count + 1)*sizeof( GList*);
	nodes = g_malloc( size);
	bzero( nodes, size);
	
	/* insert the root directory */
	dbnode = &g_array_index( gimdb->imdb->dirtree, ImDBDirectory, 0);
	gimdb->tree_root = node = gtk_ctree_insert( GTK_CTREE( gimdb->tree),
						    NULL, NULL,
						    &root_str, GTK_CTREE_SPACING, 
						    NULL, NULL, NULL, NULL,
						    FALSE, TRUE);

	nodes[0] = node;
	
	for ( node_id = 1; node_id < gimdb->imdb->header.directory_count; node_id++) {
	    dbnode = &g_array_index( gimdb->imdb->dirtree, ImDBDirectory, node_id);
	    /* parent != NULL because dbnode->parent < node_id as each "child directory"
	       is added after its parent ( cf imdb.c : imdb_append_path). */
	    parent = nodes[ dbnode->parent];
	    g_return_if_fail( parent != NULL); 
	    
	    nodes[ node_id] = gtk_ctree_insert( GTK_CTREE( gimdb->tree),
						parent, NULL,
						&dbnode->name, GTK_CTREE_SPACING,
						NULL, NULL, NULL, NULL,
						dbnode->child == 0, TRUE);
	    map = g_new( GImDBImageMap, 1);
	    map->id = node_id;
	    map->images = dbnode->images;
	    gtk_ctree_set_row_data( GTK_CTREE( gimdb->tree),
				    nodes[ node_id],
				    map);
	}

#ifdef GTK_HAVE_FEATURES_1_1_2
	gtk_ctree_sort_recursive( GTK_CTREE( gimdb->tree), gimdb->tree_root);
#endif
	g_free( nodes);

    } else if ( type == GIMDB_PROPERTY_TREE) {
	ImDBProperty *property;
	ImDBPropertyValue *value;
	GtkCTreeNode *root, *node, *child;
	GList *p, *v;
	char *root_str = "Properties";

	g_return_if_fail( gimdb->imdb->properties != NULL);

	if ( GTK_WIDGET_VISIBLE( gimdb->tree))
	    gtk_clist_freeze( GTK_CLIST( gimdb->tree));

#ifndef GTK_HAVE_FEATURES_1_1_2
	gtk_ctree_set_auto_sort( GTK_CTREE( gimdb->tree), TRUE);
#endif
	/* Insert the root item */
	gimdb->tree_root = root = gtk_ctree_insert( GTK_CTREE( gimdb->tree),
						    NULL, NULL,
						    &root_str, GTK_CTREE_SPACING,
						    NULL, NULL, NULL, NULL,
						    FALSE, TRUE);
	
	p = gimdb->imdb->properties;
	while ( p) {
	    property = p->data;
	    p = p->next;

	    node = gtk_ctree_insert( GTK_CTREE( gimdb->tree),
				     root, NULL,
				     &property->name, GTK_CTREE_SPACING,
				     NULL, NULL, NULL, NULL,
				     FALSE, TRUE);
	    /* xxx 
	    gtk_ctree_set_row_data( GTK_CTREE( gimdb->tree),
				    node,
				    GUINT_TO_POINTER( (guint) property->id));
				    */
	    v = property->values;
	    while ( v) {
		value = v->data;
		v = v->next;

		child = gtk_ctree_insert( GTK_CTREE( gimdb->tree),
				     node, NULL,
				     &value->name, GTK_CTREE_SPACING,
				     NULL, NULL, NULL, NULL,
				     TRUE, FALSE);
		/* xxx
		gtk_ctree_set_row_data( GTK_CTREE( gimdb->tree),
					child, GUINT_TO_POINTER( (guint) value->id));
					*/
	    }
	}

#ifdef GTK_HAVE_FEATURES_1_1_2
	gtk_ctree_sort_recursive( GTK_CTREE( gimdb->tree), gimdb->tree_root);
#endif

    } else {
	g_warning( "gimdb_tree_build : Unknown tree type (%d)", type);
	type = gimdb->tree_type;
    }
    
    if ( GTK_WIDGET_VISIBLE( gimdb->tree))
	gtk_clist_thaw( GTK_CLIST( gimdb->tree));

    gimdb->tree_type = type;
}
    
void
gimdb_append_images( GImDB *gimdb, gchar *path, GList *images) {
    GtkWidget *window;
    GtkCTreeNode *node;
    GImDBImageMap *map;
    GList *dbims = NULL, *item;
    Image *image;
    ImDBImage *dbim;
    gchar *pc;
    gint dbim_count = 0;
    float progress, increment;

    g_return_if_fail( gimdb != NULL);
    g_return_if_fail( images != NULL);

    window = gtk_widget_get_toplevel( gimdb->tree);
    //gtk_widget_set_state( window, GTK_STATE_INSENSITIVE);

    pc = g_strdup( path);

    item = images;
    while ( item) {
	image = item->data;
	item = item->next;

	dbim = g_new( ImDBImage, 1);
	bzero( dbim, sizeof( ImDBImage));
	dbim->image.file_name = g_strdup( image->file_name);
	dbim->image.path = pc;
	dbims = g_list_append( dbims, dbim);
	dbim_count ++;
    }

    
    if ( dbims) {
	progress = 0.0;
	increment = 1.0 / dbim_count;

	item = dbims;
	while ( item) {
	    dbim = item->data;
	    if ( !image_read_header( &dbim->image)) {
		GList *next_item;

		g_free( dbim->image.file_name);
		g_free( dbim);
		next_item = item->next;
		dbims = g_list_remove_link( dbims, item);
		item = next_item;
	    } else
		item = item->next;

	    gtk_progress_set_percentage( GTK_PROGRESS( gimdb->image_browser->progress_bar),
					 progress);
	    progress += increment;

	    while ( gtk_events_pending())
		gtk_main_iteration();

	    if ( GTK_OBJECT( window)->flags & GTK_DESTROYED) {
		item = dbims;
		while ( item) {
		    dbim = item->data;
		    item = item->next;
		    
		    g_free( dbim->image.file_name);
		    g_free( dbim);
		}
		if ( dbims)
		    g_list_free( dbims);
		
		return;
	    }
	}
	gtk_progress_set_percentage( GTK_PROGRESS( gimdb->image_browser->progress_bar), 0.0);
	imdb_append_images( gimdb->imdb, path, dbims);
	
	map = gimdb_find_map( gimdb, dbim->directory);
	if ( map) /* xxx need to uptdate the "image_box" if this directory was selected */
	    map->images = g_list_concat( map->images, dbims);
	else {
	    map = g_new( GImDBImageMap, 1);
	    map->id = dbim->directory;
	    map->images = dbims;
	    gimdb->image_map = g_list_append( gimdb->image_map, map);

	    node = gimdb_tree_append_path( gimdb, image->path);
	    gtk_ctree_set_row_data( GTK_CTREE( gimdb->tree), node, map);
	}
	

    }

    g_free( pc);

    //gtk_widget_set_state( window, GTK_STATE_NORMAL);

}

static gint        
gimdb_tree_select_row_cb( GtkWidget *w, GtkCTreeNode *node, gint column, GImDB *gimdb) {
    GImDBImageMap *map;
    ImDBDirectory *dir;

    g_return_val_if_fail( node != NULL, FALSE);
    g_return_val_if_fail( gimdb != NULL, FALSE);
    

    map = gtk_ctree_get_row_data( GTK_CTREE( gimdb->tree), node);
    if ( !map)
	return FALSE;

    dir = &g_array_index( gimdb->imdb->dirtree, ImDBDirectory, map->id);
    g_return_val_if_fail( dir->path != NULL, FALSE);

    if ( !access( dir->path, R_OK | X_OK)) 
	image_browser_set_images( gimdb->image_browser, map->images);
    else
	g_warning( "Can't access directory %s (%s)", dir->path, g_strerror( errno));

    return FALSE;
}

static GImDBImageMap*
gimdb_find_map( GImDB *gimdb, guint map_id) {
    GImDBImageMap *map;
    GList *item;

    g_return_val_if_fail( gimdb != NULL, NULL);
     
    item = gimdb->image_map;
    while ( item) {
	map = item->data;
	item = item->next;

	if ( map->id == map_id)
	    return map;
    }

    return NULL;
}

static void
gimdb_select_cell( UBox *ubox, gint cell_no, GImDB *gimdb) {

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( gimdb != NULL);


}

static gint
gimdb_image_browser_button_press( UBox *ubox, GdkEventButton *event, GImDB *gimdb) {
    GtkWidget *menu;
    gint cell_no;

    g_return_val_if_fail( ubox != NULL, FALSE);
    g_return_val_if_fail( event != NULL, FALSE);
    g_return_val_if_fail( gimdb != NULL, FALSE);

    if ( event->button == 3 && event->type == GDK_BUTTON_PRESS) {
	cell_no = ubox_cell_from_coord( ubox, event->x, event->y);
	
	if ( cell_no != -1 ) { /* xxx check if this cell is selected */
	    menu = property_menu_build( gimdb);
	    if ( menu) 
		gtk_menu_popup( GTK_MENU( menu), NULL, NULL, NULL, NULL, event->button, 0);
	}
    }
    return FALSE;
}

    
static GtkCTreeNode*     
gimdb_tree_append_path( GImDB *gimdb, char *path) {
    GtkCTreeNode *node, *child;
    char *pc, *dir;
    gint tree_column = GTK_CTREE( gimdb->tree)->tree_column;

    node = gimdb->tree_root;
    g_return_val_if_fail( node != NULL, NULL);

    pc = g_strdup( path);
    dir = strtok( pc, "/");
    while ( dir) {
	child = GTK_CTREE_ROW( node)->children;
	while ( child && strcmp( dir, GTK_CELL_PIXTEXT( GTK_CTREE_ROW( child)->row.cell[ tree_column])->text))
	    child = GTK_CTREE_ROW( node)->sibling;
	if ( !child)
	    break;

	node = child;
	dir = strtok( NULL, "/");
    }

    
    while ( dir) {
	node = gtk_ctree_insert( GTK_CTREE( gimdb->tree),
				 node, NULL,
				 &dir, GTK_CTREE_SPACING,
				 NULL, NULL, NULL, NULL,
				 FALSE, TRUE);
#ifdef GTK_HAVE_FEATURES_1_1_2
	gtk_ctree_sort_node( GTK_CTREE( gimdb->tree), node);
#endif
	dir = strtok( NULL, "/");
    }

    return node;
}


static GtkWidget*
gimdb_init_gui( GImDB *gimdb) {
    GtkWidget *window;
    GtkWidget *mainbox;
    GtkWidget *menubar;
    GtkWidget *menu;
    GtkWidget *menuitem;
    GtkWidget *paned;
    GtkWidget *frame;
    GtkItemFactory *menubar_factory, *menu_factory;
    GtkItemFactoryEntry menu_entries[] = {
    {"/File/"                        , NULL                  , NULL                    , 2, "<Separator>"  },
    {"/File/Informations"            , "<control>I"          , ifact_informations      , 2, NULL           },
    {"/File/Close"                   , "<control>C"          , ifact_close             , 2, NULL           }, 
    {"/Database/Find"                , "<control>F"          , ifact_find              , 2, NULL           },
    {"/Database/Properties"          , NULL                  , ifact_properties        , 2, NULL           } };
    char buff[64];

    menubar_factory = gtk_item_factory_new( GTK_TYPE_MENU_BAR, "<MenuBar>", NULL);
    gtk_item_factory_create_items_ac( menubar_factory, 
				      sizeof( menu_entries) / sizeof( GtkItemFactoryEntry),
				      menu_entries,
				      gimdb, 
				      2);
    menu_factory = image_browser_get_menu_factory( gimdb->image_browser);
    
    window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
    snprintf( buff, 64, "jview: %s", strrchr( gimdb->imdb->db_path, '/') + 1);
    gtk_window_set_title( GTK_WINDOW( window), buff);
    gtk_signal_connect( GTK_OBJECT( window), "destroy",
			(GtkSignalFunc) gimdb_destroy_cb, gimdb);
    gtk_window_add_accel_group( GTK_WINDOW( window), menubar_factory->accel_group);
    gtk_window_add_accel_group( GTK_WINDOW( window), menu_factory->accel_group);

    mainbox = gtk_vbox_new( FALSE, 0);
    gtk_container_add( GTK_CONTAINER( window), mainbox);
    gtk_widget_show( mainbox);

    menubar = gtk_item_factory_get_widget( menubar_factory, "<MenuBar>");
    gtk_box_pack_start( GTK_BOX( mainbox), menubar, FALSE, TRUE, 0);
    gtk_widget_show( menubar);

    menu = gtk_item_factory_get_widget( menu_factory, "<ImageBrowser>");

#ifdef GTK_HAVE_FEATURES_1_1_2
    menuitem = gtk_tearoff_menu_item_new();
    gtk_menu_prepend( GTK_MENU( menu), menuitem);
    gtk_widget_show( menuitem);
#endif

    menuitem = gtk_menu_item_new_with_label( "Browser");
    gtk_menu_item_set_submenu( GTK_MENU_ITEM( menuitem), menu);
    gtk_menu_bar_insert( GTK_MENU_BAR( menubar), menuitem, 1);
    gtk_widget_show( menuitem);

    paned = gtk_hpaned_new();
    gtk_container_add( GTK_CONTAINER( mainbox), paned);
    gtk_widget_show( paned);

    /* first frame : tree */
    frame = gtk_frame_new( NULL);
    gtk_paned_add1( GTK_PANED( paned), frame);
    gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);
    gtk_widget_show( frame);


    gtk_clist_set_policy( GTK_CLIST( gimdb->tree),
			  GTK_POLICY_AUTOMATIC,
			  GTK_POLICY_AUTOMATIC);
    gtk_signal_connect( GTK_OBJECT( gimdb->tree), "tree_select_row",
			      (GtkSignalFunc) gimdb_tree_select_row_cb, gimdb);
    gtk_container_add( GTK_CONTAINER( frame), gimdb->tree);
    gtk_widget_show( gimdb->tree);

    /* second frame : image browser */
    frame = gtk_frame_new( NULL);
    gtk_paned_add2( GTK_PANED( paned), frame);
    gtk_frame_set_shadow_type( GTK_FRAME( frame), GTK_SHADOW_IN);
    gtk_widget_show( frame);

    gtk_container_add( GTK_CONTAINER( frame), gimdb->image_browser->mainbox);
    gtk_signal_connect( GTK_OBJECT( gimdb->image_browser->preview_box), "select_cell",
			(GtkSignalFunc) gimdb_select_cell, gimdb);
    gtk_signal_connect( GTK_OBJECT( gimdb->image_browser->preview_box), "button_press_event",
			(GtkSignalFunc) gimdb_image_browser_button_press, gimdb);
    ubox_set_selection_mode( UBOX( gimdb->image_browser->preview_box), GTK_SELECTION_EXTENDED);
    gtk_widget_show( gimdb->image_browser->mainbox);

    return window;
}

/****************************** 
 *  Menu's callback functions *
 ******************************/
 
static void              
ifact_informations( GtkWidget *w, GImDB *gimdb) {
    GtkWidget *window;
    GtkWidget *mainbox;
    GtkWidget *hbox, *vbox1, *vbox2;
    GtkWidget *label;
    char buff[64];

    window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title( GTK_WINDOW( window), "Database infos");
    gtk_container_border_width( GTK_CONTAINER( window), 10);

    mainbox = gtk_vbox_new( FALSE, 0);
    gtk_container_add( GTK_CONTAINER( window), mainbox);

    hbox = gtk_hbox_new( FALSE, 0);
    gtk_box_pack_start( GTK_BOX( mainbox), hbox, FALSE, TRUE, 0);

    vbox1 = gtk_vbox_new( FALSE, 5);
    gtk_container_add( GTK_CONTAINER( hbox), vbox1);

    vbox2 = gtk_vbox_new( FALSE, 5);
    gtk_container_add( GTK_CONTAINER( hbox), vbox2);

    label = gtk_label_new( "Name :");
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_RIGHT);
    gtk_container_add( GTK_CONTAINER( vbox1), label);
    label = gtk_label_new( strrchr( gimdb->imdb->db_path, '/') + 1);
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_LEFT);
    gtk_container_add( GTK_CONTAINER( vbox2), label);

    label = gtk_label_new( "Image count :");
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_RIGHT);
    gtk_container_add( GTK_CONTAINER( vbox1), label);
    sprintf( buff, "%d", gimdb->imdb->header.image_count);
    label = gtk_label_new( buff);
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_LEFT);
    gtk_container_add( GTK_CONTAINER( vbox2), label);

    label = gtk_label_new( "Property count :");
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_RIGHT);
    gtk_container_add( GTK_CONTAINER( vbox1), label);
    sprintf( buff, "%d", gimdb->imdb->header.property_count);
    label = gtk_label_new( buff);
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_LEFT);
    gtk_container_add( GTK_CONTAINER( vbox2), label);

    label = gtk_label_new( "Directory count :");
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_RIGHT);
    gtk_container_add( GTK_CONTAINER( vbox1), label);
    sprintf( buff, "%d", gimdb->imdb->header.directory_count);
    label = gtk_label_new( buff);
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_LEFT);
    gtk_container_add( GTK_CONTAINER( vbox2), label);

    label = gtk_label_new( "Block size :");
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_RIGHT);
    gtk_container_add( GTK_CONTAINER( vbox1), label);
    sprintf( buff, "%d", gimdb->imdb->ldb->header.block_size);
    label = gtk_label_new( buff);
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_LEFT);
    gtk_container_add( GTK_CONTAINER( vbox2), label);

    label = gtk_label_new( "Block count :");
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_RIGHT);
    gtk_container_add( GTK_CONTAINER( vbox1), label);
    sprintf( buff, "%d", gimdb->imdb->ldb->header.block_count);
    label = gtk_label_new( buff);
    gtk_label_set_justify( GTK_LABEL( label), GTK_JUSTIFY_LEFT);
    gtk_container_add( GTK_CONTAINER( vbox2), label);

    gtk_widget_show_all( window);
}


static void
ifact_close( GtkWidget *w, GImDB *gimdb) {
    gtk_widget_destroy( gtk_widget_get_toplevel( gimdb->tree));
}

static void
ifact_find_remove_destroy_handler( GtkWidget *top2, GtkWidget *top1) {
    gtk_signal_disconnect_by_data( GTK_OBJECT( top1), top2);
}
 
static void              
ifact_find( GtkWidget *w, GImDB *gimdb) {
    GtkWidget *find, *top1, *top2;

    find = gimdb_find_new( g_list_append( NULL, gimdb));
    if ( find) {
	top1 = gtk_widget_get_toplevel( gimdb->tree);
	top2 = gtk_widget_get_toplevel( find);

	/* xxx the signal handler seems not to be removed when
	   the top2 widget is destroyed */
	gtk_signal_connect_object( GTK_OBJECT( top1), "destroy",
				   (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT( top2));
	gtk_signal_connect( GTK_OBJECT( top2), "destroy",
			    (GtkSignalFunc) ifact_find_remove_destroy_handler, top1);
    }

}

static void              
ifact_properties( GtkWidget *w, GImDB *gimdb) {
    PropertyPanelData *panel;
    GtkWidget *window, *mainbox, *hbox, *sep, *button;
    GtkCTreeNode *n1, *n2;
    ImDBProperty *prop;
    ImDBPropertyValue *val;
    GList *i1, *i2;
    char *root_str = "Properties";

    panel = g_new( PropertyPanelData, 1);
    panel->gimdb = gimdb;
    panel->tree = gtk_ctree_new( 1, 0);
    panel->entry = gtk_entry_new();
    panel->values = NULL;
    panel->properties = NULL;
 
    panel->tree_root = gtk_ctree_insert( GTK_CTREE( panel->tree),
					 NULL, NULL,
					 &root_str,
					 GTK_CTREE_SPACING,
					 NULL, NULL, NULL, NULL,
					 FALSE, TRUE);
    i1 = gimdb->imdb->properties;
    while ( i1) {
	prop = i1->data;
	i1 = i1->next;

	n1 = gtk_ctree_insert( GTK_CTREE( panel->tree),
			       panel->tree_root, NULL,
			       &prop->name,
			       GTK_CTREE_SPACING,
			       NULL, NULL, NULL, NULL,
			       FALSE, TRUE);
	gtk_ctree_set_row_data( GTK_CTREE( panel->tree), n1, prop);

	i2 = prop->values;
	while ( i2) {
	    val = i2->data;
	    i2 = i2->next;

	    n2 = gtk_ctree_insert( GTK_CTREE( panel->tree),
				   n1, NULL,
				   &val->name,
				   GTK_CTREE_SPACING,
				   NULL, NULL, NULL, NULL,
				   TRUE, FALSE);
	    gtk_ctree_set_row_data( GTK_CTREE( panel->tree), n2, val);
	}

    }

    /* build widgets */
    window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect( GTK_OBJECT( window), "destroy",
			(GtkSignalFunc) property_destroy, panel);
    gtk_signal_connect( GTK_OBJECT( window), "key_press_event",
			(GtkSignalFunc) property_key_press, panel);

    mainbox = gtk_vbox_new( FALSE, GTK_BOX_SPACING);
    gtk_container_add( GTK_CONTAINER( window), mainbox);

    gtk_clist_set_policy( GTK_CLIST( panel->tree),
			  GTK_POLICY_AUTOMATIC,
			  GTK_POLICY_AUTOMATIC);
#ifdef GTK_HAVE_FEATURES_1_1_2
    gtk_ctree_sort_recursive( GTK_CTREE( gimdb->tree), gimdb->tree_root);
#else 
    gtk_ctree_set_auto_sort( GTK_CTREE( panel->tree), TRUE);
#endif

    gtk_container_add( GTK_CONTAINER( mainbox), panel->tree);
    gtk_signal_connect( GTK_OBJECT( panel->tree), "tree_select_row",
			(GtkSignalFunc) property_select, panel);

    sep = gtk_hseparator_new();
    gtk_box_pack_start( GTK_BOX( mainbox), sep, FALSE, TRUE, GTK_BOX_SPACING);

    gtk_box_pack_start( GTK_BOX( mainbox), panel->entry, FALSE, TRUE, GTK_BOX_SPACING);

    hbox = gtk_hbox_new( FALSE, GTK_BOX_SPACING);
    gtk_box_pack_start( GTK_BOX( mainbox), hbox, FALSE, TRUE, GTK_BOX_SPACING);

    button = gtk_button_new_with_label( "Add Property");
    gtk_container_add( GTK_CONTAINER( hbox), button);
    gtk_signal_connect( GTK_OBJECT( button), "clicked",
			(GtkSignalFunc) property_add, panel);

    button = gtk_button_new_with_label( "Add Value");
    gtk_container_add( GTK_CONTAINER( hbox), button);
    gtk_signal_connect( GTK_OBJECT( button), "clicked",
			(GtkSignalFunc) property_add_value, panel);

    button = gtk_button_new_with_label( "Remove");
    gtk_container_add( GTK_CONTAINER( hbox), button);
    gtk_signal_connect( GTK_OBJECT( button), "clicked",
			(GtkSignalFunc) property_remove, panel);

    gtk_widget_show_all( window);
}

static void              
property_add( GtkWidget *w, PropertyPanelData *panel) {
    ImDBProperty  *prop;
    GtkCTreeNode *n1;
    gchar *name;

    name = gtk_entry_get_text( GTK_ENTRY( panel->entry));
    if ( !name || !*name)
	return;

    prop = imdb_property_add( panel->gimdb->imdb, name);

    n1 = gtk_ctree_insert( GTK_CTREE( panel->tree),
			   panel->tree_root, NULL,
			   &name,
			   GTK_CTREE_SPACING,
			   NULL, NULL, NULL, NULL,
			   FALSE, TRUE);
#ifdef GTK_HAVE_FEATURES_1_1_2
    gtk_ctree_sort_node( GTK_CTREE( panel->tree), panel->tree_root);
#endif
    gtk_ctree_set_row_data( GTK_CTREE( panel->tree), n1, prop);
    gtk_ctree_select( GTK_CTREE( panel->tree), n1);

    gtk_entry_set_text( GTK_ENTRY( panel->entry), "");
}


static void              
property_add_value( GtkWidget *w, PropertyPanelData *panel) {
    ImDBProperty  *prop;
    ImDBPropertyValue *val;
    GtkCTreeNode *n1, *n2;
    gchar *name;

    g_return_if_fail( panel != NULL);

    name = gtk_entry_get_text( GTK_ENTRY( panel->entry));
    if ( !name || !*name)
	return;

    n1 = ctree_get_selection( GTK_CTREE( panel->tree));
    if ( !n1)
	return;

    prop = gtk_ctree_get_row_data( GTK_CTREE( panel->tree), n1);
    if ( !prop)
	return;

    if ( !g_list_find( panel->gimdb->imdb->properties, prop)) {
	n1 = GTK_CTREE_ROW( n1)->parent;
	prop = gtk_ctree_get_row_data( GTK_CTREE( panel->tree), n1);
	g_return_if_fail( g_list_find( panel->gimdb->imdb->properties, prop));
	if ( !prop)
	    return;
    }

    
    val = imdb_property_add_value( panel->gimdb->imdb, prop, name);
    
    n2 = gtk_ctree_insert( GTK_CTREE( panel->tree),
			   n1, NULL,
			   &name,
			   GTK_CTREE_SPACING,
			   NULL, NULL, NULL, NULL,
			   TRUE, FALSE);
#ifdef GTK_HAVE_FEATURES_1_1_2
    gtk_ctree_sort_node( GTK_CTREE( panel->tree), n1);
#endif
    gtk_ctree_set_row_data( GTK_CTREE( panel->tree), n2, val);

    gtk_entry_set_text( GTK_ENTRY( panel->entry), "");
}

static void              
property_remove( GtkWidget *w, PropertyPanelData *panel) {
    ImDBProperty  *prop;
    GtkCTreeNode *n1;

    g_return_if_fail( panel != NULL);

    n1 = ctree_get_selection( GTK_CTREE( panel->tree));
    if ( !n1)
	return;

    prop = gtk_ctree_get_row_data( GTK_CTREE( panel->tree), n1);
    if ( !prop)
	return;

    if ( g_list_find( panel->gimdb->imdb->properties, prop)) {
	/* Remove a property */
	imdb_property_delete( panel->gimdb->imdb, prop);
	gtk_ctree_remove( GTK_CTREE( panel->tree), n1);

    } else {
	ImDBPropertyValue *val;
	GtkCTreeNode *n2;
	/* Remove a property value */
	
	n2 = n1;
	n1 = GTK_CTREE_ROW( n2)->parent;
	prop = gtk_ctree_get_row_data( GTK_CTREE( panel->tree), n1);
	val  = gtk_ctree_get_row_data( GTK_CTREE( panel->tree), n2);
	if ( !prop)
	    return;
	
	imdb_property_delete_value( panel->gimdb->imdb, prop, val);
	gtk_ctree_remove( GTK_CTREE( panel->tree), n2);
    }
    
    gtk_entry_set_text( GTK_ENTRY( panel->entry), "");
}

static void              
property_select( GtkCTree *tree, GtkCTreeNode *node, gint column, PropertyPanelData *panel) {
    gint tree_column = GTK_CTREE( tree)->tree_column;
    gchar *text;

    text = GTK_CELL_PIXTEXT( GTK_CTREE_ROW( node)->row.cell[ tree_column])->text;
    if ( text && node != panel->tree_root) 
	gtk_entry_set_text( GTK_ENTRY( panel->entry), text);
}

static void
property_key_press( GtkWidget *w, GdkEventKey *event, PropertyPanelData *panel) {

    if ( !GTK_WIDGET_HAS_FOCUS( panel->entry)) {
	gtk_entry_set_text( GTK_ENTRY( panel->entry), "");
	gtk_widget_grab_focus( panel->entry);
    }
}


static void
property_destroy( GtkWidget *w, PropertyPanelData *panel) {
    g_free( panel);
}


static GtkCTreeNode*
ctree_get_selection( GtkCTree *tree) {
    GtkCTreeNode *node = NULL;
    GtkCListRow *row;
    GList *item;
    

    item = GTK_CLIST( tree)->row_list;
    while ( item) {
	row = item->data;

	if ( row->state & GTK_STATE_SELECTED) {
	    node = (GtkCTreeNode*) item;
	    break;
	}

	item = item->next;
    }

    return node;
}
	    
/************************************************************************** 
 * Property menu : handle adding and removing properties via a popup menu *
 **************************************************************************/


static GtkWidget*
property_menu_build( GImDB *gimdb) {
    ImDBProperty *property;
    ImDBPropertyValue *value;
    GtkWidget *menu, *submenu;
    GtkWidget *menu_item;
    PropertyMenuData *data;
    GList *i1, *i2, *sel;

    g_return_val_if_fail( gimdb != NULL, FALSE);

    sel = ubox_get_selection( UBOX( gimdb->image_browser->preview_box));
    if ( !sel)
	return NULL;

    menu = gtk_menu_new();

    i1 = gimdb->imdb->properties;
    while ( i1) {
	property = i1->data;
	i1 = i1->next;

	if ( property->values) {
	    submenu = gtk_menu_new();

	    i2 = property->values;
	    while ( i2) {
		value = i2->data;
		i2 = i2->next;

		data = g_new( PropertyMenuData, 1);
		data->gimdb = gimdb;
		data->property_id = property->property_id;
		data->value_id = value->id;

		menu_item = gtk_check_menu_item_new_with_label( value->name);
		if ( images_have_property( gimdb, sel, property, value))
		    gtk_check_menu_item_set_state( GTK_CHECK_MENU_ITEM( menu_item), TRUE);
		else
		    gtk_check_menu_item_set_state( GTK_CHECK_MENU_ITEM( menu_item), FALSE);
		gtk_menu_append( GTK_MENU( submenu), menu_item);
		gtk_signal_connect( GTK_OBJECT( menu_item), "activate", 
				    (GtkSignalFunc) property_menu_activate, data);
		gtk_signal_connect( GTK_OBJECT( menu_item), "destroy",
				    (GtkSignalFunc) property_menu_destroy, data);
		gtk_widget_show( menu_item);

	    }
	    
	}

	menu_item = gtk_menu_item_new_with_label( property->name);
	gtk_menu_append( GTK_MENU( menu), menu_item);
	if ( property->values) 
	    gtk_menu_item_set_submenu( GTK_MENU_ITEM( menu_item), submenu);
	else
	    gtk_widget_set_state( menu_item, GTK_STATE_INSENSITIVE);
	gtk_widget_show( menu_item);

    }

    g_list_free( sel);

    return menu;
}


#define UBOX_CELL_IMAGE( cell) ( (ImDBImage*) ((UBoxCell*) cell)->data)

static gint
images_have_property( GImDB *gimdb, GList *images, ImDBProperty *property, ImDBPropertyValue *value) {
    
    while ( images) {
	if ( !imdb_image_find_property( gimdb->imdb, 
					UBOX_CELL_IMAGE( images->data), 
					property->property_id, value->id))
	    return FALSE;
	images = images->next;
    }

    return TRUE;
}


static void
property_menu_activate( GtkCheckMenuItem *menu_item, PropertyMenuData *data) {
    ImDBImage *image;
    GSList *sel;
    g_return_if_fail( menu_item != NULL);
    g_return_if_fail( data != NULL);
    
    sel = UBOX( data->gimdb->image_browser->preview_box)->selection;
    if ( menu_item->active) {
	GtkWidget *menu;
	GtkCheckMenuItem *w;
	GList *children, *item;
	
	/* adds a new property to all images in the selection */
	while ( sel) {
	    image = (ImDBImage*) UBOX_CELL( sel)->data;
	    sel = sel->next;
	    imdb_image_add_property( data->gimdb->imdb, image, data->property_id, data->value_id);
	}
	
	menu = GTK_WIDGET( menu_item)->parent;
	children = gtk_container_children( GTK_CONTAINER( menu));
 	
	/* make others checkmenuitem(s) not "active" */
	item = children;
	while ( item) {
	    w = item->data;
	    item = item->next;
	    
	    if ( w != menu_item && w->active)
		gtk_check_menu_item_set_state( w, FALSE);
	}
	
	if ( children)
	    g_list_free( children);
    
    } else {
	/* remove this property from selected images */
	while ( sel) {
	    image = (ImDBImage*) UBOX_CELL( sel)->data;
	    sel = sel->next;
	    imdb_image_delete_property( data->gimdb->imdb, image, data->property_id, data->value_id);
	}
	
    }

}

static void
property_menu_destroy( GtkWidget *w, PropertyMenuData *data) {
    g_free( data);
}
