/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <glib.h>
#include <gtk/gtk.h>

#include "gdkrgb/gdkrgb.h"
#include "image.h"

//#define DEBUG


#define ROW_BUFF_OPTIMAL_SIZE 131072

#define PREVIEW( image) ( (RealPreview*)(image)->preview)
#define APPLY_ALPHA( src, dst, alpha) ( (alpha * (dst - src) + (src << 8)) >> 8)

/* it refuses to compile
#define ROW_APPENDED(image,preview,dest)                                    G_STMT_START{    \            
	(preview)->data->row_buff_rows++;                                                    \
	if ( (preview)->data->row_buff_rows == (preview)->data->row_buff_size) {             \
	    image_flush_buff( image);                                                        \
	    dest = (preview)->data->row_buff;                                                \
	}                                                                                    \
	                                                                                     \
	if ( (preview)->data->progress_func) {                                               \
	    if ( !(*(preview)->data->progress_func) ( image,                                 \
						    (preview)->data->percentage,             \
						    (preview)->data->progress_func_data)) {  \
		(preview)->data->aborted_by_user = TRUE;                                     \
		return FALSE;                                                                \
	    }                                                                                \
	    (preview)->data->percentage += (preview)->data->increment;                       \
	}                                                                     }G_STMT_END */

typedef struct _RealPreview RealPreview;
typedef struct _RenderData RenderData;

                       
struct _RealPreview {
    /* 
       the public part
    */ 
    FILE *file;
    gint width;
    gint height;
    guint flags;
    ImageRenderFunc render_func;

    /* 
       for offscrenn destinations ( ie pixmaps) 
     */
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    
    guchar *image_cache;
    
    ImageColorType color_type;

    RenderData *data;
};

struct _RenderData {
    gint flags;
    
    GdkWindow *window;
    GdkGC     *gc;

    GdkColor  background;
    guchar    bg8;

    /* box_width & box_height define an imaginary box in wich the rendered image will fit */
    gint box_width;    
    gint box_height;

    /* dimensions of the source */
    gint src_width;
    gint src_height;

    /* vars used for prgress monitoring */
    gfloat percentage;
    gfloat increment;

    /* 
       some variables used for scaling 
    */
    float x_scale;
    float y_scale;
    float y_pos;

    /* where to put the pixmap in the 'window' */
    gint  x_dest;
    gint  y_dest;

    gint appended_rows;     /* number of source rows appended */
    gint next_row_to_draw;   
    
    guchar *row_buff;        /* buffer used for scaling */
    gint   row_buff_size;   /* size of this buffer in rows */
    gint   row_buff_rows;   /* number of rows in the buffer */

    /* callback used for progress monitoring */
    ImageProgressFunc progress_func; 
    gpointer progress_func_data;

    /* used to tell if the rendering process was aborted by the user 
       (ie the progress monitoring function returned FALSE) or
       by an error in the decompression process */
    gint  aborted_by_user : 1;
};

static gint            image_render_init_notcalled( void);
static gint            image_render_skip_rows( void);
static void            image_calc_preview_size( Image *image);
static gint            image_append_rows_fullsize_24( Image *image, gint num_row, guchar *data);
static gint            image_append_rows_fullsize_8( Image *image, gint num_row, guchar *data);
static gint            image_append_rows_24( Image *image, gint num_row, guchar *data);
static gint            image_append_rows_8( Image *image, gint num_row, guchar *data);

static void            image_flush_buff( Image *image);
static ImageFormat*    get_image_format_from_extension( Image *image);
static ImageFormat*    get_image_format( Image *image);
static void            stat_image( Image *image, gchar *image_path);

static GSList *images_formats = NULL;


extern inline gint
row_appended( Image *image, RealPreview *preview, guchar **dest) {
    preview->data->row_buff_rows ++; 
    if ( preview->data->row_buff_rows == preview->data->row_buff_size) {
	image_flush_buff( image);
	*dest = preview->data->row_buff;
    }
    if ( preview->data->progress_func) { 
	if ( !(*preview->data->progress_func) ( image,
						preview->data->percentage,
						preview->data->progress_func_data)) {
	    preview->data->aborted_by_user = TRUE; 
	    return FALSE;
	}
	preview->data->percentage += preview->data->increment;
    }   
    return TRUE;
}

extern inline void
apply_alpha_24( guchar *src, guchar *dst, GdkColor *bg, gint pixcount) {
    while ( pixcount--) {
	
	dst[0] = APPLY_ALPHA( src[0], bg->red, src[3]);
	dst[1] = APPLY_ALPHA( src[1], bg->green, src[3]);
	dst[2] = APPLY_ALPHA( src[2], bg->blue, src[3]);
	
	src += 4;
	dst += 3;
    }
}
	
extern inline void
apply_alpha_8( guchar *src, guchar *dst, guchar bg, gint pixcount) {
    while ( pixcount--) {
	*dst++ = APPLY_ALPHA( src[0], bg, src[1]);

	src += 2;
    }
}

void
image_format_register( ImageFormat *image_format) {
    ImageFormat *format;
    char **ext;
    gint count = 0;

    g_return_if_fail( image_format != NULL);
    g_return_if_fail( image_format->extensions != NULL);

    format = g_malloc( sizeof( ImageFormat));
    format->check = image_format->check;
    format->read_header = image_format->read_header;
    format->read = image_format->read;
    format->write = image_format->write;

    ext = image_format->extensions;
    while ( *ext++) 
	count++;

    format->extensions = g_new( char*, count + 1);
    format->extensions[ count] = NULL;
    while (  count--)
	format->extensions[count] = g_strdup( image_format->extensions[count]);

    images_formats = g_slist_prepend( images_formats, format);
}

	
Image*           
image_new( char *file_name, char *path) {
    Image *image;

    g_return_val_if_fail( file_name != NULL, NULL);

    image = g_new( Image, 1);
    image->file_name = g_strdup( file_name);
    image->file_size = 0;
    image->path = path;
    image->type = 0;
    image->width = 0;
    image->height = 0;
    image->flags = 0;
    image->preview = NULL;
    
    return image;
}

Image*
image_copy( Image *image) {
    Image *copy;

    g_return_val_if_fail( image != NULL, NULL);

    copy = g_new( Image, 1);
    copy->file_name = g_strdup( image->file_name);
    copy->file_size = image->file_size;
    copy->path = image->path;
    copy->type = image->type;
    copy->width = image->width;
    copy->height = image->height;
    copy->flags = image->flags;
    copy->preview = NULL;
 
    return copy;
}

void             
image_free( Image *image) {
    g_return_if_fail( image != NULL);

    if ( image->preview)
	image_free_preview( image);

    g_free( image->file_name);
    g_free( image);
}

void
image_free_pixmaps( Image *image) {
     g_return_if_fail( image != NULL);
     
     if ( image->preview) {
	 if ( PREVIEW( image)->pixmap) {
	     gdk_pixmap_unref( PREVIEW( image)->pixmap);
	     PREVIEW( image)->pixmap = NULL;
	 }
	 
	 if ( PREVIEW( image)->mask) {
	     gdk_bitmap_unref( PREVIEW( image)->mask);
	     PREVIEW( image)->mask = NULL;
	 }
     }
}
    
void
image_free_preview( Image *image) {
    g_return_if_fail( image != NULL);

    if ( image->preview) {
	RealPreview *preview = PREVIEW( image);

	image_free_pixmaps( image);
	
	if ( preview->image_cache)
	    g_free( preview->image_cache);
	
	g_free( image->preview);
	image->preview = NULL;
    }
}

GdkPixmap*
image_get_pixmap( Image *image) {
    g_return_val_if_fail( image != NULL, NULL);

    return image->preview ? PREVIEW( image)->pixmap : NULL;
}

GdkBitmap*
image_get_mask( Image *image) {
    g_return_val_if_fail( image != NULL, NULL);

    return image->preview ? PREVIEW( image)->mask : NULL;
}



gint
image_read_header( Image *image) {
    ImageFormat *format;
    Preview preview = { 0};
    char path[ 4096];
    gint retval = TRUE;

    g_return_val_if_fail( image != NULL, FALSE);

    if ( image->width)
	return TRUE;

    image->preview = &preview;

    snprintf( path, 4096, "%s/%s", image->path, image->file_name);
    preview.file = fopen( path, "rb");
    if ( !preview.file) {
	g_warning( "Can't read image %s (%s)", path, strerror( errno));
	retval = FALSE;
    }

    if ( retval) {
	stat_image( image, path);
	format = get_image_format( image);
	if ( !format || !format->read_header) {
	    g_warning( "File format of %s not supported\n", path);
	    retval = FALSE;
	} else if ( !(*format->read_header)( image)) 
	    retval = FALSE;
    }


    fclose( preview.file);
    image->preview = NULL;
    
    return retval;
}

gint
image_render( Image *image,
	      GdkWindow *window,
	      GdkGC *gc,
	      GdkColor *background,
	      gint  x,
	      gint  y,
	      gint  width,
	      gint  height,
	      gint  flags,
	      ImageProgressFunc func,
	      gpointer func_data) {
    ImageFormat *format;
    RealPreview *preview;
    RenderData data = { 0}; /* xxx suppose that we only render 1 image at a time */
    char path[ 4096];
    gint failure = FALSE;

    g_return_val_if_fail( image != NULL, FALSE);
    g_return_val_if_fail( window != NULL && gc != NULL, FALSE);
    g_return_val_if_fail( (width > 0) ||  (width == -1), FALSE);
    g_return_val_if_fail( height > 0, FALSE);

    if ( !image->preview) {
	preview = g_new( RealPreview, 1);
	bzero( preview, sizeof( RealPreview));

	image->preview = (Preview*) preview;

    } else
	preview = PREVIEW( image);

    
    preview->data = &data;
    preview->render_func = (ImageRenderFunc) image_render_init_notcalled;

    data.flags  = flags;
    if ( (flags & IMAGE_CACHE) && preview->image_cache)
	data.flags &= ~IMAGE_CACHE;
    data.window = window;
    data.gc     = gc;

    if ( background)
	data.background = *background;
    else {
	GdkGCValues values;

	gdk_gc_get_values( gc, &values);
	data.background = values.background;
    }
    
    data.background.red    >>= 8;
    data.background.green  >>= 8;
    data.background.blue   >>= 8;
    /* xxx I don't know if it's the right value */
    data.bg8 = data.background.red | data.background.green | data.background.blue;
    data.x_dest = x;
    data.y_dest = y;

    data.box_width  = width;
    data.box_height = height;

    data.progress_func = func;
    data.progress_func_data = func_data;

    if ( preview->pixmap) {
	gint pix_width, pix_height;
	
	image_calc_preview_size( image);

	gdk_window_get_size( preview->pixmap, &pix_width, &pix_height);

	if ( pix_width == preview->width &&
	     pix_height == preview->height)
	    return TRUE;

	image_free_pixmaps( image);
    }

    if ( preview->image_cache) {
	/* image is already in the "cache" */

	image_calc_preview_size( image);
	
	if ( preview->width  == image->width &&
	     preview->height == image->height) { 

	    data.row_buff = preview->image_cache;
	    data.row_buff_size = image->height;
	    data.row_buff_rows = image->height;
	    
	    image_flush_buff( image);
	} else {
	    image_render_init( image, 
			       image->width, image->height, 
			       preview->color_type & IMAGE_COLOR_24 ? 
			       IMAGE_COLOR_24 : IMAGE_COLOR_8);
	    image_append_rows( image,
			       image->height,
			       preview->image_cache);
	    
	    image_flush_buff( image);

	    if ( data.row_buff) {
		g_free( data.row_buff);
		data.row_buff = NULL;
	    }
	}
	
	if ( func)
	    (*func) ( image, 1.0, func_data);

	return TRUE;
    } 
	
    snprintf( path, 4096, "%s/%s", image->path, image->file_name);
    preview->file = fopen( path, "r");
    if ( !preview->file) {
	g_warning( "Can't read image %s (%s)", path, strerror( errno));
	image_free_preview( image);
	if ( func)
	    (*func) ( image, 1.0, func_data);
	return FALSE;
    }
    stat_image( image, path);
    
    format = get_image_format( image);
    if ( !format || !format->read) {
	g_warning( "File format of %s not supported\n", path);
	failure = TRUE;
    } else if ( !(*format->read)( image)) {
	if ( !data.aborted_by_user)
	    g_warning( "Can't read image %s", path);
	failure = TRUE;
    }

    if ( func)
	(*func) ( image, 1.0, func_data);

    if ( failure) {
	fclose( preview->file);
	image_free_preview( image);
	
	return FALSE;
    }

    image_flush_buff( image);
      
    g_free( data.row_buff);
    
    fclose( preview->file);
    preview->file = NULL;
    
    return TRUE;
}

gint             
image_get_render_width( Image *image) {
    RealPreview *preview = PREVIEW( image);

    if ( !preview->width)
	image_calc_preview_size( image);

    return preview->data->flags & IMAGE_CACHE ? image->width : preview->width;
}


gint             
image_get_render_height( Image *image) {
    RealPreview *preview = PREVIEW( image);

    if ( !preview->width)
	image_calc_preview_size( image);

    return preview->data->flags & IMAGE_CACHE ? image->height : preview->height;
}


static void             
image_calc_preview_size( Image *image) {
    RealPreview *preview;

    g_return_if_fail( image != NULL);
    g_return_if_fail( image->preview != NULL);
    g_return_if_fail( image->width > 0 && image->height > 0);
    
    preview = PREVIEW( image);

    if ( preview->data->box_width == -1) {
	preview->width = image->width;
	preview->height = image->height;
	preview->data->box_width = preview->width;
	preview->data->box_height = preview->height;

    } else if ( preview->data->flags & IMAGE_FILL_BOX) { 
	preview->width  = preview->data->box_width;
	preview->height = preview->data->box_height;

    } else if ( !(preview->data->flags & IMAGE_MAGNIFY_IF_SMALLER) &&
		image->width <= preview->data->box_width &&
		image->height <= preview->data->box_height) {
	/* don't magnify images that are smaller than the container box */
	preview->width  = image->width;
	preview->height = image->height;


    } else { 
	float box_scale, img_scale;
	
	box_scale = (float) preview->data->box_width / preview->data->box_height;
	img_scale = (float) image->width / image->height;
	
	if ( img_scale < box_scale) {
	    preview->height = preview->data->box_height;
	    preview->width  = (int) ( img_scale * (float) preview->data->box_height);
	} else {
	    preview->width  = preview->data->box_width;
	    preview->height = (int) ( (float) preview->data->box_width / img_scale);
	}
    } 

}

void
image_get_preview_size( Image *image, gint box_width, gint box_height, gint flags, gint *width, gint *height) {
    RealPreview preview = { 0};
    RenderData data = { 0};
    Preview *saved_preview;

    g_return_if_fail( image != NULL);
 
    saved_preview = image->preview;

    PREVIEW( image) = &preview;
    preview.data = &data;
    data.box_width = box_width;
    data.box_height = box_height;
    data.flags = flags;

    image_calc_preview_size( image);

    *width = preview.width;
    *height = preview.height;

    image->preview = saved_preview;
}

void             
image_set_infos( Image *image, ImageType type,
		 gint width, gint height, gint has_colors,
		 gint has_alpha) {
    image->type = type;
    image->width = width;
    image->height = height;
    if ( has_colors)
	image->flags |= IMAGE_HAS_COLORS;
    if ( has_alpha)
	image->flags |= IMAGE_HAS_ALPHA;
}

static gint
image_render_init_notcalled( void) {
    g_warning( "you've got to call image_render_init before appending data\n");

    return FALSE;
}

static gint            
image_render_skip_rows( void) {
    return TRUE;
}

void
image_render_init( Image *image,
		   gint width,
		   gint height,
		   ImageColorType type) {
    RealPreview *preview;

    g_return_if_fail( image != NULL);
    g_return_if_fail( width > 0 && height > 0);
  
    preview = PREVIEW( image);

    image_calc_preview_size( image);
    if ( !preview->width || !preview->height) {
	preview->render_func = (ImageRenderFunc) image_render_skip_rows;
	return;
    }

    preview->flags = image->flags;
    preview->color_type = type;
    preview->data->src_width = width;
    preview->data->src_height = height;
    preview->data->x_scale = (float) width / preview->width;
    preview->data->y_scale = (float) height / preview->height;
    preview->data->y_pos = 0.0;
    preview->data->percentage = 0.0;
    preview->data->increment  = 1.0 / preview->height;
    
    /* allocate the row buffer */
    if ( preview->flags & IMAGE_HAS_COLORS) {
	preview->data->row_buff_size = MAX( ROW_BUFF_OPTIMAL_SIZE / (preview->width * 3),
					    1); 
	preview->data->row_buff = g_malloc( preview->data->row_buff_size * preview->width * 3);
	
	if ( preview->data->flags & IMAGE_CACHE) 
	    preview->image_cache = g_malloc( image->width * image->height * 3);
	
    } else {
	preview->data->row_buff_size = MAX( ROW_BUFF_OPTIMAL_SIZE / preview->width,
					    1);
	preview->data->row_buff = g_malloc( preview->data->row_buff_size * preview->width);
	
	if ( preview->data->flags & IMAGE_CACHE) 
	    preview->image_cache = g_malloc( image->width * image->height);
    }
	
    /* create the pixmap if needed */
    if ( preview->data->flags & IMAGE_RENDER_PIXMAP)
	preview->pixmap = gdk_pixmap_new( preview->data->window,
					  preview->width,
					  preview->height,
					  -1);
    
   
    if ( width == preview->width && height == preview->height)
	preview->render_func = preview->flags & IMAGE_HAS_COLORS ?
	    image_append_rows_fullsize_24 : image_append_rows_fullsize_8;
    else
	preview->render_func = preview->flags & IMAGE_HAS_COLORS ?
	    image_append_rows_24 : image_append_rows_8;

#ifdef DEBUG
    printf( "rendering image %s %dx%dx%d at %dx%dx%d\n\t"
	    "x_scale: %f\n\ty_scale: %f\n\trow_buff_size: %d\n\tcache: %d\n"
	    "\tbackground : %d %d %d\n",
	    image->file_name, image->width, image->height, type, 
	    preview->width, preview->height, preview->flags & IMAGE_HAS_COLORS ? 3 : 1,
	    preview->data->x_scale, preview->data->y_scale, preview->data->row_buff_size, 
	    preview->data->flags & IMAGE_CACHE,
	    preview->data->background.red, preview->data->background.green, preview->data->background.blue);
    fflush( stdout);
#endif
    
    
}


static gint
image_append_rows_fullsize_24( Image *image,
			       gint num_row,
			       guchar *data) {
    RealPreview *preview = PREVIEW( image);
    guchar  *dest;

    if ( preview->data->flags & IMAGE_CACHE) {
	if ( preview->color_type == IMAGE_COLOR_24_ALPHA)
	    apply_alpha_24( data, preview->image_cache + preview->data->appended_rows * preview->width * 3,
			    &preview->data->background, num_row * preview->width);
	else  
	    memcpy( preview->image_cache + preview->data->appended_rows * preview->width * 3, 
		    data, num_row * preview->width * 3);
    }

    dest = preview->data->row_buff + preview->data->row_buff_rows * preview->width * 3;

    while ( num_row--) {
	if ( preview->color_type == IMAGE_COLOR_24_ALPHA)
	    apply_alpha_24( data, dest, &preview->data->background, preview->width);
	else 
	    memcpy( dest, data, preview->width * 3);

	dest += preview->width * 3;
	data += preview->width * preview->color_type;
	
	preview->data->appended_rows ++;
	if ( !row_appended( image, preview, &dest))
	    return FALSE;
    }

    return TRUE;
}

static gint
image_append_rows_fullsize_8( Image *image,
			      gint num_row,
			      guchar *data) {
    RealPreview *preview = PREVIEW( image);
    guchar  *dest;
 
    if ( preview->data->flags & IMAGE_CACHE) {
	if ( preview->color_type == IMAGE_COLOR_8_ALPHA) 
	    apply_alpha_8( data, preview->image_cache + preview->data->appended_rows * preview->width,
			 preview->data->bg8, num_row * preview->width);
	else
	    memcpy( preview->image_cache + preview->data->appended_rows * preview->width, 
		    data, num_row * preview->width);
    }

    dest = preview->data->row_buff + preview->data->row_buff_rows * preview->width;

    while ( num_row--) {
	if ( preview->color_type == IMAGE_COLOR_8_ALPHA)
	    apply_alpha_8( data, dest, preview->data->bg8, preview->width);
	else
	    memcpy( dest, data, preview->width);

	dest += preview->width;
	data += preview->width * preview->color_type;
	
	preview->data->appended_rows ++;
	if ( !row_appended( image, preview, &dest))
	    return FALSE;
    }

    return TRUE;
}

static gint
image_append_rows_24( Image *image,
		      gint num_row,
		      guchar *data) {
    RealPreview *preview = PREVIEW( image);
    guchar  *dest, *src;
    gint w;
    guchar *row;
    float x_pos;

    if ( preview->data->flags & IMAGE_CACHE) {
	if ( preview->color_type == IMAGE_COLOR_24_ALPHA) 
	    apply_alpha_24( data, preview->image_cache + preview->data->appended_rows * image->width * 3,
			 &preview->data->background, num_row * image->width);
	else
	    memcpy( preview->image_cache + preview->data->appended_rows * image->width * 3, 
		    data, num_row * image->width * 3);
    }

    dest = preview->data->row_buff + preview->data->row_buff_rows * preview->width * 3;
    
    while ( num_row--) {
	
	while ( preview->data->appended_rows >= (int) preview->data->y_pos) {
	    
	    src = row = data;   
	    w = preview->width;
	    x_pos = 0.0;
	    
	    if ( preview->color_type == IMAGE_COLOR_24_ALPHA) 
		while ( w--) {
		    
		    dest[0] = APPLY_ALPHA( src[0], preview->data->background.red, src[3]);
		    dest[1] = APPLY_ALPHA( src[1], preview->data->background.green, src[3]);
		    dest[2] = APPLY_ALPHA( src[2], preview->data->background.blue, src[3]);

		    dest += 3;
		    
		    src = row + (int) x_pos * 4;
		    x_pos += preview->data->x_scale;
		}
	    else 
		while ( w--) {
		    dest[0] = src[0];
		    dest[1] = src[1];
		    dest[2] = src[2];
		    dest += 3;
		    
		    src = row + (int) x_pos * 3;
		    x_pos += preview->data->x_scale;
		}
	    /* One row has been scaled */
	    
	    preview->data->y_pos += preview->data->y_scale;
	    if ( !row_appended( image, preview, &dest))
		return FALSE;
	}
	preview->data->appended_rows ++;
	data += preview->data->src_width * preview->color_type;
	
    }

    return TRUE;
}

static gint
image_append_rows_8(  Image *image,
		      gint num_row,
		      guchar *data) {
    RealPreview *preview = PREVIEW( image);
    guchar  *dest, *src;
    gint w;
    guchar *row;
    float x_pos;
    
    if ( preview->data->flags & IMAGE_CACHE) {
	if ( preview->color_type == IMAGE_COLOR_8_ALPHA) 
	    apply_alpha_8( data, preview->image_cache + preview->data->appended_rows * image->width, 
			   preview->data->bg8, num_row * image->width);
	else
	    memcpy( preview->image_cache + preview->data->appended_rows * image->width, 
		    data, num_row * image->width);
    }

    dest = preview->data->row_buff + preview->data->row_buff_rows * preview->width;
    
    while ( num_row--) {
	
	while ( preview->data->appended_rows >= (int) preview->data->y_pos) {
	    
	    src = row = data;   
	    w = preview->width;
	    x_pos = 0.0;
	    
	    if ( preview->color_type == IMAGE_COLOR_8_ALPHA)
		while ( w--) {
		    /* xxx what color should i use for the background ? */
		    *dest++ = APPLY_ALPHA( src[0], preview->data->bg8, src[1]);	
		    
		    x_pos += preview->data->x_scale;
		    src = row + (int) x_pos * 2;
		}
	    else
		while ( w--) {
		    *dest++ = src[0];	
		    
		    x_pos += preview->data->x_scale;
		    src = row + (int) x_pos;
		}
	    
	    preview->data->y_pos += preview->data->y_scale;
	    if ( !row_appended( image, preview, &dest))
		return FALSE;
	}
	preview->data->appended_rows ++;
	data += preview->data->src_width * preview->color_type;
    }

    return TRUE;
}
    
static void
image_flush_buff( Image *image) {
    RealPreview *preview = PREVIEW( image);

    if ( !preview->data->row_buff_rows)
	return;


    if ( preview->pixmap) {
	if ( preview->flags & IMAGE_HAS_COLORS)
	    gdk_draw_rgb_image( preview->pixmap,
				preview->data->gc,
				0, preview->data->next_row_to_draw,
				preview->width, 
				preview->data->row_buff_rows,
				GDK_RGB_DITHER_NORMAL,
				preview->data->row_buff,
				preview->width * 3);
	
	else
	    gdk_draw_gray_image( preview->pixmap,
				 preview->data->gc,
				 0, preview->data->next_row_to_draw,
				 preview->width, 
				 preview->data->row_buff_rows,
				 GDK_RGB_DITHER_NORMAL,
				 preview->data->row_buff,
				 preview->width);

	if ( preview->data->flags & IMAGE_RENDER_IN_WINDOW)
	    gdk_draw_pixmap( preview->data->window,
			     preview->data->gc,
			     preview->pixmap,
			     0, preview->data->next_row_to_draw,
			     preview->data->x_dest, 
			     preview->data->y_dest + preview->data->next_row_to_draw,
			     preview->width, 
			     preview->data->row_buff_rows);
       
    } else if ( preview->data->flags & IMAGE_RENDER_IN_WINDOW) {
	if ( preview->flags & IMAGE_HAS_COLORS)
	    gdk_draw_rgb_image( preview->data->window,
				preview->data->gc,
				preview->data->x_dest, 
				preview->data->y_dest + preview->data->next_row_to_draw,
				preview->width, 
				preview->data->row_buff_rows,
				GDK_RGB_DITHER_NORMAL,
				preview->data->row_buff,
				preview->width * 3);
	
	else
	    gdk_draw_gray_image( preview->data->window,
				 preview->data->gc,
				 preview->data->x_dest, 
				 preview->data->y_dest + preview->data->next_row_to_draw,
				 preview->width, 
				 preview->data->row_buff_rows,
				 GDK_RGB_DITHER_NORMAL,
				 preview->data->row_buff,
				 preview->width);
    } 
    gdk_flush();

    preview->data->next_row_to_draw += preview->data->row_buff_rows;
    preview->data->row_buff_rows = 0;

}


static ImageFormat*
get_image_format_from_extension( Image *image) {
    ImageFormat *format;
    GSList *list;
    char *extension, **exts;

    g_return_val_if_fail( image != NULL, NULL);

    extension = strrchr( image->file_name, '.');
    if ( !extension)
	return NULL;

    extension++;

    list = images_formats;
    while ( list) {
	format = list->data;
	list = list->next;
	
	exts = format->extensions;
	while ( *exts) {
	    if ( !g_strcasecmp( extension, *exts))
		return format;

	    exts++;
	}
    }

    return NULL;
}

static ImageFormat*
get_image_format( Image *image) {
    ImageFormat *format;
    GSList *list;

    g_return_val_if_fail( image != NULL, NULL);
    g_return_val_if_fail( image->preview && image->preview->file, NULL);

    list = images_formats;
    while ( list) {
	format = list->data;
	list = list->next;

	if ( format->check && (*format->check)( image))
	    return format;
    }

    return NULL;
}
    
gint             
image_extension_supported( char *file_name) {
    Image image = { 0};
    
    image.file_name = file_name;
    
    return get_image_format_from_extension( &image) != NULL;
}

gint             
image_supported( char *path, char *file_name) {
    Image image = { 0};
    
    image.file_name = file_name;
    image.path = path;

    return get_image_format( &image) != NULL;
}


static void
stat_image( Image *image, char *image_path) {
    struct stat stat_buff;

    g_return_if_fail( image != NULL);
    g_return_if_fail( image_path != NULL);

    if ( !stat( image_path, &stat_buff)) {
	image->file_size = stat_buff.st_size;
    }

}


