/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __IMAGE_H__
#define __IMAGE_H__

#include <stdio.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>

typedef struct _Image Image;
typedef struct _Preview Preview;
typedef struct _ImageFormat ImageFormat;

typedef gint (*ImageRenderFunc) ( Image *image, gint num_row, guchar *data);

/* 
   return value : zero => stop rendering the image
                  non zero => continue
		  */
typedef gint (*ImageProgressFunc) ( Image *image, gfloat percentage, gpointer data);
typedef gint (*ImageFormatFunc) ( Image *image);

enum {
    IMAGE_RENDER_PIXMAP       = 1 << 2,
    IMAGE_RENDER_IN_WINDOW    = 1 << 3,
    IMAGE_CACHE               = 1 << 4,
    IMAGE_FILL_BOX            = 1 << 5,
    IMAGE_MAGNIFY_IF_SMALLER  = 1 << 6
};

typedef enum {
    IMAGE_COLOR_8        = 1,
    IMAGE_COLOR_8_ALPHA  = 2,
    IMAGE_COLOR_24       = 3,
    IMAGE_COLOR_24_ALPHA = 4
} ImageColorType;

enum {
    IMAGE_HAS_COLORS = 1 << 1,
    IMAGE_HAS_ALPHA  = 1 << 2,
};

typedef enum {
    JPEG_IMAGE,
    PNG_IMAGE,
    TIFF_IMAGE,
    XPM_IMAGE,
    GIF_IMAGE
} ImageType;

struct _Preview {
    FILE *file;

    gint width;
    gint height;
    guint flags;

    ImageRenderFunc render_func;
};

struct _Image {
    gchar *file_name;
    gchar *path;
    guint32 file_size;

    ImageType type;
    gint width;
    gint height;
    guint flags;

    Preview *preview;
};

struct _ImageFormat {
    char **extensions; /* NULL terminated array of possibles extenssions */
    
    ImageFormatFunc check;
    ImageFormatFunc read_header;
    ImageFormatFunc read;
    ImageFormatFunc write;
};

Image*           image_new( char *file_name, char *path);
Image*           image_copy( Image *image);
void             image_free( Image *image);

/* reads the image's width height & depth. */
gint             image_read_header( Image *image);
/* free the "preview" field of an "Image". Its should be called when
   after image_render */
void             image_free_preview( Image *image);
/* unref the image's pixmap & mask. Only significant when image_render is called
   with the IMAGE_RENDER_PIXMAP flag */
void             image_free_pixmaps( Image *image);
/* after a call to image render with the IMAGE_RENDER_PIXMAP flag,
   it returns the created pixmap & mask */ 
GdkPixmap*       image_get_pixmap( Image *image);
GdkBitmap*       image_get_mask( Image *image);

gint             image_extension_supported( char *file_name);
gint             image_supported( char *path, char *file_name);

void             image_get_preview_size( Image *image, 
					 gint box_width, 
					 gint box_height, 
					 gint flags, 
					 gint *width, gint *height);

/* render "image" according to the ginven flags :
        IMAGE_RENDER_PIXMAP : render "image" in a GdkPixmap
	               'window' : GdkWindow used in the call to gdk_pixmap_new
		       'gc'     : GC used in the call to gdk_rgb_draw
		       'width & height' : output width & height of the pixmap & mask
	               (the field x and y are ignored)
		       'func' : callback used for progress monitoring. 
		       'func_data' : callback's data
	IMAGE_RENDER_IN_WINDOW : render "image" in the window "window"
	               'window' : where to draw the image
		       'gc'     : GC used in the call to gdk_rgb_draw
		       'x & y' : where to draw
		       'width & height' : output width & height of the image
		       'func' : callback used for progress monitoring. 
		       'func_data' : callback's data
  'background' is used for alpha blending
  These flags can be combined by using the "|" operator ( eg IMAGE_RENDER_PIXMAP | IMAGE_RENDER_IN_WINDOW
  Additionaly the IMAGE_CACHE can be used when yout want to render the same image
    at different sizes. What it does is to read the image and cache the read data so that
    next calls to image render will read the cached data 
    */
gint             image_render( Image *image,
			       GdkWindow *window,
			       GdkGC *gc,
			       GdkColor *background,
			       gint x,
			       gint y,
			       gint width,
			       gint height,
			       gint flags,
			       ImageProgressFunc  func,
			       gpointer           func_data);

/* 
   Used by image decompressor to append rows to an image. 
*/

void             image_set_infos( Image *image, ImageType type,
				  gint width, gint height, gint has_colors,
				  gint has_alpha);
void             image_format_register( ImageFormat *image_format);

void             image_render_init( Image *image, gint width, gint height, ImageColorType type);

gint             image_get_render_width( Image *image);
gint             image_get_render_height( Image *image);

#define          image_append_rows( image, num_row, data) \
                 ((*(image)->preview->render_func) ( image, num_row, data))

				     

#endif /* __IMAGE_H__ */
