/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>

#include "ubox.h"
#include "imageubox.h"

#define FONT_XPAD 1
#define FONT_YPAD 1
#define PREVIEW1_XPAD 4
#define PREVIEW1_YPAD 4
#define PREVIEW2_XPAD 1
#define PREVIEW2_YPAD 1

typedef struct _ImageUBox {
    GtkWidget *ubox;
    ImageUBoxType type;
    gint preview_width;
    gint preview_height;
    gint cell_width;
    gint cell_height;
    GtkSignalFunc draw_cell;

} ImageUBox;

static char           *image_ubox_key = "image_ubox";
static gint           p_image_name_width  = 150;
static gint           p_image_dimension_width  = 100;
static gint           p_image_file_size_width  = 100;
static gint           p_image_path_width = 300;
   


static void
image_ubox_draw_preview1( UBox *ubox, 
			  UBoxCell *cell, 
			  GdkRectangle *cell_area, 
			  GdkRectangle *area,
			  ImageUBox *box);
static void
image_ubox_draw_preview2( UBox *ubox, 
			  UBoxCell *cell, 
			  GdkRectangle *cell_area, 
			  GdkRectangle *area,
			  ImageUBox *box);
static void
image_ubox_draw_preview3( UBox *ubox, 
			  UBoxCell *cell, 
			  GdkRectangle *cell_area, 
			  GdkRectangle *area,
			  ImageUBox *box);
static void
image_ubox_compute_cell_dim( ImageUBox *box);
static void
image_ubox_destroy_cb( GtkWidget *w, ImageUBox *box);
static void
image_ubox_set_style( GtkWidget *ubox, GtkStyle *previous, ImageUBox *box);

GtkWidget*
image_ubox_new( gint preview_width,
		gint preview_height) {
    ImageUBox *box;

    box = g_new( ImageUBox, 1);
    box->type = IMAGE_UBOX_PREVIEW1;
    box->ubox = NULL;
    box->preview_width = preview_width;
    box->preview_height = preview_height;
    box->draw_cell = image_ubox_draw_preview1;
    image_ubox_compute_cell_dim( box);

    box->ubox = ubox_new( box->cell_width, box->cell_height);
    gtk_object_set_data( GTK_OBJECT( box->ubox), image_ubox_key, box);

    gtk_signal_connect( GTK_OBJECT( box->ubox), "destroy",
			(GtkSignalFunc) image_ubox_destroy_cb, NULL);
    gtk_signal_connect( GTK_OBJECT( box->ubox), "style_set",
			(GtkSignalFunc) image_ubox_set_style, box);
    gtk_signal_connect( GTK_OBJECT( box->ubox), "draw_cell",
			box->draw_cell, box);
    
    return box->ubox;
}

static void
image_ubox_destroy_cb( GtkWidget *w, ImageUBox *box) {
    g_free( box);
}

static void
image_ubox_set_style( GtkWidget *ubox, GtkStyle *previous, ImageUBox *box) {
    /* will compute new cell dimension and resize the ubox */
    image_ubox_set_type( UBOX( ubox), box->type);
}

UBoxCell*
image_ubox_append( UBox *ubox,
		   Image *image,
		   ImageProgressFunc func,
		   gpointer data) {
    ImageUBox *box;
    g_return_val_if_fail( ubox != NULL, NULL);
    g_return_val_if_fail( IS_UBOX( ubox), NULL);
    g_return_val_if_fail( image != NULL, NULL);
    
    box = gtk_object_get_data( GTK_OBJECT( ubox), image_ubox_key);
    g_return_val_if_fail( box != NULL, NULL);

    switch( box->type) {
    case IMAGE_UBOX_PREVIEW1:
	image_render( image,
		      ubox->cells_area,
		      GTK_WIDGET( ubox)->style->black_gc,
		      &GTK_WIDGET(ubox)->style->bg[ GTK_STATE_NORMAL],
		      0, 0,
		      box->preview_width,
		      box->preview_height,
		      IMAGE_RENDER_PIXMAP, func, data);
	return ubox_append( ubox, image);
	break;
    case IMAGE_UBOX_PREVIEW2:
    case IMAGE_UBOX_PREVIEW3:
	image_read_header( image);
	if ( func) 
	    (*func) ( image, 1.0, data);
	return ubox_append( ubox, image);
	break;
    default:
    }

    return NULL;
}

void
image_ubox_clear( UBox *ubox, gint free_preview) {
    Image *data;
    GSList *item;
   
    item = ubox->children;
    while ( item) {
	data = ((UBoxCell*) item->data)->data;
	if ( free_preview)
	    image_free_preview( data);
	item = item->next;
    }
    
    ubox_clear( ubox);
}
    
Image*           
image_ubox_get_image( UBox *ubox,
		      gint cell_no) {
    UBoxCell *cell;
    GSList *item;
    
    if ( cell_no < 0)
	return NULL;
    item = g_slist_nth( ubox->children, cell_no);
    if ( !item)
	return NULL;

    cell = item->data;

    return cell->data;
}

static void
image_ubox_compute_cell_dim( ImageUBox *box) {
    GtkStyle *style;
    gint font_height;

    if ( box->ubox) 
	style = GTK_WIDGET( box->ubox)->style;
    else
	style = gtk_widget_get_default_style();

    font_height = style->font->ascent + style->font->descent;

    switch( box->type) {
    case IMAGE_UBOX_PREVIEW1:
	box->cell_width = box->preview_width + 2*FONT_XPAD + 2*PREVIEW1_XPAD;
	box->cell_height = box->preview_height + 2*font_height + 3*FONT_YPAD + 2*PREVIEW1_YPAD;
	break;
    case IMAGE_UBOX_PREVIEW2:
	box->cell_width = p_image_name_width + 2*FONT_XPAD + 2*PREVIEW2_XPAD;
	box->cell_height = font_height + 2*FONT_YPAD + 2*PREVIEW2_YPAD;
	break;
    case IMAGE_UBOX_PREVIEW3:
	box->cell_width = p_image_name_width + 
	    p_image_dimension_width +
	    p_image_file_size_width +
	    p_image_path_width +
	    5*FONT_XPAD + 5*PREVIEW2_XPAD;
	box->cell_height = font_height + 2*FONT_YPAD + 2*PREVIEW2_YPAD;
	break;
    default:
    }
}


void
image_ubox_set_type( UBox *ubox, ImageUBoxType type) {
    ImageUBox *box;
  
    box = gtk_object_get_data( GTK_OBJECT( ubox), image_ubox_key);
    g_return_if_fail( box != NULL);

    if ( box->draw_cell)
	gtk_signal_disconnect_by_func( GTK_OBJECT( ubox), box->draw_cell, box);

    box->type = type;
    switch( type) {
    case IMAGE_UBOX_PREVIEW1:
	box->draw_cell = (GtkSignalFunc) image_ubox_draw_preview1;
	break;
    case IMAGE_UBOX_PREVIEW2:
	box->draw_cell = (GtkSignalFunc) image_ubox_draw_preview2;
	break;
    case IMAGE_UBOX_PREVIEW3:
	box->draw_cell = (GtkSignalFunc) image_ubox_draw_preview3;
	break;
    default:
    }
    
    image_ubox_compute_cell_dim( box);
    ubox_resize( UBOX( box->ubox), box->cell_width, box->cell_height);

    gtk_signal_connect( GTK_OBJECT( ubox), "draw_cell",
			box->draw_cell, box);
    
}


static void
image_ubox_draw_preview1( UBox *ubox, 
			  UBoxCell *cell, 
			  GdkRectangle *cell_area, 
			  GdkRectangle *area,
			  ImageUBox *box) {
    Image *image;
    GdkRectangle text_area;
    GdkRectangle intersect_area;
    GdkGC *fg_gc, *bg_gc;
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    gint pixmap_width, pixmap_height, font_height;
    gint xdest, ydest, xsrc, ysrc;
    char buffer[32];

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( cell != NULL);
    g_return_if_fail( cell_area != NULL);
    g_return_if_fail( area != NULL);

    image = cell->data;
    g_return_if_fail( image != NULL);
 
    fg_gc = GTK_WIDGET( ubox)->style->fg_gc[ cell->state];
    bg_gc = GTK_WIDGET( ubox)->style->bg_gc[ cell->state];

    gtk_draw_shadow( GTK_WIDGET( ubox)->style,
		     ubox->cells_area,
		     GTK_STATE_NORMAL,
		     cell->state == GTK_STATE_NORMAL ? GTK_SHADOW_OUT : GTK_SHADOW_IN,
		     cell_area->x + 2,
		     cell_area->y + 2,
		     cell_area->width - 4,
		     cell_area->height - 4);

    cell_area->x += PREVIEW1_XPAD;
    cell_area->y += PREVIEW1_YPAD;
    cell_area->width  -= 2*PREVIEW1_XPAD;
    cell_area->height -= 2*PREVIEW1_YPAD;
    
    if ( gdk_rectangle_intersect( cell_area, area, &intersect_area)) 
		gdk_window_clear_area( ubox->cells_area,
			       intersect_area.x,
			       intersect_area.y,
			       intersect_area.width,
			       intersect_area.height);

    /*
		gdk_draw_rectangle( ubox->cells_area,
			    GTK_WIDGET( ubox)->style->black_gc,
			    TRUE,
			    intersect_area.x,
			    intersect_area.y,
			    intersect_area.width,
			    intersect_area.height);
			       */

    font_height = GTK_WIDGET( ubox)->style->font->ascent +
	GTK_WIDGET( ubox)->style->font->descent;

    text_area.x = cell_area->x + FONT_XPAD;
    text_area.width  = cell_area->width - 2*FONT_XPAD;
    text_area.height = 2*font_height + 3*FONT_YPAD;
    text_area.y      = cell_area->y + (cell_area->height - text_area.height);

    pixmap = image_get_pixmap( image);
    mask = image_get_mask( image);
    if ( pixmap) {

	gdk_window_get_size( pixmap, &pixmap_width, &pixmap_height);
	
	xsrc = 0;
	ysrc = 0;
	xdest = cell_area->x + (cell_area->width - pixmap_width) / 2;
	ydest = cell_area->y + (cell_area->height - text_area.height - pixmap_height) / 2;
	
	if ( xdest < area->x) {
	    xsrc = area->x - xdest;
	    pixmap_width -= xsrc;
	    xdest = area->x;
	}
	
	if ( xdest +  pixmap_width > area->x + area->width)
	    pixmap_width = (area->x + area->width) - xdest;
	
	if ( ydest < area->y) {
	    ysrc = area->y - ydest;
	    pixmap_height -= ysrc;
	    ydest = area->y;
	}
	
	if ( ydest + pixmap_height > area->y + area->height)
	    pixmap_height = (area->y + area->height) - ydest;
	
	
	if ( pixmap_width > 0 && pixmap_height > 0) {
	    if ( mask) {
		gdk_gc_set_clip_mask( GTK_WIDGET( ubox)->style->black_gc, mask);
		gdk_gc_set_clip_origin( GTK_WIDGET( ubox)->style->black_gc, cell_area->x, cell_area->y);
	    }
	
	    
	    gdk_draw_pixmap( ubox->cells_area,
			     GTK_WIDGET( ubox)->style->white_gc,
			     pixmap,
			     xsrc,
			     ysrc,
			     xdest, 
			     ydest,
			     pixmap_width, 
			     pixmap_height);
	    
	    
	    if ( mask) {
		gdk_gc_set_clip_mask( GTK_WIDGET( ubox)->style->black_gc, NULL);
		gdk_gc_set_clip_origin( GTK_WIDGET( ubox)->style->black_gc, 0, 0);
	    }
	}
    }
    /* draw image info strings */

    if ( gdk_rectangle_intersect( area, &text_area, &intersect_area)) {
	
	gdk_draw_rectangle( ubox->cells_area,
			    bg_gc,
			    TRUE,
			    intersect_area.x,
			    intersect_area.y,
			    intersect_area.width,
			    intersect_area.height);

	gdk_gc_set_clip_rectangle( fg_gc, &intersect_area);

	xdest = text_area.x + (text_area.width - gdk_string_width( GTK_WIDGET( ubox)->style->font, image->file_name))/2;
	ydest = text_area.y + font_height + FONT_YPAD;

	gdk_draw_string( ubox->cells_area,
			 GTK_WIDGET( ubox)->style->font,
			 fg_gc,
			 xdest,
			 ydest,
			 image->file_name);
	
	sprintf( buffer, "%dx%dx%d", 
		 image->width,
		 image->height,
		 image->flags & IMAGE_HAS_COLORS ? 24 : 8);
	ydest += font_height + FONT_YPAD;
	gdk_draw_string( ubox->cells_area,
			 GTK_WIDGET( ubox)->style->font,
			 fg_gc,
			 text_area.x, 
			 ydest,
			 buffer);
	
	sprintf( buffer, "%d Kb", image->file_size / 1024);
	xdest = text_area.x + text_area.width - gdk_string_width( GTK_WIDGET( ubox)->style->font, buffer);

	gdk_draw_string( ubox->cells_area,
			 GTK_WIDGET( ubox)->style->font,
			 fg_gc,
			 xdest, 
			 ydest,
			 buffer);
	
	gdk_gc_set_clip_rectangle (fg_gc, NULL);
    }
    
}

static gint 
get_max_chars( GdkFont *font, char *s, gint width) {
    gint n;

    n = strlen( s);
    while ( gdk_text_width( font, s, n) > width)
	n--;

    return n;
}

static void
image_ubox_draw_preview2( UBox *ubox, 
			  UBoxCell *cell, 
			  GdkRectangle *cell_area, 
			  GdkRectangle *area,
			  ImageUBox *box) {
    Image *image;
    GdkGC *fg_gc, *bg_gc;
    gint xdest, ydest;
    gint l;

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( cell != NULL);
    g_return_if_fail( cell_area != NULL);
    g_return_if_fail( area != NULL);

    image = cell->data;

    fg_gc = GTK_WIDGET( ubox)->style->fg_gc[ cell->state];
    bg_gc = GTK_WIDGET( ubox)->style->bg_gc[ cell->state];
    
    gdk_draw_rectangle( ubox->cells_area,
			bg_gc,
			TRUE,
			area->x,
			area->y,
			area->width,
			area->height);
    
    gdk_draw_rectangle( ubox->cells_area,
			GTK_WIDGET( ubox)->style->bg_gc[ GTK_STATE_PRELIGHT],
			FALSE,
			cell_area->x,
			cell_area->y,
			cell_area->width,
			cell_area->height);
	
    cell_area->x += PREVIEW2_XPAD;
    cell_area->y += PREVIEW2_YPAD;
    cell_area->width -= 2*PREVIEW2_XPAD;
    cell_area->height -= 2*PREVIEW2_YPAD;
    
    xdest = cell_area->x + FONT_XPAD;
    ydest = cell_area->y + GTK_WIDGET( ubox)->style->font->ascent + FONT_YPAD;

    l = get_max_chars( GTK_WIDGET( ubox)->style->font, image->file_name,
		       p_image_name_width);
    gdk_draw_text( ubox->cells_area,
		   GTK_WIDGET( ubox)->style->font,
		   GTK_WIDGET( ubox)->style->fg_gc[ cell->state],
		   xdest, 
		   ydest,
		   image->file_name,
		   l);

}


static void
image_ubox_draw_preview3( UBox *ubox, 
			  UBoxCell *cell, 
			  GdkRectangle *cell_area, 
			  GdkRectangle *area,
			  ImageUBox *box) {
    Image *image;
    GdkGC *fg_gc, *bg_gc;
    gint xdest, ydest;
    gint l;
    char buff[ 64];
    
    g_return_if_fail( ubox != NULL);
    g_return_if_fail( cell != NULL);
    g_return_if_fail( cell_area != NULL);
    g_return_if_fail( area != NULL);

    image = cell->data;

    fg_gc = GTK_WIDGET( ubox)->style->fg_gc[ cell->state];
    bg_gc = GTK_WIDGET( ubox)->style->bg_gc[ cell->state];
    
    gdk_draw_rectangle( ubox->cells_area,
			bg_gc,
			TRUE,
			area->x,
			area->y,
			area->width,
			area->height);

    
    gdk_draw_rectangle( ubox->cells_area,
			GTK_WIDGET( ubox)->style->bg_gc[ GTK_STATE_PRELIGHT],
			FALSE,
			cell_area->x,
			cell_area->y,
			cell_area->width,
			cell_area->height);
		   
    cell_area->x += PREVIEW2_XPAD;
    cell_area->y += PREVIEW2_YPAD;
    cell_area->width -= 2*PREVIEW2_XPAD;
    cell_area->height -= 2*PREVIEW2_YPAD;
    
    xdest = cell_area->x + FONT_XPAD;
    ydest = cell_area->y + GTK_WIDGET( ubox)->style->font->ascent + FONT_YPAD;
  
    l = get_max_chars( GTK_WIDGET( ubox)->style->font, image->file_name,
		       p_image_name_width);
    gdk_draw_text( ubox->cells_area,
		   GTK_WIDGET( ubox)->style->font,
		   fg_gc,
		   xdest, 
		   ydest,
		   image->file_name,
		   l);

    xdest += p_image_name_width;
#if 0
    gdk_draw_line( ubox->cells_area,
		   GTK_WIDGET( ubox)->style->bg_gc[ GTK_STATE_PRELIGHT],
		   xdest,
		   cell_area->y,
		   xdest,
		   cell_area->y +  cell_area->height);
#endif
    xdest += FONT_XPAD + PREVIEW2_XPAD;

    sprintf( buff, "%dx%dx%d", image->width, image->height,
	     image->flags & IMAGE_HAS_COLORS ? 24 : 8);
    l = get_max_chars( GTK_WIDGET( ubox)->style->font, buff,
		       p_image_dimension_width);
    gdk_draw_text( ubox->cells_area,
		   GTK_WIDGET( ubox)->style->font,
		   fg_gc,
		   xdest, 
		   ydest,
		   buff,
		   l);

     xdest += p_image_dimension_width;
#if 0
     gdk_draw_line( ubox->cells_area,
		   GTK_WIDGET( ubox)->style->bg_gc[ GTK_STATE_PRELIGHT],
		   xdest,
		   cell_area->y,
		   xdest,
		   cell_area->y +  cell_area->height);
#endif
     xdest += FONT_XPAD + PREVIEW2_XPAD;

     sprintf( buff, "%d", image->file_size);
     l = get_max_chars( GTK_WIDGET( ubox)->style->font, buff,
			p_image_file_size_width);
     gdk_draw_text( ubox->cells_area,
		    GTK_WIDGET( ubox)->style->font,
		    fg_gc,
		    xdest, 
		    ydest,
		    buff,
		    l);

     xdest += p_image_file_size_width;
#if 0
     gdk_draw_line( ubox->cells_area,
		   GTK_WIDGET( ubox)->style->bg_gc[ GTK_STATE_PRELIGHT],
		   xdest,
		   cell_area->y,
		   xdest,
		   cell_area->y +  cell_area->height);
#endif
     xdest += FONT_XPAD + PREVIEW2_XPAD;

     l = get_max_chars( GTK_WIDGET( ubox)->style->font, image->path,
			p_image_path_width);
     gdk_draw_text( ubox->cells_area,
		    GTK_WIDGET( ubox)->style->font,
		    fg_gc,
		    xdest, 
		    ydest,
		    image->path,
		    l);

}
