/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <string.h>
#include <glib.h>

#include "ldb.h"
#include "ldbutils.h"
#include "imdb.h"
#include "gtkcompat.h"

//#define DEBUG
//#define DEBUG_ITEM

static void      imdb_init_methods( ImDB *imdb);
static void      imdb_load_dirtree( ImDB *imdb);
static void      imdb_load_properties( ImDB *imdb);
static void      imdb_load_images( ImDB *imdb);

static void      imdb_free_item( ImDB *imdb, guint16 type, void *item);

static gint      read_image ( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data);
static gint      write_image( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data);
static gint      dump_image ( FILE *dump_file, ImDBImage *dbim);
static guint     image_size ( ImDB *imdb, ImDBImage *dbim);

static gint      read_property ( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data);
static gint      write_property( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data);
static gint      dump_property ( FILE *dump_file, ImDBProperty *property);
static guint     property_size ( ImDB *imdb, ImDBProperty *property);

static gint      read_directory ( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data);
static gint      write_directory( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data);
static gint      dump_directory ( FILE *dump_file, ImDBDirectory *directory);
static guint     directory_size ( ImDB *imdb, ImDBDirectory *dir);

static char*     make_path( ImDB *imdb, ImDBDirectory *node);


#define ImDBBlockSize 2048

static void
imdb_init_methods( ImDB *imdb) {
    
    ldb_set_read ( imdb->ldb, ImDBImageType, read_image, imdb);
    ldb_set_write( imdb->ldb, ImDBImageType, write_image, imdb);
 
    ldb_set_read ( imdb->ldb, ImDBDirectoryType, read_directory, imdb);
    ldb_set_write( imdb->ldb, ImDBDirectoryType, write_directory, imdb);
 
    ldb_set_read ( imdb->ldb, ImDBPropertyType, read_property, imdb);
    ldb_set_write( imdb->ldb, ImDBPropertyType, write_property, imdb);
}


ImDB*       
imdb_load( char *path) {
    ImDB *imdb;
    LDB *ldb;
    gint fd;
    long pos;

    ldb = ldb_load( path, IMDB_HEADER_SIZE);
    if ( !ldb)
	return NULL;

    imdb = g_new( ImDB, 1);
    imdb->ldb = ldb;

    fd = ldb->fd;
    pos = ldb_get_user_header_position( ldb);
    if ( (lseek( fd, pos, SEEK_SET) == -1) ||
	 !read_uint32( fd, &imdb->header.image_count) ||
	 !read_uint16( fd, &imdb->header.property_count) ||
	 !read_uint16( fd, &imdb->header.directory_count) ||
	 !read_uint16( fd, &imdb->header.imdb_version) ||
	 !read_str( fd, imdb->header.imdb_marker, IMDB_MARKER_SIZE)) {
	close( fd);
	g_free( imdb);
	return NULL;
    }

    if ( memcmp( &imdb->header.imdb_marker, IMDB_MARKER_VALUE, IMDB_MARKER_SIZE)) {
	g_warning( "%s is not a jview image database", path);
	close( fd);
	g_free( imdb);
	return NULL;
    }

    if ( imdb->header.imdb_version > IMDB_VERSION)
	g_warning( "You should get a newer release of jview.");

    
#ifdef DEBUG
    printf( "*** loading database %s\n\timage_count: %d\n\tdirectory_count: %d\n\tproperty_count: %d\n",
	    path, imdb->header.image_count, imdb->header.directory_count, imdb->header.property_count);
    fflush( stdout);
#endif

    imdb->db_path = g_strdup( path);

    imdb_init_methods( imdb); 
    imdb_load_dirtree( imdb);
    imdb_load_properties( imdb);
    imdb_load_images( imdb);

    return imdb;
}

static void
imdb_load_dirtree( ImDB *imdb) {
    LDBCursor *cursor;
    LDBItem  item;
    ImDBDirectory *dir, node;
    gint i;

    imdb->dirtree = g_array_new_compat( ImDBDirectory);

    /* append node 0 == root */
    node.id = node.parent = node.child = node.brother = 0;
    node.name   = node.path   = NULL; node.flags = 0; node.images = NULL;
    imdb->dirtree = g_array_append_val_compat( imdb->dirtree, ImDBDirectory, node);

    i = imdb->header.directory_count - 1;
    while ( i-- > 0) 
	imdb->dirtree = g_array_append_val_compat( imdb->dirtree, ImDBDirectory, node);
    
    cursor = ldb_cursor_new( imdb->ldb, ImDBDirectoryType, LDB_READ);
    g_return_if_fail( cursor != NULL);
    while ( ldb_read_next_item( imdb->ldb, cursor, &item)) {
	dir = item.data;
	dir->brother = g_array_index( imdb->dirtree, ImDBDirectory, dir->parent).child;
	g_array_index( imdb->dirtree, ImDBDirectory, dir->parent).child = dir->id;
	dir->path = make_path( imdb, dir);
	g_array_index( imdb->dirtree, ImDBDirectory, dir->id) = *dir;

	dir->name = dir->path = NULL;
	imdb_free_item( imdb, ImDBDirectoryType, dir);
    }
    ldb_cursor_free( imdb->ldb, cursor);

}
	
static void
imdb_load_properties( ImDB *imdb) {
    LDBCursor *cursor;
    LDBItem item;

    imdb->properties = NULL;

    cursor = ldb_cursor_new( imdb->ldb, ImDBPropertyType, LDB_READ);
    if ( cursor) {
	while ( ldb_read_next_item( imdb->ldb, cursor, &item)) 
	    imdb->properties = g_list_prepend( imdb->properties, item.data);

	ldb_cursor_free( imdb->ldb, cursor);
    }
}
	
static void
imdb_load_images( ImDB *imdb) {
    LDBCursor *cursor;
    LDBItem item;
    ImDBImage *image;
    ImDBDirectory *dir;

    cursor = ldb_cursor_new( imdb->ldb, ImDBImageType, LDB_READ);
    if ( cursor) {
	while ( ldb_read_next_item( imdb->ldb, cursor, &item)) {
	    image = item.data;
	    dir = &g_array_index( imdb->dirtree, ImDBDirectory, image->directory);
	   
	    dir->images = g_list_prepend( dir->images, image);
	}

	ldb_cursor_free( imdb->ldb, cursor);
    }
}


ImDB*       
imdb_create( char *path) {
    ImDB *imdb;
    LDB *ldb;
    ImDBDirectory node;

    ldb = ldb_create( path, ImDBBlockSize, ImDBTypeCount, IMDB_HEADER_SIZE);
    if ( !ldb)
	return NULL;

    imdb = g_new( ImDB, 1);
    imdb->ldb = ldb;
    imdb->header.image_count = 0;
    imdb->header.directory_count = 1; /* the root dir is automticaly added */
    imdb->header.property_count = 0;
    imdb->header.imdb_version = IMDB_VERSION;
    memcpy( imdb->header.imdb_marker, IMDB_MARKER_VALUE, IMDB_MARKER_SIZE);
    imdb->properties = NULL;

    imdb->db_path = g_strdup( path);

    imdb_init_methods( imdb);
    imdb->dirtree = g_array_new_compat( ImDBDirectory);

    /* append node 0 == root */
    node.id = node.parent = node.child = node.brother = 0;
    node.name   = NULL; node.path   = NULL; node.flags = 0; node.images = NULL;
    imdb->dirtree = g_array_append_val_compat( imdb->dirtree, ImDBDirectory, node);

    return imdb;
}

void         
imdb_close( ImDB *imdb) {
    ImDBDirectory *dir;
    gint fd;
    GList *list;
    gint i;

    g_return_if_fail( imdb != NULL);

    fd =  imdb->ldb->fd;

    imdb_flush( imdb);

    ldb_close( imdb->ldb);

    g_free( imdb->db_path);

    list = imdb->properties;
    while ( list) {
	imdb_free_item( imdb, ImDBPropertyType, list->data);
	list = list->next;
    }
    if ( imdb->properties)
	g_list_free( imdb->properties);

    for ( i = 1; i < imdb->header.directory_count; i++) {
	dir = &g_array_index( imdb->dirtree, ImDBDirectory, i);
	if ( dir->name)
	    g_free( dir->name);
	if ( dir->path)
	    g_free( dir->path);

	/* free images in this directory */
	list = dir->images;
	while ( list) {
	    imdb_free_item( imdb, ImDBImageType, list->data);
	    list = list->next;
	}
	if ( dir->images)
	    g_list_free( dir->images);

    }
    g_array_free( imdb->dirtree, TRUE);

#ifdef DEBUG
    printf( "*** closing database \n\timage_count: %d\n\tdirectory_count: %d\n\tproperty_count: %d\n",
	    imdb->header.image_count, imdb->header.directory_count, imdb->header.property_count);
#endif
    g_free( imdb);
}
    
    
void        
imdb_dump( ImDB *imdb, char *dest_file) {

}


void         
imdb_flush( ImDB *imdb) {
    ImDBImage *image;
    ImDBProperty *property;
    ImDBDirectory *directory;
    LDBCursor *cursor;
    LDBItem item;
    GList *list;
    long pos;
    gint fd, i;

    g_return_if_fail( imdb != NULL);

    fd = imdb->ldb->fd;
    pos = ldb_get_user_header_position( imdb->ldb);
    if ( (lseek( fd, pos, SEEK_SET) == -1) || 
	 !write_uint32( fd, imdb->header.image_count) ||
	 !write_uint16( fd, imdb->header.property_count) ||
	 !write_uint16( fd, imdb->header.directory_count) ||
	 !write_uint16( fd, imdb->header.imdb_version) ||
	 !write_str( fd, imdb->header.imdb_marker, IMDB_MARKER_SIZE)) {
	g_warning( "imdb: can't write header");
	return;
    }

    /* write new & modifed images */
    cursor = ldb_cursor_new( imdb->ldb, ImDBImageType, LDB_WRITE);
    g_return_if_fail( cursor != NULL);
    for ( i = 0; i < imdb->header.directory_count; i++) {
	directory = &g_array_index( imdb->dirtree, ImDBDirectory, i);

	list = directory->images;
	while( list) {
	    image = list->data;
	    list = list->next;
	    
	    if ( image->flags & ImDBItemNew) {
		item.data = image;
		item.size = image_size( imdb, image);
		
		ldb_write_item( imdb->ldb, cursor, &item);
		
	    } else if ( image->flags & ImDBItemModified) {
		ldb_remove_item( imdb->ldb, image->ldb_id);
		
		item.data = image;
		item.size = image_size( imdb, image);
		
		ldb_write_item( imdb->ldb, cursor, &item);
	
	    }
	    
	}
    }
    ldb_cursor_free( imdb->ldb, cursor);

    /* write properties */
    cursor = ldb_cursor_new( imdb->ldb, ImDBPropertyType, LDB_WRITE);
    g_return_if_fail( cursor != NULL);
    list = imdb->properties;
    while( list) {
	property = list->data;
	list = list->next;

	if ( property->flags & ImDBItemNew) {
	    item.data = property;
	    item.size = property_size( imdb, property);

	    ldb_write_item( imdb->ldb, cursor, &item);

	} else if ( property->flags & ImDBItemModified) {
	    ldb_remove_item( imdb->ldb, property->ldb_id);
	    
	    item.data = property;
	    item.size = property_size( imdb, property);

	    ldb_write_item( imdb->ldb, cursor, &item);
	}
    }
    ldb_cursor_free( imdb->ldb, cursor);

    /* write directories */
    cursor = ldb_cursor_new( imdb->ldb, ImDBDirectoryType, LDB_WRITE);
    g_return_if_fail( cursor != NULL);
    for ( i = 0; i < imdb->header.directory_count; i++) {
	directory = &g_array_index( imdb->dirtree, ImDBDirectory, i);

	if ( directory->flags & ImDBItemNew) {
	    item.data = directory;
	    item.size = directory_size( imdb, directory);

	    ldb_write_item( imdb->ldb, cursor, &item);
	
	} else if ( directory->flags & ImDBItemModified) {
	    ldb_remove_item( imdb->ldb, directory->ldb_id);

	    item.data = directory;
	    item.size = directory_size( imdb, directory);

	    ldb_write_item( imdb->ldb, cursor, &item);
	
	}

    }
    ldb_cursor_free( imdb->ldb, cursor);
	    
    ldb_flush( imdb->ldb);
}

void         
imdb_append_images( ImDB *imdb, char *path, GList *images) {
    ImDBImage *image, *dup;
    ImDBDirectory *dir;
    GHashTable *hash;
    GList *list;
    guint dir_id;

    g_return_if_fail( imdb != NULL);
    g_return_if_fail( images != NULL);

    dir_id = imdb_append_path( imdb, path);
    dir = &g_array_index( imdb->dirtree, ImDBDirectory, dir_id);

    hash = g_hash_table_new( g_str_hash, g_str_equal);
    g_return_if_fail( hash != NULL);

    list = dir->images;
    while( list) {
	image = list->data;
	list = list->next;

	g_hash_table_insert( hash, image->image.file_name, image);
    }

    g_hash_table_freeze( hash);

    list = images;
    while ( list) {
	image = list->data;
	list = list->next;
	
	if ( !image) {
	    g_warning( "Trying to append a NULL image");
	    continue;
	}

	/* xxx what do we do if an image with the same path already
	   exists int the database and those images are different ?
	   ( ie width or height or ... differents */
	if ( (dup = g_hash_table_lookup( hash, image->image.file_name)))
	    continue;

	image->directory  = dir_id;
	image->image.path = dir->path; 
	image->ldb_id.block_no = 0;
	image->ldb_id.item_no = 0;
	image->flags = ImDBItemNew;

	dir->images = g_list_prepend( dir->images, image);
	imdb->header.image_count++;
    }

    g_hash_table_destroy( hash);
}

void         
imdb_remove_images( ImDB *imdb, GList *images) {
    ImDBImage *image;
    ImDBDirectory *dir;
    GList *item, *link;

    g_return_if_fail( imdb != NULL);
    g_return_if_fail( images != NULL);

    item = images;
    while ( item) {
	image = item->data;
	item = item->next;

	dir = &g_array_index( imdb->dirtree, ImDBDirectory, image->directory);
	link = g_list_find( dir->images, image);
	g_return_if_fail( link != NULL);

	dir->images = g_list_remove_link( dir->images, link);

	if ( !( image->flags & ImDBItemNew))
	    ldb_remove_item( imdb->ldb, image->ldb_id);
	imdb_free_item( imdb, ImDBImageType, image);
    }
}


gint                    
imdb_image_add_property( ImDB *imdb,
			 ImDBImage *image,
			 gint property_id,
			 gint value_id) {
    ImDBPropertyEntry *entry;

    g_return_val_if_fail( imdb != NULL, FALSE);
    g_return_val_if_fail( image != NULL, FALSE);
    
    if ( !imdb_image_find_property( imdb, image, property_id, value_id)) {
	entry = g_new( ImDBPropertyEntry, 1);
	entry->id = property_id;
	entry->value = value_id;
	image->properties = g_list_prepend( image->properties, entry);

	image->flags |= ImDBItemModified;
    }

    return TRUE;
}

gint                    
imdb_image_delete_property( ImDB *imdb,
			    ImDBImage *image,
			    gint property_id,
			    gint value_id) {
    ImDBPropertyEntry *entry;

    g_return_val_if_fail( imdb != NULL, FALSE);
    g_return_val_if_fail( image != NULL, FALSE);

    entry = imdb_image_find_property( imdb, image, property_id, value_id);
    if ( entry) {
	image->properties = g_list_remove( image->properties, entry);
	g_free( entry);

	image->flags |= ImDBItemModified;

	return TRUE;
    }

    return FALSE;
}

ImDBPropertyEntry*
imdb_image_find_property( ImDB *imdb,
			  ImDBImage *image,
			  gint property_id,
			  gint value_id) {
    ImDBPropertyEntry *entry;
    GList *item;

    item = image->properties;
    while ( item) {
	entry = item->data;
	item = item->next;
	if ( entry->id == property_id && entry->value == value_id)
	    return entry;
    }

    return NULL;
}

ImDBProperty*           
imdb_property_add( ImDB *imdb, gchar *name) {
    ImDBProperty *prop;

    g_return_val_if_fail( imdb != NULL, NULL);
    g_return_val_if_fail( name != NULL, NULL);
    g_return_val_if_fail( imdb->header.property_count != 65535, NULL);

    prop = g_new( ImDBProperty, 1);
    prop->name = g_strdup( name);
    prop->property_id   = ++imdb->header.property_count;
    prop->next_value_id = 0;
    prop->values = NULL;
    imdb->properties = g_list_prepend( imdb->properties, prop);
    prop->ldb_id.block_no = 0;
    prop->ldb_id.item_no = 0;
    prop->flags = ImDBItemNew;

    return prop;
}

ImDBPropertyValue*      
imdb_property_add_value( ImDB *imdb, 
			 ImDBProperty *property, 
			 gchar *name) {
    ImDBPropertyValue *val;

    g_return_val_if_fail( imdb != NULL, NULL);
    g_return_val_if_fail( property != NULL, NULL);
    g_return_val_if_fail( name != NULL, NULL);
    g_return_val_if_fail( property->next_value_id != 255, NULL);

    val = g_new( ImDBPropertyValue, 1);
    val->id = ++property->next_value_id;
    val->name = g_strdup( name);
    property->values = g_list_prepend( property->values, val);

    property->flags |= ImDBItemModified;

    return val;
}

void                    
imdb_property_delete( ImDB *imdb, ImDBProperty *property) {
    g_return_if_fail( imdb != NULL);
    g_return_if_fail( property != NULL);
    g_return_if_fail( imdb->properties != NULL);
    g_return_if_fail( g_list_find( imdb->properties, property));

    if ( !( property->flags & ImDBItemNew)) 
	ldb_remove_item( imdb->ldb, property->ldb_id);
    imdb_free_item( imdb, ImDBPropertyType, property);
}

void                    
imdb_property_delete_value( ImDB *imdb, 
			    ImDBProperty *property, 
			    ImDBPropertyValue *value) {
    
    g_return_if_fail( imdb != NULL);
    g_return_if_fail( property != NULL);
    g_return_if_fail( property->values != NULL);
    g_return_if_fail( value != NULL);
    
    if ( g_list_find( imdb->properties, property) &&
	 g_list_find( property->values, value)) {

	property->values = g_list_remove( property->values, value);

	property->flags |= ImDBItemModified;
    }
}

guint
imdb_append_path( ImDB *imdb, char *path) {
    ImDBDirectory node_d, *child, *parent;
    char  *dir, *tmp;
    guint id;
    
    g_return_val_if_fail( imdb != NULL, 0);
    g_return_val_if_fail( path != NULL, 0);
    g_return_val_if_fail( imdb->header.directory_count != 65535, 0);

    parent = &g_array_index( imdb->dirtree, ImDBDirectory, 0);

    tmp = g_strdup( path);
    dir = strtok( tmp, "/");
    while ( dir) {
	child = &g_array_index( imdb->dirtree, ImDBDirectory, parent->child);
	while ( child->id && strcmp( child->name, dir)) 
	    child = &g_array_index( imdb->dirtree, ImDBDirectory, child->brother);
 
	if ( !child->id)
	    break;
	parent = child;

	dir = strtok( NULL, "/");
    }

    
    if ( dir) {
	node_d.child = 0; node_d.images = NULL;
	while ( dir) {
	    node_d.parent = parent->id;
	    node_d.name = g_strdup( dir);
	    node_d.id = imdb->header.directory_count++;
	    node_d.path = make_path( imdb, &node_d);

	    node_d.brother = parent->child;
	    node_d.ldb_id.block_no = 0;
	    node_d.ldb_id.item_no = 0;
	    node_d.flags = ImDBItemNew;
	    parent->child  = node_d.id;
	    imdb->dirtree = g_array_append_val_compat( imdb->dirtree, ImDBDirectory, node_d);
	    
	    parent = &g_array_index( imdb->dirtree, ImDBDirectory, node_d.id);

	    dir = strtok( NULL, "/");
	}
	id = node_d.id;

    } else
	id = parent->id;

    g_free( tmp);

    return id;
}


static void         
imdb_free_item( ImDB *imdb, guint16 type, void *item) {
    g_return_if_fail( imdb != NULL);
    g_return_if_fail( item != NULL);

    if ( type == ImDBImageType) {
	ImDBImage *dbim;

	dbim = item;
	image_free_preview( &dbim->image);

	if ( dbim->image.file_name)
	    g_free( dbim->image.file_name);

	if ( dbim->properties) {
	    g_list_foreach( dbim->properties, (GFunc) g_free, NULL);
	    g_list_free( dbim->properties);
	}
	if ( dbim->keywords) {
	    g_list_foreach( dbim->keywords, (GFunc) g_free, NULL);
	    g_list_free( dbim->keywords);
	}
	g_free( dbim);

    } else if ( type == ImDBPropertyType) {
	ImDBProperty *property;
	ImDBPropertyValue *value;
	GList *list;

	property = item;
	g_free( property->name);

	list = property->values;
	while ( list) {
	    value = list->data;
	    list = list->next;
	    
	    g_free( value->name);
	    g_free( value);
	}
	if ( property->values)
	    g_list_free( property->values);

	g_free( property);

    } else if ( type == ImDBDirectoryType) {
	/* frees a malloced directory such as those returned by
	   read_directory (not those in imdb->dirtree) */
	ImDBDirectory *directory;

	directory = item;
	if ( directory->name)
	    g_free( directory->name);
	if ( directory->path)
	    g_free( directory->path);
	g_free( directory);

    } else {
	g_warning( "imdb_free_item : unknown item type !");
    }
	
}


/******************************
 *** LDB Callback functions ***
 *****************************/

/*** Images ***/


static gint 
read_image( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data) {
    ImDB *imdb;
    ImDBImage *dbim;
    ImDBPropertyEntry *pentr;
    gint i;

    imdb = data;

    dbim = g_new( ImDBImage, 1);
    dbim->image.preview = NULL;
    dbim->ldb_id = item->id;
    dbim->properties = NULL;
    dbim->keywords = NULL;
    dbim->flags = 0;

    item->data = dbim;

    dbim->image.width = ldb_read_uint16( ldb, cursor);
    dbim->image.height = ldb_read_uint16( ldb, cursor);
    dbim->image.flags = ldb_read_uint8( ldb, cursor);
    dbim->image.file_size = ldb_read_uint32( ldb, cursor);
    dbim->directory = ldb_read_uint16( ldb, cursor);
    dbim->image.file_name = ldb_read_str( ldb, cursor);
    
    if ( dbim->directory >= imdb->header.directory_count) {
	imdb_free_item( imdb, ImDBImageType, dbim);
	return FALSE;
    }
    dbim->image.path = g_array_index( imdb->dirtree, ImDBDirectory, dbim->directory).path;
    
    
#ifdef DEBUG_ITEM
    printf( "Reading image %5dx%-5dx%d %s\n\tpath :%s\n",
	    dbim->image.width, dbim->image.height, dbim->image.flags & IMAGE_HAS_COLORS ? 3 : 1, 
	    dbim->image.file_name, dbim->image.path);
    printf( "\tproperties :");
#endif

    while ( (i = ldb_read_uint8( ldb, cursor))) {
	pentr = g_new( ImDBPropertyEntry, 1);
	pentr->id = i;
	pentr->value = ldb_read_uint8( ldb, cursor);

	dbim->properties = g_list_prepend( dbim->properties, pentr);
#ifdef DEBUG_ITEM
	printf( "%d-%d ", pentr->id, pentr->value);
#endif

    }
 
    dbim->keywords = ldb_read_str_list( ldb, cursor);

#ifdef DEBUG_ITEM
    printf( "\n");
    fflush( stdout);
#endif

    return TRUE;
}

static gint      
write_image( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data) {
    ImDBImage *dbim;
    ImDBPropertyEntry *pentr;
    GList *list;

    dbim = (ImDBImage*) item->data;
    dbim->ldb_id = item->id;
    dbim->flags = 0;

    ldb_write_uint16( ldb, cursor, dbim->image.width);
    ldb_write_uint16( ldb, cursor, dbim->image.height);
    ldb_write_uint8( ldb, cursor, dbim->image.flags);
    ldb_write_uint32( ldb, cursor, dbim->image.file_size);
    ldb_write_uint16( ldb, cursor, dbim->directory);
    ldb_write_str( ldb, cursor, dbim->image.file_name);
    
    list = dbim->properties;
    while ( list) {
	pentr = list->data;
	list = list->next;

	ldb_write_uint8( ldb, cursor, pentr->id);
	ldb_write_uint8( ldb, cursor, pentr->value);
    }
    /* marks the end of the list */
    ldb_write_uint8( ldb, cursor, 0);

    ldb_write_str_list( ldb, cursor,dbim->keywords);

    return TRUE;
}

static gint      
dump_image( FILE *dump_file, ImDBImage *dbim) {
    ImDBPropertyEntry *pentr;
    GList *list;

    fprintf( dump_file, "{ %d %d %d %d %d %s { ",
	     dbim->image.width, dbim->image.height, 
	     dbim->image.flags, dbim->image.file_size,
	     dbim->directory, dbim->image.file_name);

    
    list = dbim->properties;
    while ( list) {
	pentr = list->data;
	list = list->next;

	fprintf( dump_file, "%d %d, ", pentr->id, pentr->value);
    }
    fprintf( dump_file, "} { ");

    list = dbim->keywords;
    while ( list) {
	fprintf( dump_file, "%s ,", (char*) list->data);
	list = list->next;
    }
    fprintf( dump_file, "}}\n");

    return TRUE;
}

static guint
image_size( ImDB *imdb, ImDBImage *dbim) {
    guint size;

    if ( !dbim->image.file_name)
	return 0;

    size = 11 + ldb_str_size( imdb->ldb, dbim->image.file_name);
    size += g_list_length( dbim->properties) * 2;
    size ++;

    size += ldb_str_list_size( imdb->ldb, dbim->keywords);

    return size;
}
    

/*** Properties ***/

static gint
read_property( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data) {
    ImDBProperty *property;
    ImDBPropertyValue *value;
    gint i;

    property = g_new( ImDBProperty, 1);
    property->ldb_id = item->id;
    property->values = NULL;
    property->flags = 0;

    item->data = property;

    property->property_id = ldb_read_uint8( ldb, cursor);
    property->next_value_id = ldb_read_uint8( ldb, cursor);
    property->name =  ldb_read_str( ldb, cursor);

#ifdef DEBUG_ITEM
    printf( "Reading property %d %d %s\n", 
	     property->property_id, property->next_value_id, property->name);
    printf( "\tValues :");
#endif

    while ( (i = ldb_read_uint8( ldb, cursor))) {
	value = g_new( ImDBPropertyValue, 1);
	value->id = i;
	value->name = ldb_read_str( ldb, cursor);

	property->values = g_list_prepend( property->values, value);

#ifdef DEBUG_ITEM
	printf( " %s(%d)", value->name, value->id);
#endif 
    }

#ifdef DEBUG_ITEM
    printf( "\n");
    fflush( stdout);
#endif

    return TRUE;
}

static gint      
write_property( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data) {
    ImDBProperty *property;
    ImDBPropertyValue *value;
    GList *list;

    property = item->data;
    property->ldb_id = item->id;
    property->flags = 0;

    ldb_write_uint8( ldb, cursor, property->property_id);
    ldb_write_uint8( ldb, cursor, property->next_value_id);
    ldb_write_str( ldb, cursor, property->name);


    list = property->values;
    while ( list) {
	value = list->data;
	list = list->next;
 
	ldb_write_uint8( ldb, cursor, value->id);
	ldb_write_str( ldb, cursor, value->name);
    }
    ldb_write_uint8( ldb, cursor, 0);

    return TRUE;
}

static gint      
dump_property( FILE *dump_file, ImDBProperty *property) {
    ImDBPropertyValue *value;
    GList *list;

    fprintf( dump_file, "{ %d %d %s { ", property->property_id, 
	     property->next_value_id, property->name);
    list = property->values;
    while ( list) {
	value = list->data;
	list = list->next;

	fprintf( dump_file, "%d %s, ", value->id, value->name);
    }
    fprintf( dump_file, "}}\n");

    return TRUE;
}

static guint
property_size( ImDB *imdb, ImDBProperty *property) {
    ImDBPropertyValue *value;
    guint size;
    GList *list;
    
    if ( !property->name)
	return 0;

    size = ldb_str_size( imdb->ldb, property->name) + 2;

    list = property->values;
    while ( list) {
	value = list->data;

	size += ldb_str_size( imdb->ldb, value->name) + 2;
	list = list->next;
    }
    size ++;

    return size;
}

/*** Directory ***/

static gint
read_directory( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data) {
    ImDBDirectory *directory;

    directory = g_new( ImDBDirectory, 1);
    directory->path = NULL;
    directory->ldb_id = item->id;
    directory->flags = 0;
    directory->images = NULL;

    item->data = directory;
    
    directory->id = ldb_read_uint16( ldb, cursor);
    directory->parent = ldb_read_uint16( ldb, cursor);
    directory->name = ldb_read_str( ldb, cursor);

#ifdef DEBUG_ITEM
    printf( "Reading directory %-4d<-%4d %s\n",
	     directory->id, directory->parent, directory->name );
    fflush( stdout);
#endif

    return TRUE;
}

static gint      
write_directory( LDB *ldb, LDBCursor *cursor, LDBItem *item, void *data) {
    ImDBDirectory *directory;

    directory = item->data;
    directory->ldb_id = item->id;
    directory->flags = 0;

    ldb_write_uint16( ldb, cursor, directory->id);
    ldb_write_uint16( ldb, cursor, directory->parent);
    ldb_write_str( ldb, cursor, directory->name);

    return TRUE;
}

static gint      
dump_directory( FILE *dump_file, ImDBDirectory *directory) {
    fprintf( dump_file, "{ %d %d %s }\n", directory->id, directory->parent, directory->name);

    return TRUE;
}

static guint
directory_size( ImDB *imdb, ImDBDirectory *dir) {

    if ( !dir->name)
	return 0;

    return 4 + ldb_str_size( imdb->ldb, dir->name);
}

/************
 *** Misc ***
 ************/

static char*     
make_path( ImDB *imdb, ImDBDirectory *node) {
    GList *dirs = NULL, *dir;
    guint size = 1;
    char *path, *tmp;

    
    while ( node->id) { 
	dirs = g_list_prepend( dirs, node->name);
	size += strlen( node->name) + 1;

	node = &g_array_index( imdb->dirtree, ImDBDirectory, node->parent);
    } 

    if ( size == 1)
	return NULL;

    tmp = path = g_malloc( size);
    dir = dirs;
    while ( dir) {
	sprintf( tmp, "/%s", (char*) dir->data);
	tmp += strlen( (char*) dir->data) + 1;
	dir = dir->next;
    }
    g_list_free( dirs);

    return path;
}
