/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>
#ifdef HAVE_LIBJPEG

#include <stdio.h>
#include <math.h>
#include <jpeglib.h>
#include <setjmp.h>
#include <glib.h>

#include "jpeg.h"



struct my_error_mgr {
  struct jpeg_error_mgr pub;	

  jmp_buf setjmp_buffer;
};

typedef struct my_error_mgr *my_error_ptr;

static gint           jpeg_read_headers2( Image *image, 
					  struct jpeg_decompress_struct *cinfo, 
					  struct my_error_mgr *jerr);
static void           my_error_exit( j_common_ptr cinfo);
static void           my_output_message( j_common_ptr cinfo);


void
register_jpeg() {
    char *extensions[] = { "jpg", "jpeg", NULL};
    ImageFormat jpeg_format = {
	extensions,
	is_jpeg,
	jpeg_read_headers,
	jpeg_read,
	NULL
    };

    image_format_register( &jpeg_format);
}

gint
is_jpeg( Image *image) {
    guchar buff[4];

    if ( !fread( buff, 4, 1, image->preview->file))
	return FALSE;

    fseek( image->preview->file, 0L, SEEK_SET);

    return buff[0] == 0xff && buff[1] == 0xd8 && buff[2] == 0xff;
}



static void
my_error_exit( j_common_ptr cinfo) {
    my_error_ptr myerr = (my_error_ptr) cinfo->err;
    
    (*cinfo->err->output_message) (cinfo);

    longjmp(myerr->setjmp_buffer, 1);
}

static void
my_output_message( j_common_ptr cinfo) {
    if ( cinfo->err->msg_code <= cinfo->err->last_jpeg_message)
	puts( cinfo->err->jpeg_message_table[ cinfo->err->msg_code]);
}


static gint
jpeg_read_headers2( Image *image, struct jpeg_decompress_struct *cinfo, struct my_error_mgr *jerr) {

    g_return_val_if_fail( image != NULL, FALSE);
    g_return_val_if_fail( image->preview != NULL, FALSE);
    g_return_val_if_fail( image->preview->file != NULL, FALSE);
    
    if ( setjmp(jerr->setjmp_buffer)) {
	    
	jpeg_abort_decompress( cinfo);
	return FALSE;
    }

    jpeg_stdio_src( cinfo, image->preview->file);
    jpeg_read_header( cinfo, TRUE);

    image_set_infos( image, 
		     JPEG_IMAGE,
		     cinfo->image_width, 
		     cinfo->image_height, 
		     cinfo->num_components == 3,
		     FALSE);

    return TRUE;
}
    
gint
jpeg_read_headers( Image *image) {
    struct jpeg_decompress_struct cinfo;
    struct my_error_mgr jerr;
    
    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;
    jerr.pub.output_message = my_output_message;
    
    jpeg_create_decompress( &cinfo);
    
    if ( jpeg_read_headers2( image, &cinfo, &jerr)) {
	jpeg_abort_decompress( &cinfo);
	return TRUE;
    }

    jpeg_destroy_decompress( &cinfo);
    return FALSE;
}

gint
jpeg_read( Image *image) {
    struct jpeg_decompress_struct cinfo;
    struct my_error_mgr jerr;
    JSAMPARRAY buffer;
    int sl, rw, rh;

    gint row_stride;    

    g_return_val_if_fail( image != NULL && image->preview != NULL, FALSE);
  
    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;
    jerr.pub.output_message = my_output_message;
    jpeg_create_decompress( &cinfo);
   
    if ( !jpeg_read_headers2( image, &cinfo, &jerr)) 
	return FALSE;
    
    if ( setjmp(jerr.setjmp_buffer)) {
	
	jpeg_abort_decompress( &cinfo);
	return FALSE;
    }

    rw = image_get_render_width( image);
    rh = image_get_render_height( image);

    if ( (rw < image->width) ||
	 (rh < image->height)) {
	float propw, proph;
	guint scale;

	propw = (float) image->width / rw;
	proph = (float) image->height /  rh;
	scale = MIN(rint(propw), rint(proph));
	if ( scale > 7)
	    scale = 8;
	else if ( scale > 3)
	    scale = 4;
	else if ( scale > 1)
	    scale = 2;
	else 
	    scale = 1;
	
	cinfo.scale_num = 1;
	cinfo.scale_denom = scale;
    }
  
    cinfo.dither_mode = JDITHER_NONE;

    jpeg_calc_output_dimensions( &cinfo);
    row_stride = cinfo.output_width * cinfo.output_components;

    /* the buffers must be allocated before calling jpeg_start_decompress */
    buffer = (*cinfo.mem->alloc_sarray)
    ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, cinfo.rec_outbuf_height);

    jpeg_start_decompress(&cinfo);

    switch ( cinfo.output_components) {
    case 1:
	image_render_init( image, cinfo.output_width, cinfo.output_height, IMAGE_COLOR_8);
	break;
    case 3:
	image_render_init( image, cinfo.output_width, cinfo.output_height, IMAGE_COLOR_24);
	break;
    default:
	/* This shouldn't happen */
	g_warning( "JPEG : unsupported color type");
	return FALSE;
    }

    while (cinfo.output_scanline < cinfo.output_height) {
	jpeg_read_scanlines(&cinfo, buffer, cinfo.rec_outbuf_height);
	for ( sl = 0; sl < cinfo.rec_outbuf_height; sl++)
	    if ( !image_append_rows( image, 
				     1,
				     buffer[sl])) {
		jpeg_abort_decompress( &cinfo);
		return FALSE;
	    }
    }

    jpeg_finish_decompress( &cinfo);

    jpeg_destroy_decompress( &cinfo);

    return TRUE;
}

#endif /* HAVE_LIBJPEG */

