/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <glib.h>

#include "ldb.h"
#include "ldbutils.h"
#include "gtkcompat.h"

//#define DEBUG
//#define DEBUG_ITEM
#define DEFAULT_BLOCK_SIZE 1024
#define FIRST_ITEM_ID 1
#define LAST_ITEM_ID 0xFFFF

static void       ldb_block_corrupted( LDB *ldb, LDBBlock *block);
static void       ldb_error( LDB *ldb);
static void       empty_item( LDB *ldb, LDBBlock *block, guint16 item_no);
static void       empty_deleted_items( LDB *ldb, LDBBlock *block);
static void       delete_empty_items( LDB *ldb, LDBBlock *block);

LDB*        
ldb_create( char *path, 
	    guint block_size,
	    gint  types,
	    gint header_size) {

    LDB  *ldb;
    guint i;
    gint fd;
    struct flock lock;

    g_return_val_if_fail( path != NULL, FALSE);
    g_return_val_if_fail( types > 0 && header_size > 0, NULL);

#ifdef DEBUG
    printf( "LDB: Creating database :\n\tblock size: %d\n\ttypes: %d\n",
	     block_size, types);
#endif

    fd = open( path, O_CREAT | O_EXCL | O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
    if ( fd == -1) {
	g_warning( "ldb: Can't open %s: %s", path, strerror( errno));
	return NULL;
    }
    
    lock.l_type = F_WRLCK;
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;
    if ( fcntl( fd, F_SETLK, &lock) == -1) {
	g_warning( "ldb: Can't obtain a lock on %s: %s", path, strerror( errno));
	close( fd);
	return NULL;
    }

    ldb = g_new( LDB, 1);
    ldb->fd = fd;
    ldb->header.user_header_size = header_size;
    ldb->header.types = types;
    ldb->header.version = LDB_VERSION;
    ldb->header.block_size = block_size ? block_size : DEFAULT_BLOCK_SIZE;
    ldb->header.block_count = 0;
    memcpy( ldb->header.marker, LDB_MARKER, LDB_MARKER_SIZE);
    ldb->header.type_info = g_new( LDBTypeInfo, types);
    ldb->methods = g_new( LDBTypeMethods, types);
    bzero( ldb->methods,  types*sizeof( LDBTypeMethods));
    ldb->blocks = g_array_new_compat( LDBBlock);

    for ( i = 0; i < types; i++) {
	ldb->header.type_info[ i].last_block = ldb->header.type_info[ i].first_block = 0xffffffff;
	ldb->header.type_info[ i].block_count = ldb->header.type_info[ i].item_count = 0;
    }
    

    ldb->first_block_pos = ldb->header.user_header_size + LDB_HEADER_SIZE + types * LDB_TYPE_INFO_SIZE; 
    for ( i = 0; i < ldb->first_block_pos; i++)
	if ( !write_char( fd, 0)) {
	    ldb_error( ldb);
	    g_free( ldb);
	    return NULL;
	}
    


#ifdef DEBUG
    printf( "\tfirst_block_pos: %ld\n", ldb->first_block_pos);
    fflush( stdout);
#endif
    return ldb;
}


LDB*        
ldb_load( char *path, gint header_size) {
    LDB   *ldb;
    LDBBlock block = { 0};
    struct flock lock;
    glong pos;
    guint i;
    gint fd;
    
    
    g_return_val_if_fail( path != NULL, NULL);
    g_return_val_if_fail( header_size >= 0, NULL);

#ifdef DEBUG
    printf( "LDB: Loading database ...\n");
#endif

    fd = open( path, O_RDWR, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
    if ( fd == -1) {
	g_warning( "ldb: Can't load %s: %s", path, strerror( errno));
	return NULL;
    }

    lock.l_type = F_WRLCK;
    lock.l_whence = SEEK_SET;
    lock.l_start = 0;
    lock.l_len = 0;
    if ( fcntl( fd, F_SETLK, &lock) == -1) {
	g_warning( "ldb: Can't obtain a lock on %s: %s", path, strerror( errno));
	close( fd);
	return NULL;
    }

    ldb = g_new( LDB, 1);
    ldb->fd = fd;

    if ( !read_uint16( fd, &ldb->header.user_header_size) || 
	 !read_uint16( fd, &ldb->header.types) ||
	 !read_uint16( fd, &ldb->header.version) ||
	 !read_uint16( fd, &ldb->header.block_size) ||
	 !read_uint32( fd, &ldb->header.block_count) ||
	 !read_str( fd, ldb->header.marker, LDB_MARKER_SIZE)) {
	ldb_error( ldb);
	g_free( ldb);
	close( fd);
	return NULL;
    }

    if ( memcmp( ldb->header.marker, LDB_MARKER, LDB_MARKER_SIZE)) {
	g_warning( "This is not an ldb database");
	g_free( ldb);
	close( fd);
	return NULL;
    }

    if ( ldb->header.version > LDB_VERSION) {
	g_warning( "Database is newer than this program");
	g_free( ldb);
	close( fd);
	return NULL;
    }

    ldb->first_block_pos = ldb->header.user_header_size + LDB_HEADER_SIZE + 
	ldb->header.types * LDB_TYPE_INFO_SIZE;

#ifdef DEBUG
    printf( "\tblock_size: %d\n\ttypes: %d\n\tversion: %d\n\tblock_count: %d\n",
	     ldb->header.block_size, ldb->header.types, ldb->header.version,
	     ldb->header.block_count);
    printf( "tfirst_block_pos %ld\n", ldb->first_block_pos);
    fflush( stdout);
#endif

    ldb->header.type_info = g_new( LDBTypeInfo, ldb->header.types);
    for ( i = 0; i < ldb->header.types; i++) {
	if ( !read_uint32( fd, &ldb->header.type_info[ i].first_block) ||
	     !read_uint32( fd, &ldb->header.type_info[ i].last_block) ||
	     !read_uint32( fd, &ldb->header.type_info[ i].block_count) ||
	     !read_uint32( fd, &ldb->header.type_info[ i].item_count)) {
	    ldb_error( ldb);
	    g_free( ldb->header.type_info);
	    g_free( ldb);
	    close( fd);
	    return NULL;
	}
#ifdef DEBUG
	printf( "\ttype %d (first block: %d last block: %d block_count: %d item count: %d\n",
		i,
		ldb->header.type_info[ i].first_block,
		ldb->header.type_info[ i].last_block,
		ldb->header.type_info[ i].block_count,
		ldb->header.type_info[ i].item_count);
    
	fflush( stdout);
#endif
    }

    ldb->methods = g_new( LDBTypeMethods, ldb->header.types);
    bzero( ldb->methods, sizeof(  LDBTypeMethods) * ldb->header.types);
    
    pos = ldb->first_block_pos + ldb->header.block_count * ldb->header.block_size;
    if ( lseek( ldb->fd, pos, SEEK_SET) == -1) {
	ldb_error( ldb);
	g_array_free( ldb->blocks, TRUE);
	g_free( ldb->header.type_info);
	g_free( ldb->methods);
	g_free( ldb);
	close( fd);
	return NULL;
    }

    ldb->blocks = g_array_new_compat( LDBBlock);
   
    for ( i = 0; i < ldb->header.block_count; i++) {
	block.block_no = i;
	if ( !read_uint16( fd, &block.header.free_space) ||
	     !read_uint32( fd, &block.header.prev_block) ||
	     !read_uint16( fd, &block.header.next_item_id)) {
	    ldb_error( ldb);
	    g_array_free( ldb->blocks, TRUE);
	    g_free( ldb->header.type_info);
	    g_free( ldb->methods);
	    g_free( ldb);
	    close( fd);
	    return NULL;
	}
	ldb->blocks = g_array_append_val_compat( ldb->blocks, LDBBlock, block);

	if ( block.header.free_space > ldb->header.block_size) {
#ifdef DEBUG
	    printf( "*** Free space in block %5d is bigger than the block size\n", i);
	    fflush( stdout);
#endif
	    ldb_block_corrupted( ldb, &g_array_index( ldb->blocks, LDBBlock, i));
	}

#ifdef DEBUG
	printf( "Reading header of block %d ( prev: %d free_space: %d next_item_id: %d)\n", 
		i, block.header.prev_block, block.header.free_space, block.header.next_item_id);
	fflush( stdout);
#endif
    }

    return ldb;
}

    
void        
ldb_close( LDB *ldb) {
    LDBBlock *block;
    guint i;

    g_return_if_fail( ldb != NULL);

#ifdef DEBUG
    printf( "LDB: Closing Database\n");
    fflush( stdout);
#endif

    for ( i = 0; i < ldb->header.block_count; i++) {
	block = &g_array_index( ldb->blocks, LDBBlock, i);
	if ( block->deleted_items) {
	    ldb_read_block( ldb, block->block_no);
	    empty_deleted_items( ldb, block);
	    delete_empty_items( ldb, block);
	}
	ldb_release_block( ldb, block);
    }

    ldb_flush( ldb);

    close( ldb->fd);

    g_array_free( ldb->blocks, TRUE);
    g_free( ldb->methods);
    g_free( ldb);

}

gint
ldb_flush( LDB *ldb) {
    guint i;
    LDBBlock *block;
    gint fd;
    long pos;
    
#ifdef DEBUG
    printf( "LDB: Flushing database\n");
    fflush( stdout);
#endif

    for ( i = 0; i < ldb->header.block_count; i++) {
	block = &g_array_index( ldb->blocks, LDBBlock, i);
	if ( block->flags & LDB_BLOCK_MODIFIED)
	    ldb_write_block( ldb, block);
    }

    /* write the header */
    fd = ldb->fd;

    if ( (lseek( fd, 0, SEEK_SET) == -1) ||
	 !write_uint16( fd, ldb->header.user_header_size) ||
	 !write_uint16( fd, ldb->header.types) ||
	 !write_uint16( fd, ldb->header.version) ||
	 !write_uint16( fd, ldb->header.block_size) ||
	 !write_uint32( fd, ldb->header.block_count) ||
	 !write_str( fd, ldb->header.marker, LDB_MARKER_SIZE)) {
	ldb_error( ldb);
	return FALSE;
    }

    for ( i = 0; i < ldb->header.types; i++)
	if ( !write_uint32( fd, ldb->header.type_info[i].first_block) ||
	     !write_uint32( fd, ldb->header.type_info[i].last_block) ||
	     !write_uint32( fd, ldb->header.type_info[i].block_count) ||
	     !write_uint32( fd, ldb->header.type_info[i].item_count)) {
	    ldb_error( ldb);
	    return FALSE;
	}

	
    /* write each block header at the end of the file */
    pos = ldb->first_block_pos + ldb->header.block_count * ldb->header.block_size;
    if ( lseek( ldb->fd, pos, SEEK_SET) == -1) {
	ldb_error( ldb);
	return FALSE;
    }

    for ( i = 0; i < ldb->header.block_count; i++) {
	block = &g_array_index( ldb->blocks, LDBBlock, i);
	if ( !write_uint16( fd, block->header.free_space) ||
	     !write_uint32( fd, block->header.prev_block) ||
	     !write_uint16( fd, block->header.next_item_id)) {
	    ldb_error( ldb);
	    return FALSE;
	}
    }

    return TRUE;
}

glong
ldb_get_user_header_position( LDB *ldb) {
    g_return_val_if_fail( ldb != NULL, -1);

    return ldb->first_block_pos - ldb->header.user_header_size;
}

LDBBlock*
ldb_new_block( LDB *ldb, 
	       guint16 type) {
    LDBBlock block;
    register gint count;
    long pos;

    g_return_val_if_fail( ldb != NULL, NULL);
    g_return_val_if_fail( type <= ldb->header.types, NULL);
    g_return_val_if_fail( ldb->header.block_count != 0xffffffff /* ouch */, NULL);

    block.block_no = ldb->header.block_count;
    block.flags = LDB_BLOCK_MODIFIED | LDB_BLOCK_ALREADY_READ;
    block.deleted_items = NULL;
    block.header.prev_block = ldb->header.type_info[type].last_block;
    block.header.free_space = ldb->header.block_size;
    block.header.next_item_id = FIRST_ITEM_ID;
    block.data = g_malloc( ldb->header.block_size);
    memset( block.data, '-', ldb->header.block_size);

    pos = ldb->first_block_pos + 
	block.block_no * ldb->header.block_size;
    
    if ( lseek( ldb->fd, pos, SEEK_SET) == -1) {
	g_free( block.data);
	ldb_error( ldb);
	return NULL;
    }

    while ( (count = write( ldb->fd, block.data, ldb->header.block_size)) == -1 &&
	    errno == EINTR)
	;
    
    if ( count == -1) {
	g_free( block.data);
	ldb_error( ldb);
	return NULL;
    }

    ldb->blocks = g_array_append_val_compat( ldb->blocks, LDBBlock, block);
   
    if ( !~ldb->header.type_info[type].first_block)
	ldb->header.type_info[type].first_block = block.block_no;

    ldb->header.type_info[type].last_block = block.block_no;

#ifdef DEBUG
    g_return_val_if_fail( g_array_index( ldb->blocks, LDBBlock, block.block_no).block_no == block.block_no, NULL);
    printf( "LDB: new block %d (%d)\n",
	    ldb->header.type_info[ type].block_count,
	    type);
    fflush( stdout);
#endif

    ldb->header.type_info[ type].block_count++;
    ldb->header.block_count++;

    return &g_array_index( ldb->blocks, LDBBlock, block.block_no);
}


LDBBlock* 
ldb_read_block( LDB *ldb, 
		guint32 block_no) {
    LDBBlock *block;
    register gint count;
    long pos;
    
    g_return_val_if_fail( ldb != NULL, NULL);
    g_return_val_if_fail( block_no <= ldb->header.block_count, NULL);

    block = &g_array_index( ldb->blocks, LDBBlock, block_no);

    if ( block->data) {
#ifdef DEBUG
	printf( "LDB: Reading block %d from cache\n",
		block_no);
	fflush( stdout);
#endif
	return block;

    }

    pos = ldb->first_block_pos + 
	block_no * ldb->header.block_size;

    if ( lseek( ldb->fd, pos, SEEK_SET) == -1) {
	ldb_error( ldb);
	return NULL;
    }


#ifdef DEBUG
    printf( "LDB: Reading block %d at %ld ( free: %d prev %d)\n",
	    block_no, pos, block->header.free_space, block->header.prev_block);
    fflush( stdout);
#endif

    block->data = g_malloc( ldb->header.block_size);
    while ( (count = read( ldb->fd, block->data, ldb->header.block_size)) == -1 &&
	    (errno == EAGAIN || errno == EINTR))
	;
     
    if ( count == -1) {
	ldb_error( ldb);
	g_free( block->data);
	block->data = NULL;
	return NULL;
    }

    if ( !(block->flags & LDB_BLOCK_ALREADY_READ)) {
	block->flags |= LDB_BLOCK_ALREADY_READ;
	delete_empty_items( ldb, block);
    }

    return block;
}


gint        
ldb_write_block( LDB *ldb, 
		 LDBBlock *block) {
    long pos;
    register gint count;

    g_return_val_if_fail( ldb != NULL, FALSE);
    g_return_val_if_fail( block != NULL, FALSE);
    g_return_val_if_fail( block->data != NULL, FALSE);

    empty_deleted_items( ldb, block);

    pos = ldb->first_block_pos + 
	block->block_no * ldb->header.block_size;
    
    if ( lseek( ldb->fd, pos, SEEK_SET) == -1) {
	ldb_error( ldb);
	return FALSE;
    }

#ifdef DEBUG
    printf( "LDB: Writing block %d at %ld ( free: %d prev %d)\n",
	    block->block_no, pos, block->header.free_space, block->header.prev_block);
    fflush( stdout);
#endif


    while ( (count = write( ldb->fd, block->data, ldb->header.block_size)) == -1 &&
	    errno == EINTR)
	;

    if ( count == -1) {
	ldb_error( ldb);
	return FALSE;
    }

    block->flags &= ~LDB_BLOCK_MODIFIED;

    return TRUE;
}


void        
ldb_release_block( LDB *ldb, LDBBlock *block) {
    g_return_if_fail( ldb != NULL);
    
    if ( block->flags & LDB_BLOCK_MODIFIED)
	ldb_write_block( ldb, block);

    g_free( block->data);
    block->data = NULL;
}

static void
ldb_block_corrupted( LDB *ldb, LDBBlock *block) {
    g_return_if_fail( ldb != NULL);
    g_return_if_fail( block != NULL);
     
    g_warning( "LDB: ERROR: Block %5d is corrupted. Deleting this block.", block->block_no);

    if ( block->deleted_items) {
	g_slist_free( block->deleted_items);
	block->deleted_items = NULL;
    }

    block->header.free_space = ldb->header.block_size;
    block->header.next_item_id = FIRST_ITEM_ID;

    block->flags |= LDB_BLOCK_MODIFIED;

}

/***************
 *** Methods ***
 ***************/

void        
ldb_set_read( LDB *ldb, 
	      gint type,
	      LDBRead read, 
	      void *read_data) {

    g_return_if_fail( ldb != NULL);
    g_return_if_fail( type < ldb->header.types);
    g_return_if_fail( read != NULL);

    ldb->methods[ type].read = read;
    ldb->methods[ type].read_data = read_data;
}

void        
ldb_set_write( LDB *ldb, 
	       gint type,
	       LDBWrite write, 
	       void *write_data) {

    g_return_if_fail( ldb != NULL);
    g_return_if_fail( type < ldb->header.types);
    g_return_if_fail( write != NULL);

    ldb->methods[ type].write = write;
    ldb->methods[ type].write_data = write_data;
}




/***************
 *** Cursors ***
 ***************/

/* xxx only one cursor should be used at a time: 
   if you add/remove some data in one block and has 
   another cursor (c2) opened for that block then the
   ldb->private_pos pointer of c2 will probably
   be invalid. If c2 is an LDB_WRITE cursor than 
   it will surely corupt the block.
   */
LDBCursor*  
ldb_cursor_new( LDB *ldb, guint16 type, LDBCursorMode mode) {
    LDBCursor *cursor;

    g_return_val_if_fail( ldb != NULL, NULL);
    g_return_val_if_fail( type < ldb->header.types, NULL);
   

    cursor = g_new( LDBCursor, 1); 
    cursor->mode = mode;
    if ( ~ldb->header.type_info[ type].last_block)
	cursor->block = ldb_read_block( ldb, 
					ldb->header.type_info[ type].last_block);
    else if ( mode == LDB_WRITE)
	cursor->block = ldb_new_block( ldb, type);
    else
	cursor->block = NULL;

    if ( !cursor->block) {
	g_free( cursor);
	return NULL;
    }

    if ( mode == LDB_READ) 
	cursor->private_pos = cursor->block->data;
    else if ( mode == LDB_WRITE)
	cursor->private_pos = cursor->block->data + 
	    ldb->header.block_size - cursor->block->header.free_space;

    cursor->type = type;
    cursor->item_no = FIRST_ITEM_ID; /* used only by ldb_read_next_item */

    return cursor;
}

void
ldb_cursor_free( LDB *ldb, 
		 LDBCursor *cursor) {

    g_return_if_fail( ldb != NULL);
    g_return_if_fail( cursor != NULL);

    if ( cursor->block)
	ldb_release_block( ldb, cursor->block);

    g_free( cursor);
}


gint
ldb_read_next_item( LDB *ldb, 
		    LDBCursor *cursor,
		    LDBItem *item) {
    
    g_return_val_if_fail( ldb != NULL, FALSE);
    g_return_val_if_fail( cursor != NULL, FALSE);
    g_return_val_if_fail( cursor->mode == LDB_READ, FALSE);
    g_return_val_if_fail( ldb->methods[ cursor->type].read != NULL, FALSE);


    do {
	do {
	    
	    if ( cursor->item_no == cursor->block->header.next_item_id) {
		
		ldb_release_block( ldb, cursor->block);
		
		if ( !~cursor->block->header.prev_block) {
		    cursor->block = NULL;
		    return FALSE;
		}
		
		cursor->block = ldb_read_block( ldb, 
						cursor->block->header.prev_block);
		
		g_return_val_if_fail( cursor->block != NULL, FALSE);

		cursor->item_no = FIRST_ITEM_ID;
		cursor->private_pos = cursor->block->data;

		if ( cursor->block->header.next_item_id == FIRST_ITEM_ID)
		    /* empty block */
		    continue;
		
#ifdef DEBUG
		printf( "LDB: cursor now on block %d\n", cursor->block->block_no);
		fflush( stdout);
#endif	
	    }
	    
	    item->id.block_no = cursor->block->block_no;
	    item->id.item_no  = cursor->item_no++; 
	    
	    item->size = GUINT16BE( GUINT16(cursor->private_pos));
	    if ( (cursor->private_pos + item->size) > (cursor->block->data + ldb->header.block_size)) {
#ifdef DEBUG
		printf( "*** LDB: Item %5d-%-4d is too big to fit in block (at pos=+%d size=%d free=%d)\n", 
			item->id.block_no,
			item->id.item_no,
			cursor->private_pos - cursor->block->data,
			item->size, cursor->block->header.free_space);
		fflush( stdout);
#endif
		ldb_block_corrupted( ldb, cursor->block);
		
		cursor->item_no = cursor->block->header.next_item_id;
		continue;
	    }

	    if ( item->size > 2 && ( !cursor->block->deleted_items ||
		 !g_slist_find( cursor->block->deleted_items, GUINT_TO_POINTER( (guint) item->id.item_no))))
		break;
	    
	    cursor->private_pos += item->size;
	    
	}  while ( 1); 
	
	cursor->pos  = cursor->private_pos + 2;
	cursor->private_pos += item->size;
	
#ifdef DEBUG_ITEM
	printf( "LDB: Reading item %5d-%-4d size: %d at block +%d\n",
		item->id.block_no, item->id.item_no, item->size, cursor->pos - cursor->block->data);
	fflush( stdout);
#endif
    
    } while ( !(*ldb->methods[ cursor->type].read)( ldb, 
						    cursor, 
						    item,
						    ldb->methods[ cursor->type].read_data)); 
	      
    if ( cursor->pos > cursor->private_pos)
	    g_warning( "Reading after the end of this item");

    return TRUE;
}

/* xxx TODO: How do we choose where to write an item ? Simple solution:
   Find the smaller piece of free space available in the database
   ( such that the item can be written there), load the block
   and write the item ?
   Other solution: Keep blocks with a defined amount of free space
   in memory and write the new items in those blocks. If there's no
   such blocks than apply previous method (or allocate a new block ?).
   Other solution: ???

   The current method is too write new items in a new block without
   using free space in others blocks.
   This need to be changed.
   */

gint        
ldb_write_item( LDB *ldb, 
		LDBCursor *cursor, 
		LDBItem *item) {
    gint retval;

    g_return_val_if_fail( ldb != NULL, FALSE);
    g_return_val_if_fail( cursor != NULL, FALSE);
    g_return_val_if_fail( cursor->mode == LDB_WRITE, FALSE);
    g_return_val_if_fail( ldb->methods[ cursor->type].write != NULL, FALSE);
    g_return_val_if_fail( item->size > 0, FALSE);

    item->size += 2;
    if ( (cursor->block->header.free_space < item->size) ||
	 (cursor->block->header.next_item_id == LAST_ITEM_ID) ) { 
	/* next_item_id cannot be greater than 0xFFFF
	   ( because item->id.item_no is a guint16)
	   */
	
	ldb_release_block( ldb, cursor->block);
	
	if ( !(cursor->block = ldb_new_block( ldb, cursor->type)))
	    return FALSE;

	if ( cursor->block->header.free_space < item->size) {
	    g_warning( "LDB: Block size (%db) too small to old this item (%db)",
		       ldb->header.block_size, item->size);
	    return FALSE;
	}

#ifdef DEBUG
	printf( "LDB: cursor now on block %d\n", cursor->block->block_no);
	fflush( stdout);
#endif
	
	cursor->private_pos = cursor->block->data;
	
    }
    
    if ( !(cursor->block->flags & LDB_BLOCK_MODIFIED))
	cursor->block->flags |= LDB_BLOCK_MODIFIED;
    
    ldb->header.type_info[ cursor->type].item_count++;
    item->id.block_no = cursor->block->block_no;
    item->id.item_no  = cursor->block->header.next_item_id++;
    GUINT16(  cursor->private_pos) = GUINT16BE( item->size);
    
    cursor->pos = cursor->private_pos + 2;
    cursor->private_pos += item->size;

#ifdef DEBUG_ITEM
    printf( "LDB: Writing item %5d-%-4d size: %d at block +%d free_space:  %d\n",
	    cursor->block->block_no, item->id.item_no, item->size, 
	    cursor->pos - cursor->block->data, cursor->block->header.free_space);
    fflush( stdout);
#endif

    cursor->block->header.free_space -= item->size;

    retval = (*ldb->methods[ cursor->type].write)( ldb, 
						   cursor, 
						   item, 
						   ldb->methods[ cursor->type].write_data); 

    if ( (cursor->pos > cursor->private_pos) || 
	 (cursor->block->header.free_space != 0 && *cursor->private_pos != '-'))
	g_warning( "LDB: Writing more data than requested");


    return retval;
}

    

gint
ldb_unread_item( LDB *ldb, 
		 LDBCursor *cursor) {
    
    g_return_val_if_fail( ldb != NULL, FALSE);
    g_return_val_if_fail( cursor != NULL, FALSE);
    g_return_val_if_fail( cursor->block != NULL, FALSE);
    


    return TRUE;
}


/* ldb_remove_item : */
gint        
ldb_remove_item( LDB *ldb, 
		 LDBItemId id) {
    LDBBlock *block;

    g_return_val_if_fail( ldb != NULL, FALSE);
    g_return_val_if_fail( id.block_no != 0xffffffff, FALSE);
    g_return_val_if_fail( id.item_no >= FIRST_ITEM_ID, FALSE);

    block = &g_array_index( ldb->blocks, LDBBlock, id.block_no);

    g_return_val_if_fail( id.item_no < block->header.next_item_id, FALSE);

    /* 
       when closing or flushing the database, all
       items marked for deletion will be emptied.

       emptied items will be deleted the next time 
       the database is loaded.

       ( xxx blah items could be emptied when the db is closed !!!)

       This method was choosed to avoid
       that two calls to this function with the same "id"
       delete two different items.
       this may happen if "ids" of deleted blocks are reused.
       One disaventage is that if the number of items in this
       block plus the number of deleted items is geater than 65535
       than there'll be no more valid ids to insert new items
       (as the item number (id->item_no) is coded with a guint16).
       In this case new items will be inserted on a new block
       while there may be some free space on this block.

       */

    block->deleted_items = g_slist_prepend( block->deleted_items, 
					    GUINT_TO_POINTER( (guint) id.item_no));
    
#ifdef DEBUG_ITEM
    printf( "LDB: Marking item %5d-%-4d for deletion\n", id.block_no, id.item_no);
    fflush( stdout);
#endif

    return TRUE;
}



gint        
ldb_dump( LDB *ldb) {

    return TRUE;
}



static void
ldb_error( LDB *ldb) {
    printf( "*** LDB: ERROR: %s\n", strerror( errno));
    fflush( stdout);
}


static void
empty_item( LDB *ldb, LDBBlock *block, guint16 item) {
    gchar *pos;
    guint16 item_no, item_size;
    gint32 size;

    g_return_if_fail( item < block->header.next_item_id);
#ifdef DEBUG_ITEM
	printf( "Truncating size of item  %5d-%-4d\n", block->block_no, item);
	fflush( stdout);
#endif
    pos = block->data;

    size = ldb->header.block_size - block->header.free_space;

    for ( item_no = FIRST_ITEM_ID; item_no < item && size > 0; item_no++) {
	item_size = GUINT16BE( GUINT16( pos));
	pos += item_size;
	size -= item_size;
    }
    
    item_size = GUINT16BE( GUINT16( pos));
    size -= item_size;

    if ( item_size > 2) {
	GUINT16( pos) = GUINT16BE( 2);

	if ( size > 0)
	    memmove( pos + 2, pos + item_size, size);

	block->header.free_space += item_size - 2;	

	if ( !(block->flags & LDB_BLOCK_MODIFIED))
	    block->flags |= LDB_BLOCK_MODIFIED;

	

    } else if ( item_size == 2 && size >= 0)
	/* xxx This shouldn't happen unless this block is corrupted ??? */
	g_warning( "*** LDB: Item %5d-%-4d has already been deleted\n", block->block_no, item);
    else {
#ifdef DEBUG
	printf( "*** LDB: Item %5d-%-4d is either too big to fit in block or too small to be a valid item\n", 
		block->block_no, item);
#endif
	ldb_block_corrupted( ldb, block);
    }

}

static void
empty_deleted_items( LDB *ldb, LDBBlock *block) {
    GSList *list;

    list = block->deleted_items;
    while ( list) {
	empty_item( ldb, block, GPOINTER_TO_UINT( list->data));
	list = list->next;
    }
    if ( block->deleted_items) 
	g_slist_free( block->deleted_items);

    block->deleted_items = NULL;
}


static void
delete_empty_items( LDB *ldb, LDBBlock *block) {
    char *pos;
    guint16 item_size, items;
    gint32 size;

    g_return_if_fail( ldb && block && block->data);

#ifdef DEBUG_ITEM
    printf( "LDB: deleting empty items in block %5d\n", block->block_no);
    fflush( stdout);
#endif

    pos = block->data;
    size = ldb->header.block_size - block->header.free_space;

    items = block->header.next_item_id;
    while ( items-- > FIRST_ITEM_ID && size > 0) {
	item_size = GUINT16BE( GUINT16( pos));
	size -= item_size;
	if ( item_size == 2) { /* it's an empty item */
	    memmove( pos, pos + 2, size);
	    block->header.next_item_id--;
	    block->header.free_space += 2;

	    if ( !(block->flags & LDB_BLOCK_MODIFIED))
		block->flags |= LDB_BLOCK_MODIFIED;
	} else if ( item_size > 2)
	    pos  += item_size;
	else
	    size = -1;
    }

    if ( size < 0) {
#ifdef DEBUG
	printf( "*** LDB: Item %5d-%-4d is (%d bytes) too big too fit in block\n",
		block->block_no,
		block->header.next_item_id - ++items + FIRST_ITEM_ID,
		-size);
	fflush( stdout);
#endif
	ldb_block_corrupted( ldb, block);
    }
}
    
