/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <config.h>

#ifdef HAVE_LIBPNG

#include <stdio.h>
#include <errno.h>
#include <png.h>

#include "image.h"
#include "pngsup.h"

#define PNG_BYTES_TO_CHECK 8

static void             my_error_func( png_structp png, png_const_charp err);
static void             my_warning_func( png_structp png, png_const_charp err);

static void
my_error_func( png_structp png, png_const_charp err) {
    Image *image;

    image = png_get_error_ptr( png);

}

static void
my_warning_func( png_structp png, png_const_charp err) {
    Image *image;

    image = png_get_error_ptr( png);
}


void
register_png() {
    char *extensions[] = { "png", NULL};
    ImageFormat png_format = {
	extensions,
	is_png,
	png_read_header,
	png_read,
	NULL
    };

    image_format_register( &png_format);
}

gint
is_png( Image *image) {
    char buff[ 8];

    g_return_val_if_fail( image->preview->file != NULL, FALSE);

    if ( !fread( buff, 8, 1, image->preview->file)) 
	return FALSE;

    fseek( image->preview->file, 0L, SEEK_SET);

    return png_check_sig( buff, 8);
}

gint 
png_read_header( Image *image) {
    png_structp png;
    png_infop info;
    gint color_type;

    png = png_create_read_struct( PNG_LIBPNG_VER_STRING,
				  (void*) image,
				  my_error_func,
				  my_warning_func);

    if ( !png)
	return FALSE;
    
    info = png_create_info_struct( png);
    if ( !info) {
	png_destroy_read_struct(&png, NULL, NULL);
	return FALSE;
    }
    
    if ( setjmp(png->jmpbuf)) {
        png_destroy_read_struct( &png, &info, NULL);
        return FALSE;
    }
    
    
    png_init_io( png, image->preview->file);
    
    png_read_info( png, info);

    color_type = png_get_color_type( png, info);
    image_set_infos( image, 
		     PNG_IMAGE,
		     png_get_image_width( png, info),
		     png_get_image_height( png, info),
		     png_get_bit_depth( png, info) > 1,
		     color_type == PNG_COLOR_TYPE_RGB_ALPHA || color_type == PNG_COLOR_TYPE_GRAY_ALPHA);

    png_destroy_read_struct( &png, &info, NULL);

    return TRUE;
}    

gint
png_read( Image *image) {
    png_structp png;
    png_infop info;
    png_bytep row;
    gint color_type, bit_depth, y;

    png = png_create_read_struct( PNG_LIBPNG_VER_STRING,
				  (void*) image,
				  my_error_func,
				  my_warning_func);
    if ( !png)
	return FALSE;
    
    info = png_create_info_struct( png);
    if ( !info) {
	png_destroy_read_struct(&png, NULL, NULL);
	return FALSE;
    }
    
    if ( setjmp(png->jmpbuf)) {
        png_destroy_read_struct( &png, &info, NULL);
        return FALSE;
    }
    
    
    png_init_io( png, image->preview->file);
    png_set_invert_alpha( png);
    png_read_info( png, info);


    bit_depth  = png_get_bit_depth( png, info);
    color_type = png_get_color_type( png, info);
    
    
    if ( color_type == PNG_COLOR_TYPE_PALETTE)
	png_set_expand( png);
    
    if ( color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
	png_set_expand( png);
    
    png_set_strip_16( png);
    png_set_packing( png);

    if ( png_get_valid( png, info, PNG_INFO_tRNS)) 
	png_set_expand( png);


    png_read_update_info( png, info);
    color_type = png_get_color_type( png, info);
    bit_depth  = png_get_bit_depth( png, info);
    image_set_infos( image,
		     PNG_IMAGE,
		     png_get_image_width( png, info),
		     png_get_image_height( png, info),
		     color_type == PNG_COLOR_TYPE_RGB || color_type == PNG_COLOR_TYPE_RGB_ALPHA,
		     color_type == PNG_COLOR_TYPE_RGB_ALPHA || color_type == PNG_COLOR_TYPE_GRAY_ALPHA);


    switch( color_type) {
    case PNG_COLOR_TYPE_GRAY:
	image_render_init( image, image->width, image->height, IMAGE_COLOR_8);
	break;
    case PNG_COLOR_TYPE_GRAY_ALPHA:
	image_render_init( image, image->width, image->height, IMAGE_COLOR_8_ALPHA);
	break;
    case PNG_COLOR_TYPE_RGB:
	image_render_init( image, image->width, image->height, IMAGE_COLOR_24);
	break;
    case PNG_COLOR_TYPE_RGB_ALPHA:
	image_render_init( image, image->width, image->height, IMAGE_COLOR_24_ALPHA);
	break;
	
    default:
	g_warning( "PNG : Unsupported color type");
	return FALSE;
	
    }

    row = g_malloc( png_get_rowbytes( png, info));

    for ( y = 0; y < image->height; y++) {
	png_read_rows( png, &row, NULL, 1);
	if ( !image_append_rows( image,
				 1,
				 row)) {
	    png_destroy_read_struct( &png, &info, NULL);
	    return FALSE;
	}
    }

    g_free( row);

    png_read_end( png, info);

    png_destroy_read_struct( &png, &info, NULL);

    return TRUE;
}


#endif

