/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* UBox is an adaptation of GtkCList with the following features :
       - An ubox contains a collection of items which have the same size
       - In an Ubox, the number of columns is
       not fixed. The number of columns depends of the width allocated
       to the widget such that all columns are visible.
   */

#include <stdio.h>
#include <math.h>
#include <gtk/gtk.h>
#include "ubox.h"

#define SCROLLBAR_SPACING 0
#define SCROLL_TIME 100

#define CELL_ROW( ubox, cell) ( (ubox)->cells_per_row ? (cell)/(ubox)->cells_per_row : 0)
#define CELL_COLUMN( ubox, cell) ( (ubox)->cells_per_row ? (cell)%(ubox)->cells_per_row : 0)


#define ROW_TOP( ubox, row) ( (ubox)->cell_height*(row) + (ubox)->voffset)
#define COLUMN_RIGHT( ubox, column) ( (ubox)->cell_width*(column) + (ubox)->hoffset)

#define UBOX_WIDTH( ubox) ( (ubox)->cells_per_row * (ubox)->cell_width)
#define UBOX_HEIGHT( ubox) ( (int) ceil( (double) (ubox)->child_count / (ubox)->cells_per_row) * (ubox)->cell_height)

#define NEAREST_MODULO_X( val, x) ( (val / x) * x)

static void         ubox_destroy( GtkObject *object);
static void         ubox_finalize( GtkObject *object);

static void         ubox_class_init( UBoxClass *klass);
static void         ubox_init( UBox *ubox);
static void         ubox_map( GtkWidget *widget);
static void         ubox_unmap( GtkWidget *widget);
static void         ubox_draw( GtkWidget *widget, GdkRectangle *area);

static void         ubox_realize( GtkWidget *widget);
static void         ubox_unrealize( GtkWidget *widget);
static gint         ubox_expose( GtkWidget *widget, GdkEventExpose *event);
static void         ubox_size_request( GtkWidget *widget, 
					   GtkRequisition  *requisition);
static void         ubox_size_allocate( GtkWidget *widget,
					    GtkAllocation  *allocation);
static gint         ubox_button_press( GtkWidget *widget,
					   GdkEventButton *event);
static gint         ubox_button_release( GtkWidget *widget,
					 GdkEventButton *event);
static gint         horizontal_timeout( UBox *ubox);
static gint         vertical_timeout( UBox *ubox);
static gint         ubox_motion( GtkWidget *widget, GdkEventMotion *event);

static void         set_anchor( UBox *ubox, gint anchor);
static void         move_horizontal( UBox *ubox, gint diff);
static void         move_vertical( UBox *ubox, gint diff);

static void         ubox_draw_focus( GtkWidget *widget);
static void         ubox_focus_cell( UBox *ubox, gint cell_no);
static void         update_extended_selection( UBox *ubox, gint cell_no);


static void         ubox_real_select_cell( UBox *ubox, gint cell_no);
static void         ubox_real_unselect_cell( UBox *ubox, gint cell_no);

static void         create_scrollbars( UBox *ubox);
static void         vadjustment_changed( GtkAdjustment *adjustment, gpointer data);
static void         hadjustment_changed( GtkAdjustment *adjustment, gpointer data);
static void         vadjustment_value_changed( GtkAdjustment *adjustment, gpointer data);
static void         hadjustment_value_changed( GtkAdjustment *adjustment, gpointer data);
static void         adjust_scrollbars( UBox *ubox);
static gint         cell_from_coord( UBox *ubox, gint x, gint y);
static gint         closest_cell_from_coord( UBox *ubox, gint x, gint y);

enum {
    DRAW_CELL,
    SELECT_CELL,
    UNSELECT_CELL,
    LAST_SIGNAL
};

static GtkContainerClass *parent_class;
static guint ubox_signals[ LAST_SIGNAL] = { 0};


GtkType
ubox_get_type (void)
{
  static GtkType ubox_type = 0;

  if (!ubox_type)
    {
      GtkTypeInfo ubox_info =
      {
	"UBox",
	sizeof (UBox),
	sizeof (UBoxClass),
	(GtkClassInitFunc) ubox_class_init,
	(GtkObjectInitFunc) ubox_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL,
      };

      ubox_type = gtk_type_unique (GTK_TYPE_CONTAINER, &ubox_info);
    }

  return ubox_type;
}


static void         
ubox_class_init( UBoxClass *klass) {
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;
    GtkContainerClass *container_class;
    
    object_class = (GtkObjectClass *) klass;
    widget_class = (GtkWidgetClass *) klass;
    container_class = (GtkContainerClass *) klass;
    
    parent_class = gtk_type_class( gtk_container_get_type());

    object_class->destroy = ubox_destroy;
    object_class->finalize = ubox_finalize;
    
    ubox_signals[ DRAW_CELL] = 
	gtk_signal_new( "draw_cell",
			GTK_RUN_FIRST,
			object_class->type,
			GTK_SIGNAL_OFFSET( UBoxClass, draw_cell),
			gtk_marshal_NONE__POINTER_POINTER_POINTER,
			GTK_TYPE_NONE, 3,
			GTK_TYPE_POINTER, GTK_TYPE_POINTER, GTK_TYPE_POINTER);
    ubox_signals[ SELECT_CELL] = 
	gtk_signal_new( "select_cell",
			GTK_RUN_FIRST,
			object_class->type,
			GTK_SIGNAL_OFFSET( UBoxClass, select_cell),
			gtk_marshal_NONE__INT,
			GTK_TYPE_NONE, 1,
			GTK_TYPE_INT);
    ubox_signals[ UNSELECT_CELL] = 
	gtk_signal_new( "unselect_cell",
			GTK_RUN_FIRST,
			object_class->type,
			GTK_SIGNAL_OFFSET( UBoxClass, unselect_cell),
			gtk_marshal_NONE__INT,
			GTK_TYPE_NONE, 1,
			GTK_TYPE_INT);
    gtk_object_class_add_signals( object_class, ubox_signals, LAST_SIGNAL);

    widget_class->map   = ubox_map;
    widget_class->unmap = ubox_unmap;
    widget_class->draw  = ubox_draw;
    widget_class->realize = ubox_realize;
    widget_class->unrealize = ubox_unrealize;
    widget_class->expose_event  = ubox_expose;
    widget_class->size_request  = ubox_size_request;
    widget_class->size_allocate = ubox_size_allocate;
    widget_class->button_press_event = ubox_button_press;
    widget_class->button_release_event = ubox_button_release;
    widget_class->motion_notify_event = ubox_motion;

    klass->draw_cell = NULL;
    klass->select_cell = ubox_real_select_cell;
    klass->unselect_cell = ubox_real_unselect_cell;
}

static void
ubox_init( UBox *ubox) {
    GTK_WIDGET_UNSET_FLAGS( ubox, GTK_NO_WINDOW);
    
    ubox->flags  = 0;

    ubox->cells_area = NULL;
    ubox->cells_area_width = 0;
    ubox->cells_area_height = 0;

    ubox->children = NULL;
    ubox->child_count = 0;
    ubox->cell_width = 0;
    ubox->cell_height = 0;
 
    ubox->hoffset = 0;
    ubox->voffset = 0;

    ubox->hscrollbar = NULL;
    ubox->vscrollbar = NULL;
    ubox->hscrollbar_policy = GTK_POLICY_AUTOMATIC;
    ubox->vscrollbar_policy = GTK_POLICY_AUTOMATIC;

    ubox->selection_mode = GTK_SELECTION_SINGLE;
    ubox->selection = NULL;

    ubox->focus_cell = -1;
 
    ubox->anchor = -1;
    ubox->drag_pos = -1;
    ubox->focus_cell = -1;
    ubox->htimer = 0;
    ubox->vtimer = 0;
}

GtkWidget*
ubox_new( gint cell_width,
	      gint cell_height) {
    UBox *ubox;

    g_return_val_if_fail( cell_width > 0, NULL);
    g_return_val_if_fail( cell_height > 0, NULL);

    ubox = gtk_type_new( ubox_get_type());

    
    ubox->cell_width = cell_width;
    ubox->cell_height = cell_height;

    create_scrollbars( ubox);

    return GTK_WIDGET( ubox);
}
    
static void
ubox_destroy( GtkObject *object) {
    UBox *ubox;

    g_return_if_fail( object != NULL);
    g_return_if_fail( IS_UBOX( object));

    ubox = UBOX( object);

    if ( ubox->vscrollbar) {
	gtk_widget_unparent( ubox->vscrollbar);
	ubox->vscrollbar = NULL;
    }
    if ( ubox->hscrollbar) {
	gtk_widget_unparent( ubox->hscrollbar);
	ubox->hscrollbar = NULL;
    }

    if ( GTK_OBJECT_CLASS( parent_class)->destroy)
	( *GTK_OBJECT_CLASS( parent_class)->destroy) (object);
}

static void
ubox_finalize( GtkObject *object) {
    UBox *ubox;

    g_return_if_fail( object != NULL);
    g_return_if_fail( IS_UBOX( object));

    ubox = UBOX( object);

    if ( ubox->children)
	g_slist_free( ubox->children);
    if ( ubox->selection)
	g_slist_free( ubox->selection);

    if ( GTK_OBJECT_CLASS( parent_class)->finalize)
	( *GTK_OBJECT_CLASS( parent_class)->finalize) (object);
}


static void         
ubox_map( GtkWidget *widget) {
    UBox *ubox;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_UBOX( widget));

    ubox = UBOX( widget);
    GTK_WIDGET_SET_FLAGS( widget, GTK_MAPPED);

    gdk_window_show( widget->window);
    gdk_window_show( ubox->cells_area);
    
    if (GTK_WIDGET_VISIBLE( ubox->vscrollbar) &&
	!GTK_WIDGET_MAPPED( ubox->vscrollbar))
	gtk_widget_map( ubox->vscrollbar);
    
    if (GTK_WIDGET_VISIBLE( ubox->hscrollbar) &&
	!GTK_WIDGET_MAPPED( ubox->hscrollbar))
	gtk_widget_map( ubox->hscrollbar);

}

static void         
ubox_unmap( GtkWidget *widget) {
    UBox *ubox;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_UBOX( widget));

    ubox = UBOX( widget);
    GTK_WIDGET_UNSET_FLAGS ( widget, GTK_MAPPED);

    gdk_window_hide( widget->window);
    gdk_window_hide( ubox->cells_area);
    
    if ( GTK_WIDGET_MAPPED( ubox->vscrollbar))
	gtk_widget_unmap( ubox->vscrollbar);

    if ( GTK_WIDGET_MAPPED( ubox->hscrollbar))
	gtk_widget_unmap( ubox->hscrollbar);
    

}

static void
ubox_draw( GtkWidget *widget,
	       GdkRectangle *area) {
    UBox *ubox;
    UBoxCell *child;
    GdkRectangle child_area;
    GdkRectangle cells_area;
    GdkRectangle intersect_area;
    GSList *list;
    gint first_cell, last_cell, column, row, i;
    gint width, height;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_UBOX( widget));

    if ( !GTK_WIDGET_DRAWABLE( widget))
	return;

    ubox = UBOX( widget);

    gdk_window_get_size( ubox->cells_area, &width, &height);
    cells_area.x = 0;
    cells_area.y = 0;
    cells_area.width = width;
    cells_area.height = height;
	
    
    if ( !area)
	area = &cells_area;

 
    if ( !gdk_rectangle_intersect( area, &cells_area, &intersect_area))
	return;
    
    
    gdk_window_clear_area( ubox->cells_area,
			   intersect_area.x,
			   intersect_area.y,
			   intersect_area.width,
			   intersect_area.height);
			   
			   
    last_cell  = closest_cell_from_coord( ubox, 
					  area->x + area->width,
					  area->y + area->height);

    
    first_cell  = closest_cell_from_coord( ubox,
					   area->x,
					   area->y);


    column = CELL_COLUMN( ubox, first_cell);
    row    = CELL_ROW( ubox, first_cell);
    
    i = last_cell - first_cell + 1;

    list = g_slist_nth( ubox->children, first_cell);
    while ( list && i--) {
	
	child = list->data;
	list = list->next;

	child_area.y = ROW_TOP( ubox, row); 
	child_area.x = COLUMN_RIGHT( ubox, column);
	child_area.width = ubox->cell_width;
	child_area.height = ubox->cell_height;
	
	if ( gdk_rectangle_intersect( area, &child_area, &intersect_area)) {
			     
	    if ( gdk_rectangle_intersect( area, &child_area, &intersect_area)) 
		gtk_signal_emit( GTK_OBJECT( widget), ubox_signals[ DRAW_CELL],
				 child, &child_area, &intersect_area);
	} 

	column++;
	if ( column == ubox->cells_per_row) {
	    column = 0;
	    row++;
	} 
	
    }
    

}


static void
ubox_redraw_cell( UBox *ubox,
		  gint cell_no) {
    GdkRectangle child_area;	
    gint row, column;

    adjust_scrollbars( ubox);
    
    row = CELL_ROW( ubox, cell_no);
    column = CELL_COLUMN( ubox, cell_no);
    
    child_area.x = COLUMN_RIGHT( ubox, column);
    child_area.y = ROW_TOP( ubox, row);
    child_area.width = ubox->cell_width;
    child_area.height = ubox->cell_height;
    
	
    ubox_draw( GTK_WIDGET( ubox), &child_area);
}

static void         
ubox_realize( GtkWidget *widget) {
    UBox *ubox;
    GdkWindowAttr attributes;
    gint attributes_mask;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_UBOX( widget));

    GTK_WIDGET_SET_FLAGS( widget, GTK_REALIZED);

    ubox = UBOX( widget);

    /* widget window */
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual( widget);
    attributes.colormap = gtk_widget_get_colormap( widget);
    attributes.event_mask = (GDK_EXPOSURE_MASK |
			     GDK_BUTTON_PRESS_MASK |
			     GDK_BUTTON_RELEASE_MASK |
			     GDK_KEY_PRESS_MASK |
			     GDK_KEY_RELEASE_MASK);
    
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    
    widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), 
				     &attributes, attributes_mask);
    gdk_window_set_user_data (widget->window, widget);

    widget->style = gtk_style_attach (widget->style, widget->window);
    gtk_style_set_background( widget->style, widget->window, GTK_STATE_NORMAL);

    /* ubox window */
    attributes.x = GTK_CONTAINER( widget)->border_width;
    attributes.y = GTK_CONTAINER( widget)->border_width;
    attributes.width  -= 2*GTK_CONTAINER( widget)->border_width;
    if ( GTK_WIDGET_VISIBLE( ubox->vscrollbar))
	attributes.width  -= ubox->vscrollbar->allocation.width - SCROLLBAR_SPACING;
    attributes.height -= 2*GTK_CONTAINER( widget)->border_width;
    if ( GTK_WIDGET_VISIBLE( ubox->hscrollbar))
	attributes.height  -= ubox->hscrollbar->allocation.height - SCROLLBAR_SPACING;
    ubox->cells_area = gdk_window_new ( widget->window, 
					&attributes, attributes_mask);
    
    gdk_window_set_user_data ( ubox->cells_area, widget);
    gdk_window_set_background ( ubox->cells_area, 
				&widget->style->bg[GTK_STATE_NORMAL]);
    gdk_window_show( ubox->cells_area);
    
    ubox->cells_area_width = attributes.width;
    ubox->cells_area_height = attributes.height;
}

static void
ubox_unrealize( GtkWidget *widget) {
    UBox *ubox;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_UBOX( widget));

    ubox = UBOX( widget);

    gdk_window_set_user_data( ubox->cells_area, NULL);
    gdk_window_destroy( ubox->cells_area);
    ubox->cells_area = NULL;

    if (GTK_WIDGET_CLASS (parent_class)->unrealize)
	(* GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
}


static gint
ubox_expose( GtkWidget	*widget,
		 GdkEventExpose *event) {
    UBox *ubox;

    g_return_val_if_fail( widget != NULL, FALSE);
    g_return_val_if_fail( IS_UBOX( widget), FALSE);
  
    ubox = UBOX( widget);

    if ( GTK_WIDGET_DRAWABLE( widget)) {
   
	if ( event->window == ubox->cells_area)
	    ubox_draw( widget, &event->area);

    }

    return FALSE;
}

static void
ubox_size_request( GtkWidget       *widget,
		   GtkRequisition  *requisition) {

    UBox *ubox;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_UBOX( widget));
    g_return_if_fail( requisition != NULL);

    ubox = UBOX( widget);
    requisition->width = ubox->cell_width;
    requisition->height = ubox->cell_height;

    /* add the vscrollbar space */
    if (( ubox->vscrollbar_policy == GTK_POLICY_AUTOMATIC) ||
	GTK_WIDGET_VISIBLE ( ubox->vscrollbar)) {
	gtk_widget_size_request ( ubox->vscrollbar, 
				  &ubox->vscrollbar->requisition);
	
	requisition->width += ubox->vscrollbar->requisition.width + 
	    SCROLLBAR_SPACING;
	requisition->height = MAX (requisition->height,
				   ubox->vscrollbar->requisition.height);
    }
    
    /* add the hscrollbar space */
    if (( ubox->hscrollbar_policy == GTK_POLICY_AUTOMATIC) ||
	GTK_WIDGET_VISIBLE ( ubox->hscrollbar)) {
	gtk_widget_size_request ( ubox->hscrollbar, 
				  &ubox->hscrollbar->requisition);
	
	requisition->height += ubox->hscrollbar->requisition.height + 
	    SCROLLBAR_SPACING;
	requisition->width = MAX (requisition->width,
				  ubox->hscrollbar->requisition.width);/*, 
				  requisition->width - 
				  ubox->vscrollbar->requisition.width);*/
	
    }
    
    requisition->width += widget->style->klass->xthickness * 2 +
	GTK_CONTAINER (widget)->border_width * 2;
    requisition->height += widget->style->klass->ythickness * 2 +
	GTK_CONTAINER (widget)->border_width * 2;
}


static void
ubox_size_allocate( GtkWidget      *widget,
			GtkAllocation  *allocation) {
    UBox *ubox;
    GtkAllocation widgets_allocation;
    GtkAllocation child_allocation;
    gint hscrollbar_vis = TRUE, vscrollbar_vis = TRUE, i;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_UBOX( widget));
    g_return_if_fail( allocation != NULL);

    ubox = UBOX( widget);
    widget->allocation = *allocation;

    if ( GTK_WIDGET_REALIZED( widget))
	gdk_window_move_resize (widget->window,
				allocation->x + GTK_CONTAINER (widget)->border_width,
				allocation->y + GTK_CONTAINER (widget)->border_width,
				allocation->width - 2*GTK_CONTAINER (widget)->border_width, 
				allocation->height - 2*GTK_CONTAINER (widget)->border_width);

    
    widgets_allocation.x = GTK_CONTAINER (widget)->border_width;
    widgets_allocation.y = GTK_CONTAINER (widget)->border_width;
    widgets_allocation.width = allocation->width - 
	ubox->vscrollbar->requisition.width - SCROLLBAR_SPACING - 
	2 * GTK_CONTAINER( widget)->border_width;
    widgets_allocation.height = allocation->height - 
	ubox->hscrollbar->requisition.height - SCROLLBAR_SPACING - 
	2 * GTK_CONTAINER( widget)->border_width;

    if ( widgets_allocation.width < 1)
	widgets_allocation.width = 1;
    if ( widgets_allocation.height < 1)
	widgets_allocation.height = 1;

    
    /* allocate scrollbars areas and ajust its value */
    ubox->cells_per_row = widgets_allocation.width / ubox->cell_width;
    if ( !ubox->cells_per_row)
	ubox->cells_per_row = 1;

    for ( i = 0; i < 2; i++) {
	if ( vscrollbar_vis) {
	    vscrollbar_vis = widgets_allocation.height < UBOX_HEIGHT( ubox);
	    if ( !vscrollbar_vis) {
		widgets_allocation.width += ubox->vscrollbar->requisition.width + SCROLLBAR_SPACING;
		ubox->cells_per_row = widgets_allocation.width / ubox->cell_width;
		if ( !ubox->cells_per_row)
		    ubox->cells_per_row = 1;
	    }
	}

	if ( hscrollbar_vis) {
	    hscrollbar_vis = widgets_allocation.width < UBOX_WIDTH( ubox);
	    if ( !hscrollbar_vis)
		widgets_allocation.height += ubox->hscrollbar->requisition.height + SCROLLBAR_SPACING;
	}
    }

    if ( GTK_WIDGET_REALIZED( widget))
	gdk_window_move_resize( ubox->cells_area,
				widgets_allocation.x,
				widgets_allocation.y,
				widgets_allocation.width,
				widgets_allocation.height);

    if ( vscrollbar_vis) {

	if ( !GTK_WIDGET_VISIBLE( ubox->vscrollbar))
	    gtk_widget_show( ubox->vscrollbar);

	child_allocation.x = //allocation->x +
	    allocation->width -
	    ubox->vscrollbar->requisition.width -
	    GTK_CONTAINER( widget)->border_width;
	child_allocation.y = //allocation->y +
	    GTK_CONTAINER( widget)->border_width;
	child_allocation.width  = ubox->vscrollbar->requisition.width;
	child_allocation.height = allocation->height - 
	    SCROLLBAR_SPACING - 2*GTK_CONTAINER( widget)->border_width;
	if ( hscrollbar_vis)
	    child_allocation.height -= ubox->hscrollbar->requisition.height;

	gtk_widget_size_allocate( ubox->vscrollbar, &child_allocation);
	
    } else if ( GTK_WIDGET_VISIBLE( ubox->vscrollbar))
	    gtk_widget_hide( ubox->vscrollbar);
    
    if ( hscrollbar_vis) {
	if ( !GTK_WIDGET_VISIBLE( ubox->hscrollbar))
	    gtk_widget_show( ubox->hscrollbar);

	child_allocation.x = //allocation->x +
	    GTK_CONTAINER( widget)->border_width;
	child_allocation.y = allocation->y +
	    allocation->height - 
	    ubox->hscrollbar->requisition.height +
	    GTK_CONTAINER( widget)->border_width;
	child_allocation.width = allocation->width - 
	    SCROLLBAR_SPACING - 2*GTK_CONTAINER( widget)->border_width;
	child_allocation.height = ubox->hscrollbar->requisition.height;

	if ( vscrollbar_vis)
	    child_allocation.width -= ubox->vscrollbar->requisition.width;
	
	gtk_widget_size_allocate( ubox->hscrollbar, &child_allocation);
    
    } else if ( GTK_WIDGET_VISIBLE( ubox->hscrollbar))
	    gtk_widget_hide( ubox->hscrollbar);
    
    ubox->cells_area_width = widgets_allocation.width;
    ubox->cells_area_height = widgets_allocation.height;

    adjust_scrollbars( ubox);

    ubox_draw( widget, NULL);
}
    
static gint         
ubox_button_press( GtkWidget *widget,
		   GdkEventButton *event) {
    UBox *ubox;
    gint cell, old_cell;

    g_return_val_if_fail( widget != NULL, FALSE);
    g_return_val_if_fail( IS_UBOX( widget), FALSE);
    g_return_val_if_fail( event != NULL, FALSE);

    if ( event->button != 1)
	return FALSE;

    ubox = UBOX( widget);
    
    if ( event->window == ubox->cells_area) {
	old_cell = ubox->focus_cell;

	cell = ubox_cell_from_coord( ubox, event->x, event->y);
	if ( cell == -1)
	    return FALSE;

	if (event->type == GDK_BUTTON_PRESS) {
	    UBOX_SET_FLAG( ubox, UBOX_DRAG_SELECTION);
	    gdk_pointer_grab( ubox->cells_area, FALSE,
			      GDK_POINTER_MOTION_HINT_MASK |
			      GDK_BUTTON1_MOTION_MASK |
			      GDK_BUTTON_RELEASE_MASK,
			      NULL, NULL, event->time);
	    gtk_grab_add( widget);

	} else if ( gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (widget)) {
	    UBOX_UNSET_FLAG( ubox, UBOX_DRAG_SELECTION);
	    gtk_grab_remove( widget);
	    gdk_pointer_ungrab( event->time);
	}
	
	if ( UBOX_FLAGS( ubox) & UBOX_ADD_MODE) {
	    UBOX_UNSET_FLAG( ubox, UBOX_ADD_MODE);
	    if ( GTK_WIDGET_HAS_FOCUS( widget)) {
		ubox_draw_focus( widget);
		//gdk_gc_set_line_attributes( ubox->xor_gc, 1,
					      //GDK_LINE_SOLID, 0, 0);
		ubox->focus_cell = cell;
		ubox_draw_focus( widget);
	    } else {
		//gdk_gc_set_line_attributes( ubox->xor_gc, 1,
					      //GDK_LINE_SOLID, 0, 0);
		ubox->focus_cell = cell;
	    }
	} else if (cell != ubox->focus_cell) {
	      if (GTK_WIDGET_HAS_FOCUS( widget)) {
		  ubox_draw_focus( widget);
		  ubox->focus_cell = cell;
		  ubox_draw_focus( widget);
		}
	      else
		ubox->focus_cell = cell;
	}
	
	switch ( ubox->selection_mode) {
	case GTK_SELECTION_SINGLE:
	case GTK_SELECTION_MULTIPLE:
	    if (event->type != GDK_BUTTON_PRESS)
		ubox_real_select_cell( ubox, cell);
	    else
		ubox->anchor = cell;
	    break;

	case GTK_SELECTION_BROWSE:
	    ubox_real_select_cell( ubox, cell);
	    break;
	    
	case GTK_SELECTION_EXTENDED:
	    if (event->type != GDK_BUTTON_PRESS) {
		/*if (ubox->anchor != -1) {
		    update_extended_selection( ubox, ubox->focus_cell);
		    }*/
		ubox->anchor = -1;
		ubox_real_select_cell( ubox, cell);
		break;
	    }
	    
	    if (event->state & GDK_CONTROL_MASK) {
		if (event->state & GDK_SHIFT_MASK) {
		    if ( ubox->anchor < 0)
			ubox->drag_pos = old_cell;

		    update_extended_selection (ubox, ubox->focus_cell);
		} else {
		    if (ubox->anchor == -1) {
			set_anchor( ubox, cell);
		    } else
			update_extended_selection( ubox, ubox->focus_cell);
		}
		break;
	    }
	    
	    if (event->state & GDK_SHIFT_MASK) {
		set_anchor( ubox, cell);
		ubox->drag_pos = -1;
		update_extended_selection( ubox, ubox->focus_cell);
		break;
	    }
	    
	    if (ubox->anchor == -1) {
		GSList *list;
		UBoxCell *c;
		gint i = 0;

		list = ubox->children;
		while ( list) {
		    c = list->data;
		    list = list->next;
		    if ( c->state == GTK_STATE_SELECTED)
			ubox_real_unselect_cell( ubox, i);
		    i++;
		}
		
		set_anchor( ubox, cell);
	    } else
		update_extended_selection( ubox, ubox->focus_cell);
	    break;
	    
        default:
	    break;
        }
    }
    
    return FALSE;
}

static gint         
ubox_button_release( GtkWidget *widget,
		     GdkEventButton *event) {
    UBox *ubox;
    UBoxCell *cell;
    GSList *item;
    gint cell_no;

    g_return_val_if_fail( widget != NULL, FALSE);
    g_return_val_if_fail( IS_UBOX( widget), FALSE);
    g_return_val_if_fail( event != NULL, FALSE);

    
    ubox = UBOX( widget);
    
    if ( event->button != 1)
	return FALSE;

    
    if ( UBOX_FLAGS( ubox) & UBOX_DRAG_SELECTION) {
	UBOX_UNSET_FLAG( ubox, UBOX_DRAG_SELECTION);

	gtk_grab_remove( widget);
	gdk_pointer_ungrab( event->time);
	
	if ( ubox->htimer) {
	    gtk_timeout_remove( ubox->htimer);
	    ubox->htimer = 0;
	}
	if ( ubox->vtimer) {
	    gtk_timeout_remove( ubox->vtimer);
	    ubox->vtimer = 0;
	}

	switch ( ubox->selection_mode) {
	case GTK_SELECTION_SINGLE:
	case GTK_SELECTION_BROWSE:
	    if ( (event->window != ubox->cells_area))
	    return FALSE;

	    item = g_slist_nth( ubox->children, cell_no);
	    if ( item)
		cell = item->data;
	    cell_no = cell_from_coord( ubox, (int) event->x, (int) event->y);
	    if ( cell_no == -1)
		return FALSE;
	    if ( ubox->selection && cell->state != GTK_STATE_SELECTED)
		gtk_signal_emit( GTK_OBJECT( widget), 
				 ubox_signals[ UNSELECT_CELL],
				 g_slist_index( ubox->children, ubox->selection->data));
	    if ( cell->state == GTK_STATE_SELECTED) 
		gtk_signal_emit( GTK_OBJECT( widget), 
				 ubox_signals[ UNSELECT_CELL],
				 cell_no);
	    else 
		gtk_signal_emit( GTK_OBJECT( widget), 
				 ubox_signals[ SELECT_CELL],
				 cell_no);
	    break;
	case GTK_SELECTION_MULTIPLE:
	case GTK_SELECTION_EXTENDED:
	    ubox->anchor = -1;
	    break;
	}
	
    }
    
    return FALSE;
}

static void
set_anchor( UBox *ubox, gint anchor) {
    UBoxCell *cell;
    
    cell = g_slist_nth_data( ubox->children, anchor);
    g_return_if_fail( cell != NULL);
    if ( cell->state == GTK_STATE_SELECTED)
	ubox_real_unselect_cell( ubox, anchor);
    else
	ubox_real_select_cell( ubox, anchor);

    ubox->anchor = anchor;
}


static gint
horizontal_timeout( UBox *ubox) {
    gint x, y;
    GdkEventMotion event;
    GdkModifierType mask;
    
    g_return_val_if_fail( IS_UBOX( ubox), FALSE);
    
    ubox->htimer = 0;
    gdk_window_get_pointer( ubox->cells_area, &x, &y, &mask);
    
    event.is_hint = 0;
    event.x = x;
    event.y = y;
    event.state = mask;
    
    ubox_motion( GTK_WIDGET( ubox), &event);
    
    return FALSE;
}

static gint
vertical_timeout( UBox *ubox) {
    gint x, y;
    GdkEventMotion event;
    GdkModifierType mask;
    
    g_return_val_if_fail( IS_UBOX( ubox), FALSE);
    
    ubox->vtimer = 0;
    gdk_window_get_pointer( ubox->cells_area, &x, &y, &mask);
    
    event.is_hint = 0;
    event.x = x;
    event.y = y;
    event.state = mask;
    
    ubox_motion( GTK_WIDGET( ubox), &event);
    
    return FALSE;
}

static gint         
ubox_motion( GtkWidget *widget, GdkEventMotion *event) {
    UBox *ubox;
    gint cell;

    g_return_val_if_fail( widget != NULL, FALSE);
    g_return_val_if_fail( IS_UBOX( widget), FALSE);
 
    ubox = UBOX( widget);

    if (!(gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB( widget)))
	return FALSE;

    if ( UBOX_WIDTH( ubox) > ubox->cells_area_width &&
	 (event->x < 0 || event->x > ubox->cells_area_width)) {
	
	if ( ubox->htimer)
	    return FALSE;

	ubox->htimer = gtk_timeout_add( SCROLL_TIME, (GtkFunction) horizontal_timeout, ubox);

	if ( !( event->x < 0 && GTK_RANGE( ubox->hscrollbar)->adjustment->value == 0) ||
	     ( event->x > ubox->cells_area_width && 
	       GTK_RANGE( ubox->hscrollbar)->adjustment->value == ubox->cells_area_width)) {
	    
	    if ( event->x < 0) 
		move_horizontal( ubox, -1 + event->x / 2);
	    else
		move_horizontal( ubox, 1 + (event->x - ubox->cells_area_width) / 2);
	}
    }

    if ( UBOX_HEIGHT( ubox) > ubox->cells_area_height &&
	 (event->y < 0 || event->y > ubox->cells_area_height)) {

	if ( ubox->vtimer)
	    return FALSE;

	ubox->vtimer = gtk_timeout_add( SCROLL_TIME, (GtkFunction) vertical_timeout, ubox);

	if ( !( event->y < 0 && GTK_RANGE( ubox->vscrollbar)->adjustment->value == 0) ||
	     ( event->y > ubox->cells_area_height && 
	       GTK_RANGE( ubox->vscrollbar)->adjustment->value == ubox->cells_area_height)) {
	    
	    if ( event->y < 0) 
		move_vertical( ubox, -1 + event->y / 2);
	    else
		move_vertical( ubox, 1 + (event->y - ubox->cells_area_height) / 2);
	}
    }
	
    cell = closest_cell_from_coord( ubox, event->x, event->y);

    if ( UBOX_FLAGS( ubox) & UBOX_DRAG_SELECTION) {
	if (cell == ubox->focus_cell)
	    return FALSE;
	
	ubox_draw_focus( widget);
	ubox->focus_cell = cell;
	ubox_draw_focus( widget);
	
      switch (ubox->selection_mode) {
      case GTK_SELECTION_BROWSE:
	  ubox_real_select_cell( ubox, ubox->focus_cell);
	  break;
	  
      case GTK_SELECTION_EXTENDED:
	  update_extended_selection( ubox, ubox->focus_cell);
	  break;
	  
      default:
	  break;
      }
    }
    

    return FALSE;
}

	
static void
move_horizontal( UBox *ubox,
		 gint diff) {
    gfloat upper;
    GtkAdjustment *adj;
    
    adj = GTK_RANGE( ubox->hscrollbar)->adjustment;
    
    adj->value += diff;
    
    upper = adj->upper - adj->page_size;
    adj->value = CLAMP( adj->value, 0.0, upper);

    gtk_signal_emit_by_name( GTK_OBJECT( adj), "value_changed");
}

static void
move_vertical( UBox *ubox,
		 gint diff) {
    gfloat upper;
    GtkAdjustment *adj;
    
    adj = GTK_RANGE( ubox->vscrollbar)->adjustment;
    
    adj->value += diff;
    
    upper = adj->upper - adj->page_size;
    adj->value = CLAMP( adj->value, 0.0, upper);

    gtk_signal_emit_by_name( GTK_OBJECT( adj), "value_changed");
}


static void
ubox_draw_focus( GtkWidget *widget) {
    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_UBOX( widget));

}

static void
ubox_focus_cell( UBox *ubox, gint cell_no) {
    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));
    g_return_if_fail( cell_no >= 0);


}

static void
update_extended_selection( UBox *ubox, gint cell_no) {
    UBoxCell *cell;
    GSList *list, *scell, *ecell;
    gint i, s, e, state;

    if ( cell_no < 0)
	return;

    if ( ubox->anchor < cell_no) {
	s = ubox->anchor;
	e = cell_no;
	scell = g_slist_nth( ubox->children, s);
	ecell = g_slist_nth( scell, e - s);
	if ( !scell || !ecell) 
	    return;

	state = ((UBoxCell*) scell->data)->state;
    } else {
	s = cell_no;
	e = ubox->anchor;
	scell = g_slist_nth( ubox->children, s);
	ecell = g_slist_nth( scell, e - s);
	if ( !scell || !ecell) 
	    return;

	state = ((UBoxCell*) ecell->data)->state;
    }
    

    if ( ubox->drag_pos < s) {
        list = g_slist_nth( ubox->children, ubox->drag_pos);
	i = ubox->drag_pos;
	while ( list && list != scell) {
	    cell = list->data;
	    list = list->next;
	    if ( cell->state == state) {
		if ( cell->state == GTK_STATE_NORMAL)
		    ubox_real_select_cell( ubox, i);
		else
		    ubox_real_unselect_cell( ubox, i);

	    }
	    i++;
	} 
    }

    if ( ubox->drag_pos > e) {
	GSList *drag_pos_cell;
	drag_pos_cell = g_slist_nth( ubox->children, ubox->drag_pos);
	if ( drag_pos_cell)
	    drag_pos_cell = drag_pos_cell->next;
	list = ecell;
	i = e;
	while ( list && list != drag_pos_cell) {
	    cell = list->data;
	    list = list->next;
	    if ( cell->state == state) {
		if ( cell->state == GTK_STATE_NORMAL)
		    ubox_real_select_cell( ubox, i);
		else
		    ubox_real_unselect_cell( ubox, i);

	    }
	    i++;
	}
    }

    list = scell;
    i = s;
    while ( list && list != ecell) {
	cell = list->data;
	list = list->next;
	if ( cell->state != state) {
	    if ( state == GTK_STATE_NORMAL)
		ubox_real_unselect_cell( ubox, i);
	    else
		ubox_real_select_cell( ubox, i);
	}
	i++;
    }
    if ( list) {
	cell = list->data;
	if ( cell->state != state) {
	    if ( state == GTK_STATE_NORMAL)
		ubox_real_unselect_cell( ubox, i);
	    else
		ubox_real_select_cell( ubox, i);
	}
    }
    
    ubox->drag_pos = cell_no;
}


UBoxCell*        
ubox_append( UBox *ubox, 
	     gpointer data) {
    UBoxCell *cell;

    g_return_val_if_fail( ubox != NULL, NULL);
    g_return_val_if_fail( IS_UBOX( ubox), NULL);
 
    cell = g_new( UBoxCell, 1);
    cell->data = data;
    cell->state = GTK_STATE_NORMAL;

    ubox->children = g_slist_append( ubox->children, cell);

    if ( GTK_WIDGET_VISIBLE( ubox))
	ubox_redraw_cell( ubox, ubox->child_count);
    
    ubox->child_count++;

    return cell;
}

void         
ubox_remove( UBox *ubox,
	     UBoxCell *cell) {
    GSList *child_item;
    
    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));

    child_item = g_slist_find( ubox->children, cell);
    g_return_if_fail( child_item != NULL);

    cell = child_item->data;

    ubox->children = g_slist_remove_link( ubox->children, child_item);
 
    ubox->selection = g_slist_remove( ubox->selection, cell);

    if ( GTK_WIDGET_VISIBLE( ubox)) {
	/* TODO : only redraw what's needed */
	ubox_draw( GTK_WIDGET( ubox), NULL);
    }

    ubox->child_count--;
}


static void         
ubox_real_select_cell( UBox *ubox, 
		       gint cell_no) {
    UBoxCell *cell;
    GSList *item;

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));
    g_return_if_fail( cell_no >= 0);  

    item = g_slist_nth( ubox->children, cell_no);
    if ( item) {
	cell = item->data;
	if ( !(cell->state == GTK_STATE_SELECTED)) {
	    ubox->selection = g_slist_prepend( ubox->selection, cell);
	    cell->state = GTK_STATE_SELECTED;
	    ubox_redraw_cell( ubox, cell_no);
	}
    }
	
}

static void         
ubox_real_unselect_cell( UBox *ubox, 
			 gint cell_no) {
    UBoxCell *cell;
    GSList *item;

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));
    g_return_if_fail( cell_no >= 0);

    item = g_slist_nth( ubox->children, cell_no);
    if ( item) {
	cell = item->data;
	if ( cell->state == GTK_STATE_SELECTED) {
	    ubox->selection = g_slist_remove( ubox->selection, cell);
	    cell->state = GTK_STATE_NORMAL;
	    ubox_redraw_cell( ubox, cell_no);
	}
    }

}

void
ubox_resize( UBox *ubox,
	     gint cell_width,
	     gint cell_height) {
    
    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));
    g_return_if_fail( cell_width > 0 && cell_height > 0);

    ubox->cell_width = cell_width;
    ubox->cell_height = cell_height;

    if ( GTK_WIDGET_VISIBLE( ubox))
	gtk_widget_queue_resize( GTK_WIDGET( ubox));
}

void
ubox_clear( UBox *ubox) {
    GSList *child_item;
    UBoxCell *cell;

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));

    child_item = ubox->children;
    while ( child_item) {
	cell = child_item->data;
	child_item = child_item->next;

	g_free( cell);
    }

    g_slist_free( ubox->children);
    ubox->children = NULL;
    ubox->child_count = 0;

    if ( ubox->selection) {
	g_slist_free( ubox->selection);
	ubox->selection = NULL;
    }

    if ( GTK_WIDGET_VISIBLE( ubox))
	ubox_draw( GTK_WIDGET( ubox), NULL);
}

void           
ubox_set_one_cell_per_row( UBox *ubox, gint one) {
    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));


    if ( one)
	ubox->flags |= UBOX_ONE_CELL_PER_ROW;
    else
	ubox->flags &= ~UBOX_ONE_CELL_PER_ROW;

    if ( GTK_WIDGET_VISIBLE( ubox))
	ubox_draw( GTK_WIDGET( ubox), NULL);
}

    
void           
ubox_select_cell( UBox *ubox, UBoxCell *cell) {
    gint cell_no;

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));
    g_return_if_fail( cell != NULL);
    
    cell_no = g_slist_index( ubox->children, cell);
    g_return_if_fail( cell_no >= 0);

    switch ( ubox->selection_mode) {
	case GTK_SELECTION_SINGLE:
	case GTK_SELECTION_BROWSE:
	    if ( ubox->selection)
		gtk_signal_emit( GTK_OBJECT( ubox), 
				 ubox_signals[ UNSELECT_CELL],
				 g_slist_index( ubox->children, ubox->selection->data));
	    break;
	case GTK_SELECTION_MULTIPLE:
	case GTK_SELECTION_EXTENDED:
	    break;
	default:
	    break;
    }
    
    ubox_moveto( ubox, cell);

    gtk_signal_emit( GTK_OBJECT( ubox), 
		     ubox_signals[ SELECT_CELL],
		     cell_no);
    
}

void           
ubox_unselect_cell( UBox *ubox, UBoxCell *cell) {
    gint cell_no;

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));
    g_return_if_fail( cell != NULL);
    
    cell_no = g_slist_index( ubox->children, cell);
    g_return_if_fail( cell_no >= 0);

    gtk_signal_emit( GTK_OBJECT( ubox), 
		     ubox_signals[ UNSELECT_CELL],
		     cell_no);
    
}

void           
ubox_moveto( UBox *ubox, UBoxCell *cell) {
    gint first_vis_cell, last_vis_cell, cell_no;

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));
    g_return_if_fail( cell != NULL);
    
    cell_no = g_slist_index( ubox->children, cell);
    g_return_if_fail( cell_no >= 0);

    first_vis_cell = closest_cell_from_coord( ubox, 0, 0);
    last_vis_cell = closest_cell_from_coord( ubox, ubox->cells_area_width, ubox->cells_area_height);
    
    if ( (cell_no <= --first_vis_cell) || (cell_no >= --last_vis_cell)) {
	GTK_RANGE( ubox->vscrollbar)->adjustment->value = CELL_ROW( ubox, cell_no) * ubox->cell_height;
	GTK_RANGE( ubox->hscrollbar)->adjustment->value = CELL_COLUMN( ubox, cell_no) * ubox->cell_width;
	
	gtk_signal_emit_by_name( GTK_OBJECT( GTK_RANGE( ubox->vscrollbar)->adjustment), 
				 "value_changed");
	gtk_signal_emit_by_name( GTK_OBJECT( GTK_RANGE( ubox->hscrollbar)->adjustment), 
				 "value_changed");
	
    }
}

GList*          
ubox_get_selection( UBox *ubox) {
    UBoxCell *cell;
    GList *list = NULL;
    GSList *sel;
    
    g_return_val_if_fail( ubox != NULL, NULL);
    g_return_val_if_fail( IS_UBOX( ubox) , NULL);
    
    sel = ubox->selection;
    while ( sel) {
	cell = sel->data;
	sel = sel->next;

	list = g_list_append( list, cell);
    }

    return list;
}

    
void           
ubox_set_selection_mode( UBox *ubox, GtkSelectionMode mode) {
    UBoxCell *cell;
    GSList *sel;

    g_return_if_fail( ubox != NULL);
    g_return_if_fail( IS_UBOX( ubox));

    if ( mode == ubox->selection_mode)
	return;

    switch( mode) {
    case GTK_SELECTION_MULTIPLE:
    case GTK_SELECTION_EXTENDED:
	break;
    case GTK_SELECTION_BROWSE:
    case GTK_SELECTION_SINGLE:
	sel = ubox->selection;
	while ( sel) {
	    cell = sel->data;
	    sel = sel->next;
	    
	    gtk_signal_emit( GTK_OBJECT( ubox), 
			     ubox_signals[ UNSELECT_CELL],
			     g_slist_index( ubox->children, cell));
	}
	break;
    }
    ubox->selection_mode = mode;
}

gint           
ubox_cell_from_coord( UBox *ubox, gint x, gint y) {
    g_return_val_if_fail( ubox != NULL, FALSE);
    g_return_val_if_fail( IS_UBOX( ubox), FALSE);

    return cell_from_coord( ubox, x, y);
}


gint           
ubox_get_cell_coord( UBox *ubox, gint cell_no, gint *x, gint *y) {
    gint top, right, bottom, left;

    g_return_val_if_fail( ubox != NULL, FALSE);
    g_return_val_if_fail( IS_UBOX( ubox), FALSE);
    g_return_val_if_fail( cell_no >= 0 && cell_no < ubox->child_count, FALSE);

    top = ROW_TOP( ubox, CELL_ROW( ubox, cell_no));
    right = COLUMN_RIGHT( ubox, CELL_COLUMN( ubox, cell_no));

    /* checks if a part of the cell is visible */
    if ( (top + ubox->cell_height < 0) || (top > ubox->cells_area_height))
	return FALSE;

    if ( (right + ubox->cell_width < 0) || (right > ubox->cells_area_width))
	return FALSE;

    /* arrange top & right so that there are >= 0 */
    bottom = MIN( top + ubox->cell_height, ubox->cells_area_height);
    top += (bottom - top) / 2;

    left = MIN( right + ubox->cell_width, ubox->cells_area_width);
    right += (left - right) / 2;

    *x = right;
    *y = top;

    return TRUE;
}

/******************
 *** Scrollbars ***
 ******************/


static void
create_scrollbars( UBox *ubox) {
      GtkAdjustment *adjustment;
      
      ubox->vscrollbar = gtk_vscrollbar_new( NULL);
      
      adjustment = gtk_range_get_adjustment( GTK_RANGE( ubox->vscrollbar));
      adjustment->step_increment = ubox->cell_height;
      gtk_signal_connect( GTK_OBJECT( adjustment), "changed",
			  (GtkSignalFunc) vadjustment_changed,
			  (gpointer) ubox);
      
      gtk_signal_connect( GTK_OBJECT( adjustment), "value_changed",
			  (GtkSignalFunc) vadjustment_value_changed,
			  (gpointer) ubox);
      
      gtk_widget_set_parent( ubox->vscrollbar, GTK_WIDGET( ubox));
      gtk_widget_show( ubox->vscrollbar);
      
      ubox->hscrollbar = gtk_hscrollbar_new( NULL);
      
      adjustment = gtk_range_get_adjustment( GTK_RANGE( ubox->hscrollbar));
      adjustment->step_increment = ubox->cell_width;

      gtk_signal_connect( GTK_OBJECT( adjustment), "changed",
			  (GtkSignalFunc) hadjustment_changed,
			  (gpointer) ubox);
      
      gtk_signal_connect( GTK_OBJECT( adjustment), "value_changed",
			  (GtkSignalFunc) hadjustment_value_changed,
			  (gpointer) ubox);
      
      gtk_widget_set_parent( ubox->hscrollbar, GTK_WIDGET( ubox));
      gtk_widget_show( ubox->hscrollbar);
}

static void
vadjustment_changed( GtkAdjustment *adjustment,
		     gpointer data) {
  UBox *ubox;

  g_return_if_fail( adjustment != NULL);
  g_return_if_fail( data != NULL);

  ubox = UBOX( data);
}

static void
hadjustment_changed( GtkAdjustment *adjustment,
		     gpointer data) {
  UBox *ubox;

  g_return_if_fail( adjustment != NULL);
  g_return_if_fail( data != NULL);

  ubox = UBOX( data);
}

static void
vadjustment_value_changed( GtkAdjustment *adjustment,
			   gpointer data) {
    UBox *ubox;
    GdkRectangle area;
    gint value, diff, width, height;

    g_return_if_fail( adjustment != NULL);
    g_return_if_fail( data != NULL);
    g_return_if_fail( IS_UBOX( data));
  
    if (!GTK_WIDGET_DRAWABLE( data))
	return;

    ubox = UBOX( data);
    g_return_if_fail( ubox->cells_area != NULL);

    width = ubox->cells_area_width;
    height = ubox->cells_area_height;

    value = adjustment->value;

    if ( adjustment == gtk_range_get_adjustment( GTK_RANGE( ubox->vscrollbar))) {
	if ( value > -ubox->voffset) {
	    /* scroll up */
	    
	    diff = value + ubox->voffset;
	    
	    if ( diff >= height) {
		ubox->voffset = -value;
		ubox_draw( GTK_WIDGET( ubox), NULL);
		return;
	    }
	    
	    if ( diff != 0)
		gdk_window_copy_area( ubox->cells_area,
				      GTK_WIDGET( ubox)->style->fg_gc[ GTK_STATE_NORMAL],
				      0, 0,
				      ubox->cells_area,
				      0, diff,
				      width,
				      height - diff);
	    area.x = 0;
	    area.y = height - diff;
	    area.width = width;
	    area.height = diff;
	    
	} else {
	    /* scroll down */
	    
	    diff = -ubox->voffset - value;
	    
	    if ( diff >= height) {
		ubox->voffset = -value;
		ubox_draw( GTK_WIDGET( ubox), NULL);
		return;
	    }
	    
	    if ( diff != 0) 
		gdk_window_copy_area( ubox->cells_area,
				      GTK_WIDGET( ubox)->style->fg_gc[ GTK_STATE_NORMAL],
				      0, diff,
				      ubox->cells_area,
				      0, 0,
				      width,
				      height - diff);
	    
	    
	    area.x = 0;
	    area.y = 0;
	    area.width = width;
	    area.height = diff;
	    
	    diff = -diff;
	}
	
	ubox->voffset = -value;
	ubox_draw( GTK_WIDGET( ubox), &area);
    }

}

static void
hadjustment_value_changed( GtkAdjustment * adjustment,
			   gpointer data) {
    UBox *ubox;
    GdkRectangle area;
    gint value, diff, width, height;

    g_return_if_fail( adjustment != NULL);
    g_return_if_fail( data != NULL);
    g_return_if_fail( IS_UBOX( data));
  
    if (!GTK_WIDGET_DRAWABLE( data))
	return;

    ubox = UBOX( data);
    g_return_if_fail( ubox->cells_area != NULL);

    width = ubox->cells_area_width;
    height = ubox->cells_area_height;

    value = adjustment->value;

    if ( adjustment == gtk_range_get_adjustment( GTK_RANGE( ubox->hscrollbar))) {
	if ( value > -ubox->hoffset) {
	    /* scroll left */
	    
	    diff = value + ubox->hoffset;
	    
	    if ( diff >= width) {
		ubox->hoffset = -value;
		ubox_draw( GTK_WIDGET( ubox), NULL);
		return;
	    }
	    
	    if ( diff != 0)
		gdk_window_copy_area( ubox->cells_area,
				      GTK_WIDGET( ubox)->style->fg_gc[ GTK_STATE_NORMAL],
				      0, 0,
				      ubox->cells_area,
				      diff, 0,
				      width - diff,
				      height);
	    area.x = width - diff;
	    area.y = 0;
	    area.width = diff;
	    area.height = height;
	    
	} else {
	    /* scroll rihgt */
	    
	    diff = -ubox->hoffset - value;
	    
	    if ( diff >= width) {
		ubox->hoffset = -value;
		ubox_draw( GTK_WIDGET( ubox), NULL);
		return;
	    }
	    
	    if ( diff != 0) 
		gdk_window_copy_area( ubox->cells_area,
				      GTK_WIDGET( ubox)->style->fg_gc[ GTK_STATE_NORMAL],
				      diff, 0,
				      ubox->cells_area,
				      0, 0,
				      width - diff,
				      height);
	    
	    
	    area.x = 0;
	    area.y = 0;
	    area.width = diff;
	    area.height = height;
	    
	    diff = -diff;
	}
	
	ubox->hoffset = -value;
	ubox_draw( GTK_WIDGET( ubox), &area);
    }

}

static void
adjust_scrollbars( UBox *ubox) {
 
    g_return_if_fail( ubox != NULL);
    
    if ( (ubox->cells_area == NULL) && (ubox->cells_area_width == 0))
	return;


    GTK_RANGE( ubox->vscrollbar)->adjustment->page_size = ubox->cells_area_height;
    GTK_RANGE( ubox->vscrollbar)->adjustment->page_increment = 
	NEAREST_MODULO_X( ubox->cells_area_height, ubox->cell_height);
    if ( !GTK_RANGE( ubox->vscrollbar)->adjustment->page_increment)
	GTK_RANGE( ubox->vscrollbar)->adjustment->page_increment = ubox->cell_height;
    GTK_RANGE( ubox->vscrollbar)->adjustment->lower = 0;
    GTK_RANGE( ubox->vscrollbar)->adjustment->upper = UBOX_HEIGHT( ubox);
   
    if ( GTK_WIDGET_VISIBLE( ubox->vscrollbar) &&
	 (ubox->cells_area_height - ubox->voffset > UBOX_HEIGHT( ubox))) {
	GTK_RANGE( ubox->vscrollbar)->adjustment->value = MAX( 0, UBOX_HEIGHT( ubox) - ubox->cells_area_height);
      gtk_signal_emit_by_name( GTK_OBJECT( GTK_RANGE( ubox->vscrollbar)->adjustment), 
			       "value_changed");
    }
    
    GTK_RANGE( ubox->hscrollbar)->adjustment->page_size = ubox->cells_area_width;
    GTK_RANGE( ubox->hscrollbar)->adjustment->page_increment = 
	NEAREST_MODULO_X( ubox->cells_area_width, ubox->cell_width);
    if ( !GTK_RANGE( ubox->hscrollbar)->adjustment->page_increment)
	GTK_RANGE( ubox->hscrollbar)->adjustment->page_increment = ubox->cell_width;
    GTK_RANGE( ubox->hscrollbar)->adjustment->lower = 0;
    GTK_RANGE( ubox->hscrollbar)->adjustment->upper = UBOX_WIDTH( ubox);
    
    if ( GTK_WIDGET_VISIBLE( ubox->hscrollbar) &&
	 (ubox->cells_area_width - ubox->hoffset > UBOX_WIDTH( ubox))) {
	GTK_RANGE( ubox->hscrollbar)->adjustment->value = MAX( 0, UBOX_WIDTH( ubox) - ubox->cells_area_width);
      gtk_signal_emit_by_name( GTK_OBJECT( GTK_RANGE( ubox->hscrollbar)->adjustment), 
			       "value_changed");
    }
    
    
    if ( UBOX_HEIGHT( ubox) <= ubox->cells_area_height &&
	 ubox->vscrollbar_policy == GTK_POLICY_AUTOMATIC) {
	
	if ( GTK_WIDGET_VISIBLE( ubox->vscrollbar)) {
	    gtk_widget_hide( ubox->vscrollbar);
	    gtk_widget_queue_resize( GTK_WIDGET( ubox));
	}
    } else {
	if ( ubox->cells_area_height < UBOX_HEIGHT( ubox) && !GTK_WIDGET_VISIBLE( ubox->vscrollbar)) {

	    gtk_widget_show( ubox->vscrollbar);
	    gtk_widget_queue_resize( GTK_WIDGET( ubox));
	}
    }
    
    if ( UBOX_WIDTH( ubox) <= ubox->cells_area_width &&
	 ubox->hscrollbar_policy == GTK_POLICY_AUTOMATIC) {
	if ( GTK_WIDGET_VISIBLE( ubox->hscrollbar)) {
	    gtk_widget_hide( ubox->hscrollbar);
	    gtk_widget_queue_resize( GTK_WIDGET( ubox));
	}
    } else {
	if (!GTK_WIDGET_VISIBLE( ubox->hscrollbar)) {
	    gtk_widget_show( ubox->hscrollbar);
	    gtk_widget_queue_resize( GTK_WIDGET( ubox));
	}
    }
    gtk_signal_emit_by_name( GTK_OBJECT( GTK_RANGE( ubox->hscrollbar)->adjustment), "changed");
    gtk_signal_emit_by_name( GTK_OBJECT( GTK_RANGE( ubox->vscrollbar)->adjustment), "changed");

    
}


static gint
cell_from_coord( UBox *ubox, gint x, gint y) {
    gint cell_no;

    g_return_val_if_fail( ubox != NULL, -1);

    if ( ubox->cells_per_row*ubox->cell_width < x) 
	cell_no = -1;
    else { 
	cell_no = ((y-ubox->voffset) / ubox->cell_height) * ubox->cells_per_row +
	    ((x-ubox->hoffset) / ubox->cell_width);
	if ( cell_no > ubox->child_count) 
	    cell_no = -1; 
    }     
    
    return cell_no;
}

static gint
closest_cell_from_coord( UBox *ubox, gint x, gint y) {
    gint cell_no;

    g_return_val_if_fail( ubox != NULL, -1);
 

    if ( ubox->cells_per_row*ubox->cell_width < x) {
	x = ubox->cells_per_row*ubox->cell_width;
	cell_no = ((y-ubox->voffset) / ubox->cell_height) * ubox->cells_per_row +
	    ((x-ubox->hoffset) / ubox->cell_width);
    } else 
	cell_no = ((y-ubox->voffset) / ubox->cell_height) * ubox->cells_per_row +
	    (x-ubox->hoffset) / ubox->cell_width;

    if ( cell_no > ubox->child_count)
	cell_no = ubox->child_count;

    return cell_no;
}






