/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <strings.h>
#include <glib.h>
#include "image.h"
#include "xpm.h"

/* xxx limits the size of each string */
#define XPM_BUFF_SIZE 65536

typedef enum {
    XPM_VALUES,
    XPM_COLORS,
    XPM_PIXELS,
    XPM_EXTENSIONS,
    XPM_START
} XpmSection;

typedef struct _CMapEntry {
    char *id;
    guint32 color;
} CMapEntry;

typedef struct _Xpm {
    Image *image;
    gint width, height, color_type, ncolors, cpp;
    CMapEntry *cmap;

    char buff[ XPM_BUFF_SIZE];
    gint buff_size;
} Xpm;

static gint        read_string( int fd, char *buff, gint size);
static long        read_long( char *nptr, char **endptr, int base);
static void        xpm_free( Xpm *xpm);

void
register_xpm() {
    char *extensions[] = { "xpm", NULL};
    ImageFormat xpm_format = {
	extensions, 
	is_xpm,
	xpm_read_header,
	xpm_read,
	NULL,
    };

    image_format_register( &xpm_format);
}

gint             
is_xpm( Image *image) {
    char buff[ 16];

    if ( !fread( buff, 1, 16, image->preview->file))
	return FALSE;
    
    fseek( image->preview->file, 0L, SEEK_SET);

    buff[9] = 0;
    return strcmp( buff, "/* XPM */") == 0;
}

/* xxx we run into troubles if a color spceification is bigger than that */
#define MAX_COLOR_SIZE 65536

static gint
_xpm_read_header( Xpm *xpm) {
    XpmSection section = XPM_START;
    char *string, *endptr;
    gint done = FALSE, color_no = 0;
    char color_spec[ MAX_COLOR_SIZE], i;

    while ( !done) {
	if ( !read_string( fileno( xpm->image->preview->file), xpm->buff, xpm->buff_size))
	    return FALSE;

	string = xpm->buff;
	if ( !strcmp( string, ","))
	    continue;

	switch ( section) {
	case XPM_START:
	    if ( !strcmp( string, "{"))
		section = XPM_VALUES;
	    break;
	case XPM_VALUES:
	    /* scan values */
	    endptr = string;
	    xpm->width = read_long( endptr, &endptr, 10);
	    xpm->height = read_long( endptr, &endptr, 10);
	    xpm->ncolors = read_long( endptr, &endptr, 10);
	    xpm->cpp = read_long( endptr, &endptr, 10);
	    if ( xpm->width < 0 || 
		 xpm->height < 0 || 
		 xpm->ncolors < 0 || 
		 xpm->cpp <= 0) {
		done = TRUE;
		break;
	    }

	    xpm->cmap = g_new( CMapEntry, xpm->ncolors);
	    if ( !xpm->cmap) {
		done = TRUE;
		break;
	    }
	    bzero( xpm->cmap, sizeof( CMapEntry) * xpm->ncolors);
	    section = XPM_COLORS;
	    break;
	case XPM_COLORS:
	    xpm->cmap[ color_no].id = g_malloc( xpm->cpp + 1);
	    xpm->cmap[ color_no].id[ xpm->cpp] = 0;

	    memcpy( xpm->cmap[ color_no].id, string, xpm->cpp);
	    
	    string += xpm->cpp + 1; i = 0;
	    string += strspn( string, " \t");
	    *color_spec = 0;
	    while ( *string) {
		
		if ( ((*string == 'c' || *string == 'm' || *string == 's') && string[1] == ' ') ||
		     ( *string == 'g' && string[1] == '4' && string[2] == ' ')) {
   
		    if ( *color_spec)
			break;

		    if ( *string == 'c' && string[1] == ' ') {
			/* xxx only scan 'c' colors */
			string++;
			string += strspn( string, " \t");
			
			i = strcspn( string, " \t");
			strncat( color_spec, string, i);

		    } else
			string += strspn( string, " \t");

		} else if ( *color_spec) {
		    i = strcspn( string, " \t");
		    strcat( color_spec, " "); 
		    strncat( color_spec, string, i);
		}
		string += strcspn( string, " \t");
		string += strspn( string, " \t");
	    }

	    if ( !*color_spec) {
		done = TRUE;
		break;
	    }

	    if ( !strcasecmp( color_spec, "none")) {
		xpm->color_type = IMAGE_COLOR_24_ALPHA;
		xpm->cmap[ color_no].color = 0xff000000;
		
	    } else {
		GdkColor c;
		
		if ( !gdk_color_parse( color_spec, &c)) {
		    g_warning( "xpm: can't parse color %s", color_spec);
		    
		    /*
		      done = TRUE;
		      break;
		      */
		} else 
		    xpm->cmap[ color_no].color = (c.red >> 8 & 0xff) | (c.green & 0xff00) | (c.blue << 8 & 0xff0000);
		
	    }
	    
	    color_no ++;
	    if ( color_no == xpm->ncolors) {
		section = XPM_PIXELS;
		done = TRUE;
	    }
	    break;
	case XPM_PIXELS:
	case XPM_EXTENSIONS:
	default:
	    done = TRUE;
	    break;
	}
    }
    

    if ( section == XPM_PIXELS) {
	image_set_infos( xpm->image,
			 XPM_IMAGE,
			 xpm->width,
			 xpm->height,
			 TRUE,
			 xpm->color_type == IMAGE_COLOR_24_ALPHA);
	return TRUE;
    }

    return FALSE;
}
  
gint 
xpm_read_header( Image *image) {
    Xpm xpm = { 0};
 
    xpm.image = image;
    xpm.color_type = IMAGE_COLOR_24;
    xpm.buff_size = XPM_BUFF_SIZE;
    
    if ( lseek( fileno( image->preview->file), 0, SEEK_SET) == -1)
	return FALSE;

    xpm_free( &xpm);

    return _xpm_read_header( &xpm);
}

static CMapEntry*
xpm_find_color( Xpm *xpm, char *id) {
    gint i;

    for ( i = 0; i < xpm->ncolors; i++)
	if ( !strncmp( xpm->cmap[ i].id, id, xpm->cpp))
	    return &xpm->cmap[ i];

    return NULL;
}

static void
xpm_free( Xpm *xpm) {
    gint i;

    for ( i = 0; i < xpm->ncolors; i++)
	if ( xpm->cmap[ i].id)
	    g_free( xpm->cmap[ i].id);

    g_free( xpm->cmap);

}

gint
xpm_read( Image *image) {
    Xpm xpm = { 0};
    guchar *line_buff;
    register guchar *dst, *src;
    register CMapEntry *ce;
    gint i, p, err = FALSE;

    xpm.image = image;
    xpm.color_type = IMAGE_COLOR_24;
    xpm.buff_size = XPM_BUFF_SIZE;
 
    if ( lseek( fileno( image->preview->file), 0, SEEK_SET) == -1)
	return FALSE;

    if ( !_xpm_read_header( &xpm)) {
	xpm_free( &xpm);
	return FALSE;
    }

    if ( xpm.color_type == IMAGE_COLOR_24) 
	line_buff = g_malloc( xpm.width * 3);
    else 
	line_buff = g_malloc( xpm.width * 4);


    if ( !line_buff)
	err = TRUE;

    image_render_init( image, xpm.width, xpm.height, xpm.color_type);

    for ( i = 0; i < xpm.height && !err;) {
	if ( !read_string( fileno( xpm.image->preview->file), xpm.buff, xpm.buff_size)) {
	    /* there's not enough scanline */
	    g_warning( "xpm: not enought scanlines");
	    err = TRUE;
	    break;
	}

	src = xpm.buff;

	if ( !strcmp( src, ","))
	    continue;

	i++;
	if ( strlen( src) < xpm.width * xpm.cpp) {
	    /* xpm string too short */
	    g_warning( "xpm: string too short");
	    err = TRUE;
	    break;
	}

	dst = line_buff;
	if ( xpm.color_type == IMAGE_COLOR_24) {
	    for ( p = 0; p < xpm.width; p++, src += xpm.cpp) {
		ce = xpm_find_color( &xpm, src);
		if ( !ce) {
		    err = TRUE;
		    g_warning( "xpm: color not found");
		    break;
		}
		dst[0] = ((char*) &ce->color)[0];
		dst[1] = ((char*) &ce->color)[1];
		dst[2] = ((char*) &ce->color)[2];
		dst += 3;
	    }
	} else {
	    for ( p = 0; p < xpm.width; p++, src += xpm.cpp) {
		ce = xpm_find_color( &xpm, src);
		if ( !ce) {
		    err = TRUE;
		    g_warning( "xpm: color not found");
		    break;
		}
		dst[0] = ((char*) &ce->color)[0];
		dst[1] = ((char*) &ce->color)[1];
		dst[2] = ((char*) &ce->color)[2];
		dst[3] = ((char*) &ce->color)[3];
		dst += 4;
	    }
	}
	if ( !image_append_rows( image, 1, line_buff)) {
	    err = TRUE;
	    break;
	}
    }

    xpm_free( &xpm);

    return err == FALSE;
}

static inline gint
read_char( gint fd, char *c) {
    register gint count;

    do {
	count = read( fd, c, 1);
    } while ( count == -1 &&
	      (errno == EINTR ||
	       errno == EAGAIN));

    if ( count <= 0)
	g_warning( "%s", strerror( errno));

    return count > 0;
}

    
static gint
read_string( int fd, char *buff, gint size) {
    register char *dest;
    char c;
    register gint s;
    gint sq = FALSE, dq = FALSE;

    dest = buff; s = size;
    
    while( read_char( fd, &c))
	if ( !(c == ' ' || c == '\t' || c == '\n' || c == '\r'))
	    break;
    do { 
	
	switch( c) {
	case '\"':
	    if ( dest != buff) {
		*dest = 0;
		   
		if ( !dq)
		    lseek( fd, -1, SEEK_CUR);
		
		return TRUE;
	    }
	    dq = TRUE;
	    break;
	case '\'':
	    if ( dq) {
		*dest++ = c; s--;
		break;
	    }

	    if ( dest != buff) {
		*dest = 0;
			
		if ( !sq)
		    lseek( fd, -1, SEEK_CUR);
		
		return TRUE;
	    }
	    sq = TRUE;
	    break;
	case '/':
	    *dest++ = c; s--;
	    if ( !read_char( fd, &c))
		return FALSE;
	    if ( c == '*') {
		register char prev = 0;

		dest--; s++;
		while ( read_char( fd, &c))
		    if ( c == '/' && prev == '*')
			break;
		    else
			prev = c;

		while( read_char( fd, &c))
		    if ( !(c == ' ' || c == '\t' || c == '\n' || c == '\r'))
			break;
		
	    } 
	    lseek( fd, -1, SEEK_CUR);

	    break;
	case ' ':
	case '\t':
	case '\n':
	case '\r':
	    if ( !sq && !dq) {
		*dest = 0;
		return TRUE;
	    }
	    *dest++ = c; s--;
	    break;
	default:
	    *dest++ = c; s--;
	}


    } while ( s && read_char( fd, &c));
    
    return FALSE;
}
		       
static long 
read_long( char *nptr, char **endptr, int base) {
    long n;

    n = strtol( nptr, endptr, 10);
    if ( errno == ERANGE) {
	g_warning( "xpm: number is too big");
	return -1;
    }

    return n;
}

